/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/**
 * @file c_rohc_v2.h
 * @brief ROHC compression context for the RoHCv2 profile.
 * @author Didier Barvaux <didier.barvaux@toulouse.viveris.com>
 * @author The hackers from ROHC for Linux
 * @author FWX <rohc_team@dialine.fr>
 */

#ifndef C_ROHC_V2_H
#define C_ROHC_V2_H

#include <netinet/ip.h>

#include "rohc_v2.h"

#include "c_generic.h"
#include "ts_sc_comp.h"
#include <sys/timeb.h>

/**
 * @brief Define the RoHCv2-specific temporary variables in the profile compression
 *        context.
 *
 * This object must be used by the RoHCv2-specific compression context sc_v2_context.
 *
 * @see sc_v2_context
 */
#ifdef LMQJSLMQJs  // DBX
struct v2_tmp_variables
{
	/// The number of TCP fields that changed in the TCP header
	int send_tcp_dynamic;
};
#endif

#define MAX_IPV6_OPTION_LENGTH        6   // FOR Destination/Hop-by-Hop/Routing/ah
#define MAX_IPV6_CONTEXT_OPTION_SIZE  (2 + ((MAX_IPV6_OPTION_LENGTH + 1) << 3))

/**
 * @brief Define the IPv6 generic option context.
 *
 */

typedef struct __attribute__((packed)) ipv6_option_context
{
	u_int8_t context_length;
	u_int8_t option_length;

	u_int8_t next_header;
	u_int8_t length;

	u_int8_t value[1];
} ipv6_option_context_t;

/**
 * @brief Define the IPv6 GRE option context.
 *
 */

typedef struct __attribute__((packed)) ipv6_gre_option_context
{
	u_int8_t context_length;
	u_int8_t option_length;

	u_int8_t next_header;

	u_int8_t c_flag : 1;
	u_int8_t k_flag : 1;
	u_int8_t s_flag : 1;
	u_int8_t padding : 5;

	u_int16_t protocol;

	u_int32_t key;               // if k_flag set
	u_int32_t sequence_number;   // if s_flag set

} ipv6_gre_option_context_t;

/**
 * @brief Define the IPv6 MIME option context.
 *
 */

typedef struct __attribute__((packed)) ipv6_mime_option_context
{
	u_int8_t context_length;
	u_int8_t option_length;

	u_int8_t next_header;

	u_int8_t s_bit : 1;
	u_int8_t res_bits : 7;
	u_int16_t checksum;
	u_int32_t orig_dest;
	u_int32_t orig_src;         // if s_bit set

} ipv6_mime_option_context_t;

/**
 * @brief Define the IPv6 AH option context.
 *
 */

typedef struct __attribute__((packed)) ipv6_ah_option_context
{
	u_int8_t context_length;
	u_int8_t option_length;

	u_int8_t next_header;

	u_int8_t length;
	u_int32_t spi;
	u_int32_t sequence_number;
	u_int32_t icv[1];
} ipv6_ah_option_context_t;

/**
 * @brief Define the common IP header context to IPv4 and IPv6.
 *
 */

typedef struct __attribute__((packed)) ipvx_context
{
	u_int8_t version : 4;
	u_int8_t unused : 1;
	u_int8_t inner : 3;

	u_int8_t next_header;

	u_int8_t tos_tc;
	u_int8_t ttl_hopl;

	u_int8_t ip_id_behavior;
	u_int8_t last_ip_id_behavior;

} ipvx_context_t;

/**
 * @brief Define the IPv4 header context.
 *
 */

typedef struct __attribute__((packed)) ipv4_context
{
	u_int8_t version : 4;
	u_int8_t df : 1;
	u_int8_t inner : 3;

	u_int8_t protocol;

	u_int8_t tos_tc;
	u_int8_t ttl_hopl;

	u_int8_t ip_id_behavior;
	u_int8_t last_ip_id_behavior;
	WB_t last_ip_id;

	u_int32_t src_addr;
	u_int32_t dst_addr;

} ipv4_context_t;

/**
 * @brief Define the IPv6 header context.
 *
 */

typedef struct __attribute__((packed)) ipv6_context
{
	u_int8_t version : 4;
	u_int8_t unused : 1;
	u_int8_t inner : 3;

	u_int8_t next_header;

#if __BYTE_ORDER == __LITTLE_ENDIAN
	u_int8_t tos_tc1 : 4;
	u_int8_t tos_tc2 : 4;
#elif __BYTE_ORDER == __BIG_ENDIAN
	u_int8_t tos_tc;
#endif
	u_int8_t ttl_hopl;

	u_int8_t ip_id_behavior;
	u_int8_t last_ip_id_behavior;

	u_int8_t flow_label1 : 4;
	u_int16_t flow_label2;

	u_int32_t src_addr[4];
	u_int32_t dest_addr[4];

} ipv6_context_t;

/**
 * @brief Define the IP-Only header context.
 *
 */

typedef struct __attribute__((packed)) ip_context
{
	u_int8_t protocol;

} ip_context_t;

/**
 * @brief Define the UDP context
 *
 */

typedef struct __attribute__((packed)) udp_context
{
	u_int8_t protocol;

	u_int16_t src_port;
	u_int16_t dst_port;

	u_int8_t checksum_used;

} udp_context_t;

/**
 * @brief Define the RTP context
 *
 */

typedef struct __attribute__((packed)) rtp_context
{
	u_int8_t protocol;

	u_int32_t ssrc;

	u_int32_t ts_stride;
	u_int32_t time_stride;
	u_int32_t last_ts_stride;
	u_int32_t last_time_stride;
	u_int32_t ts_scaled;
	u_int32_t ts_offset;

	u_int32_t last_timestamp;    // A VOIR

	u_int8_t marker : 1;
	u_int8_t pad_bit : 1;
	u_int8_t extension : 1;
	u_int8_t reserved1 : 5;
	u_int8_t reserved2 : 1;
	u_int8_t payload_type : 7;

	u_int8_t max_csrc_index;
	u_int32_t csrc_list[16];

#ifdef LKQSKQJHKSh
// cf c_rtp.h, struct rtp_tmp_variables :

	/// The number of UDP/RTP fields that changed in the UDP/RTP headers
	int send_rtp_dynamic;

	/// The number of bits needed to encode ts_send
	size_t nr_ts_bits;

	/// The number of bits of TS to place in the extension 3 header
	size_t nr_ts_bits_ext3;

	/// The real timestamp of the last RTP message
	uint32_t timestamp;

	/// The TS field to send (ts_scaled or ts)
	uint32_t ts_send;

	/// Whether the M bit is set in the RTP header or not
	int m_set;

	/// Whether the Payload Type (PT) field changed or not
	int rtp_pt_changed;
#endif

//	struct ts_sc_comp ts_comp;

	int nb_element_in_window;
	int put_in_window;
	int get_in_window;

#define TS_WINDOW_SIZE 4  // Power of 2!
	struct timeb timeb_in_window[TS_WINDOW_SIZE];
	u_int32_t timestamp_in_window[TS_WINDOW_SIZE];
	u_int16_t sequence_number_in_window[TS_WINDOW_SIZE];

} rtp_context_t;

/**
 * @brief Define the UDP-Lite context
 *
 */

typedef struct __attribute__((packed)) udp_lite_context
{
	u_int8_t protocol;

	u_int16_t src_port;
	u_int16_t dst_port;

	u_int8_t checksum_used;

	u_int8_t coverage_behavior;
	u_int8_t last_coverage_behavior;

	u_int16_t last_checksum_coverage;
} udp_lite_context_t;

/**
 * @brief Define the ESP context
 *
 */

typedef struct __attribute__((packed)) esp_context
{
	u_int8_t protocol;

	u_int32_t spi;

	u_int32_t sequence_number;

} esp_context_t;

/**
 * @brief Define union of IP contexts pointers.
 *
 */

typedef union
{
	u_int8_t *uint8;
	ipvx_context_t *vx;
	ipv4_context_t *v4;
	ipv6_context_t *v6;
	ipv6_option_context_t *v6_option;
	ipv6_gre_option_context_t *v6_gre_option;
	ipv6_mime_option_context_t *v6_mime_option;
	ipv6_ah_option_context_t *v6_ah_option;
	ip_context_t *ip;
	udp_context_t *udp;
	rtp_context_t *rtp;
	udp_lite_context_t *udp_lite;
	esp_context_t *esp;
} ip_context_ptr_t;

/**
 * @brief Define the RoHCv2 part of the profile compression context.
 *
 * This object must be used with the generic part of the compression
 * context c_generic_context.
 *
 * @see c_generic_context
 */
struct sc_v2_context
{
	/// The number of times the sequence number field was added to the compressed header
	int v2_seq_number_change_count;

	// The Master Sequence Number
	u_int16_t msn;

	u_int8_t reorder_ratio;

	u_int32_t v2_last_seq_number;

	/// @brief TCP-specific temporary variables that are used during one single
	///        compression of packet
// DBX	struct v2_tmp_variables tmp_variables;

	u_int8_t ip_context[1];
};

typedef struct
{
	struct c_profile profile;
	int context_size;
} c_v2_profile_t;

/*
 * Function prototypes.
 */

int c_rohc_v2_create(struct c_context *const context, const struct ip_packet *ip);

int c_rohc_v2_check_context(const struct c_context *context,
                            const struct ip_packet *ip);

int c_rohc_v2_encode(struct c_context *const context,
                     const struct ip_packet *ip,
                     const int packet_size,
                     unsigned char *const dest,
                     const int dest_size,
                     rohc_packet_t *const packet_type,
                     int *const payload_offset);

void tcp_decide_state(struct c_context *const context);

#endif

