/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/**
 * @file rfc5225_encoding.c
 * @brief Library of encoding methods from RFC4997, RFC4996 and RFC5225
 * @author FWX <rohc_team@dialine.fr>
 */

#include "c_generic.h"
#include "config.h" /* for RTP_BIT_TYPE and ROHC_DEBUG_LEVEL definitions */

#if SUPPORT_ROHC_V2

#include "rohc_traces.h"
#include "rohc_debug.h"
#include "rohc_bit_ops.h"
#include "crc.h"

//#include "tcp.h" // For IP_ID_BEHAVIOR
#include "rohc_v2.h"
#include "c_rohc_v2.h"
#include "rfc5225_encoding.h"
#include "trace.h" // FWX2

#include <math.h>
#include <assert.h>


#ifdef CODE_ALREADY_DEFINED

/**
 * @brief Table of the mask for lsb()
 */

unsigned int lsb_masks[] =
{
	0x00000,
	0x00001, 0x00003, 0x00007, 0x0000F,
	0x0001F, 0x0003F, 0x0007F, 0x000FF,
	0x001FF, 0x003FF, 0x007FF, 0x00FFF,
	0x01FFF, 0x03FFF, 0x07FFF, 0x0FFFF,
	0x1FFFF, 0x3FFFF, 0x7FFFF, 0xFFFFF
};

/* idem lsb_masks[]
static unsigned int lsb_power[] =
{
   ( 1 << 0 ) - 1,
   ( 1 << 1 ) - 1,
   ( 1 << 2 ) - 1,
   ( 1 << 3 ) - 1,
   ( 1 << 4 ) - 1,
   ( 1 << 5 ) - 1,
   ( 1 << 6 ) - 1,
   ( 1 << 7 ) - 1,
   ( 1 << 8 ) - 1,
   ( 1 << 9 ) - 1,
   ( 1 << 10 ) - 1,
   ( 1 << 11 ) - 1,
   ( 1 << 12 ) - 1,
   ( 1 << 13 ) - 1,
   ( 1 << 14 ) - 1,
   ( 1 << 15 ) - 1,
   ( 1 << 16 ) - 1,
   ( 1 << 17 ) - 1,
   ( 1 << 18 ) - 1
};
*/

/**
 * @brief Compress the lower bits of the given value.
 *
 * See RFC4997 page 27
 *
 * @param num_lsbs_param   The number of bits
 * @param offset_param     The offset
 * @param context_value    The value of the context
 * @param original_value   The value to compress
 * @return                 The compressed value with num_lsbs_param bits
 */

u_int32_t c_lsb( int num_lsbs_param, unsigned int offset_param, unsigned int context_value,
                 unsigned int original_value )
{
	unsigned int lower_bound;
	unsigned int upper_bound;
	unsigned int value;

	rohc_debugf(3, "c_lsb() num_lsb %d offset_param %d context_value %Xh original_value %Xh\n",
	            num_lsbs_param,offset_param,context_value,original_value);

	assert( num_lsbs_param > 0 && num_lsbs_param <= 18 );

	lower_bound = context_value - offset_param;
//	upper_bound = context_value + lsb_power[num_lsbs_param] - offset_param;
	upper_bound = context_value + lsb_masks[num_lsbs_param] - offset_param;

	value = original_value & lsb_masks[num_lsbs_param];

//	rohc_debugf(3, "c_lsb() %u < value %u < %d return %u\n",lower_bound,original_value,upper_bound,value);
	rohc_debugf(3, "c_lsb() %Xh < value %Xh < %Xh return %Xh\n",lower_bound,original_value,
	            upper_bound,
	            value);

//	assert( ( ( context_value & ( ~lsb_masks[num_lsbs_param] ) ) | value ) >= lower_bound );
//	assert( ( ( context_value & ( ~lsb_masks[num_lsbs_param] ) ) | value ) <= upper_bound );
//	assert( ( context_value - value ) >= lower_bound );
//	assert( ( context_value + value ) <= upper_bound );

//	assert( original_value >= lower_bound );
//	assert( original_value <= upper_bound );

	return value;
}


#endif

/**
 * @brief Compress the 8 bits given, depending of the context value.
 *
 * See RFC4996 page 46
 *
 * @param pmptr            The destination pointer, where to store the compressed value
 * @param context_value    The value of the context
 * @param value            The value to compress
 * @return                 0 if context value equal current value, 1 otherwise
 */

int c2_static_or_irreg8( multi_ptr_t *pmptr, u_int8_t *context_value, u_int8_t value )
{
	if(value == *context_value)
	{
		return 0;
	}
	else
	{
		*(pmptr->uint8)++ = value;
		*context_value = value;
		return 1;
	}
}


#ifdef CODE_UNUSED

/**
 * @brief Compress the 16 bits given, depending of the context value.
 *
 * @param pmptr            The destination pointer, where to store the compressed value
 * @param context_value    The value of the context
 * @param value            The value to compress
 * @return                 0 if context value equal current value, 1 otherwise
 */

int c_static_or_irreg16( multi_ptr_t *pmptr, u_int16_t *context_value, u_int16_t value )
{
	if(value == *context_value)
	{
		return 0;
	}
	else
	{
		WRITE16_TO_PMPTR(pmptr,value);
		*context_value = value;
		return 1;
	}
}


#endif

#ifdef CODE_UNUSED

/**
 * @brief Compress the 8 bits value, regarding if null or not
 *
 * @param pmptr            The destination of the compress value
 * @param value            The value to compress
 * @return                 1 if null value, 0 otherwise
 */

u_int8_t c_zero_or_irreg8( multi_ptr_t *pmptr, u_int8_t value )
{
	if(value != 0)
	{
		*(pmptr->uint8)++ = value;
		return 0;
	}
	else
	{
		return 1;
	}
}


#endif

#ifdef CODE_UNUSED

/**
 * @brief Compress the 16 bits value, regarding if null or not
 *
 * @param pmptr            The destination of the compress value
 * @param value            The value to compress
 * @return                 1 if null value, 0 otherwise
 */

u_int16_t c_zero_or_irreg16( multi_ptr_t *pmptr, u_int16_t value )
{
	if(value != 0)
	{
		WRITE16_TO_PMPTR(pmptr,value);
		return 0;
	}
	else
	{
		return 1;
	}
}


#endif

#ifdef CODE_ALREADY_DEFINED

/**
 * @brief Compress a 32 bits value
 *
 * See RFC4996 page 46
 *
 * @param pmptr            The destination for the compressed value
 * @param puint32          Pointer to the 32 bits value to compress
 * @return                 Size of the compressed value, in octets
 */

unsigned int variable_length_32_enc( multi_ptr_t *pmptr, u_int32_t *puint32 )
{
	multi_ptr_t mptr;

	if(*puint32 == 0)
	{
		return 0;
	}
	mptr.uint32 = puint32;
	if(READNI16_FROM_MPTR(mptr) == 0)
	{
		++mptr.uint16;
		if(*mptr.uint8 == 0)
		{
			*(pmptr->uint8)++ = *(++mptr.uint8);
			return 1;
		}
		else
		{
			WRITE16_TO_PMPTR(pmptr,READNI16_FROM_MPTR(mptr));
			return 2;
		}
	}
	else
	{
		WRITE32_TO_PMPTR(pmptr,READNI32_FROM_MPTR(mptr));
		return 3;
	}
}


#endif

#ifdef CODE_ALREADY_DEFINED

/**
 * @brief Compress a 32 bits value, regarding the context value
 *
 * See RFC4996 page 47
 *
 * @param pmptr            The destination for the compressed value
 * @param context_value    The context value
 * @param value            The value to compress
 * @return                 Indicator 1 if compressed, 0 if same than context value
 */

unsigned int c_optional32( multi_ptr_t *pmptr, u_int32_t context_value, u_int32_t value )
{
	if(value == context_value)
	{
		return 0;
	}
	WRITE32_TO_PMPTR(pmptr,value);
	return 1;
}


#endif

#ifdef CODE_ALREADY_DEFINED

/**
 * @brief Compress a 32 bits value to 7 or 31 bits
 *
 * See RFC4996 page 47
 *
 * @param pmptr            The destination for the compressed value
 * @param value            The value to compress
 * @return                 Nothing
 */

void c_lsb_7_31( multi_ptr_t *pmptr, u_int32_t value )
{
	if(value > 0x7F)
	{
		WRITE32_TO_PMPTR(pmptr, ( htonl(value) & htonl(0x7FFFFFFF) ) | htonl(0x80000000) );
	}
	else
	{
		*(pmptr->uint8++) = value & 0x7F;
	}
}


#endif

#ifndef USE_ROHC_V2_MACROS

/**
 * @brief Calculate the scaled and residue values from unscaled value and scaling factor
 *
 * See RFC4996 page 49
 *
 * @param scaled_value     The scaled value
 * @param residue_field    The residue field of the compressed value
 * @param scaling_factor   The scaling factor
 * @param unscaled_value   The value to scale
 * @return                 Nothing
 */

void c_field_scaling( u_int32_t *scaled_value, u_int32_t *residue_field, u_int32_t scaling_factor,
                      u_int32_t unscaled_value )
{
	if(scaling_factor == 0)
	{
		*residue_field = unscaled_value;
		*scaled_value = 0;
	}
	else
	{
		*residue_field = unscaled_value % scaling_factor;
		*scaled_value = unscaled_value / scaling_factor;
		assert( unscaled_value == ( *scaled_value * scaling_factor ) + *residue_field );
	}
}


#endif

#ifdef TCP_H

/**
 * @brief Calculate the rsf_index from the rsf flags
 *
 * See RFC4996 page 71
 *
 * @param rsf_flags        The RSF flags
 * @return                 The rsf index
 */

unsigned int rsf_index_enc( unsigned int rsf_flags )
{
	switch(rsf_flags)
	{
		case RSF_NONE:
			return 0;
		case RSF_RST_ONLY:
			return 1;
		case RSF_SYN_ONLY:
			return 2;
		case RSF_FIN_ONLY:
			return 3;
		default:
			rohc_debugf(3, "TCP RSF_UNKNOWN!\n");
			return 0;
	}
}


#endif

#ifdef CODE_ALREADY_DEFINED

/**
 * @brief Compress the lower bits of IP-ID
 *
 * See RFC4996 page 75
 *
 * @param behavior         The IP-ID behavior
 * @param k                The num_lsbs_param parameter for c_lsb()
 * @param p                The offset parameter for c_lsb()
 * @param context_ip_id    The context value of IP-ID
 * @param ip_id            The IP-ID value to compress
 * @param msn              The Master Sequence Number
 * @return                 The lsb of offset between IP-ID and MSN
 */

unsigned int c_ip_id_lsb( int behavior, unsigned int k, unsigned int p, WB_t context_ip_id,
                          WB_t ip_id,
                          u_int16_t msn )
{
	u_int16_t ip_id_offset;
	WB_t ip_id_nbo;
	WB_t swapped_context_ip_id;

	assert( behavior == IP_ID_BEHAVIOR_SEQUENTIAL ||
	        behavior == IP_ID_BEHAVIOR_SEQUENTIAL_SWAPPED );

	rohc_debugf(3, "behavior %d context_ip_id %4.4Xh ip_id=%4.4Xh msn=%4.4Xh\n",behavior,
	            context_ip_id.uint16,ip_id.uint16,
	            msn);

	switch(behavior)
	{
		case IP_ID_BEHAVIOR_SEQUENTIAL:
			ip_id_offset = ip_id.uint16 - msn;
			rohc_debugf(3, "ip_id_offset = %Xh - %Xh = %Xh\n",ip_id.uint16,msn,ip_id_offset);
			//  rohc_debugf(3, "context_ip_id - msn = %Xh - %Xh = %Xh\n",context_ip_id,msn,context_ip_id-msn);
			//   ip_id_offset = c_lsb( k , p , context_ip_id - msn , ip_id_offset );
			ip_id_offset = c_lsb( k, p, context_ip_id.uint16 - msn, ip_id_offset );
			rohc_debugf(3, "ip_id_offset = %Xh\n",ip_id_offset);
			break;
		case IP_ID_BEHAVIOR_SEQUENTIAL_SWAPPED:
			// ip_id_nbo = ( ip_id / 256 ) + ( ( ip_id % 256 ) * 256 );
			ip_id_nbo.uint8[0] = ip_id.uint8[1];
			ip_id_nbo.uint8[1] = ip_id.uint8[0];
			rohc_debugf(3, "ip_id_nbo = %Xh\n",ip_id_nbo.uint16);
			ip_id_offset = ip_id_nbo.uint16 - msn;
			rohc_debugf(3, "ip_id_offset = %Xh\n",ip_id_offset);
			swapped_context_ip_id.uint8[0] = context_ip_id.uint8[1];
			swapped_context_ip_id.uint8[1] = context_ip_id.uint8[0];
			ip_id_offset = c_lsb( k, p, swapped_context_ip_id.uint16 - msn, ip_id_offset );
			rohc_debugf(3, "ip_id_offset = %Xh\n",ip_id_offset);
			break;
	}
	return ip_id_offset;
}


#endif

#ifdef CODE_ALREADY_DEFINED

/**
 * @brief Compress or not the IP-ID
 *
 * See RFC4996 page 76
 *
 * @param pmptr            The destination for the compressed value
 * @param behavior         The IP-ID behavior
 * @param context_ip_id    The context value of IP-ID
 * @param ip_id            The IP-ID value to compress
 * @param msn              The Master Sequence Number
 * @return                 Indicator : 0 if short, 1 if long
 */

unsigned int c_optional_ip_id_lsb( multi_ptr_t *pmptr, int behavior, WB_t context_ip_id, WB_t ip_id,
                                   u_int16_t msn )
{

	rohc_debugf(3, "behavior %Xh context_ip_id %Xh ip_id %Xh msn %Xh\n",behavior,
	            context_ip_id.uint16,ip_id.uint16,
	            msn);

	switch(behavior)
	{
		case IP_ID_BEHAVIOR_SEQUENTIAL:
			if( ( context_ip_id.uint16 & 0xFF00 ) == ( ip_id.uint16 & 0xFF00 ) )
			{
				*(pmptr->uint8)++ = c_ip_id_lsb(behavior,8,3,context_ip_id,ip_id,msn);
				rohc_debugf(3, "write ip_id %Xh\n",*(pmptr->uint8 - 1));
				return 0;
			}
			else
			{
				WRITE16_TO_PMPTR(pmptr,htons(ip_id.uint16));
				rohc_debugf(3, "write ip_id %Xh\n",ip_id.uint16);
				return 1;
			}
			break;
		case IP_ID_BEHAVIOR_SEQUENTIAL_SWAPPED:
			if( ( context_ip_id.uint16 & 0x00FF ) == ( ip_id.uint16 & 0x00FF ) )
			{
				*(pmptr->uint8)++ = c_ip_id_lsb(behavior,8,3,context_ip_id,ip_id,msn);
				rohc_debugf(3, "write ip_id %Xh\n",*(pmptr->uint8 - 1));
				return 0;
			}
			else
			{
				WB_t swapped_ip_id;
				swapped_ip_id.uint8[0] = ip_id.uint8[1];
				swapped_ip_id.uint8[1] = ip_id.uint8[0];
				WRITE16_TO_PMPTR(pmptr,htons(swapped_ip_id.uint16));
				rohc_debugf(3, "write ip_id %Xh\n",swapped_ip_id.uint16);
				return 1;
			}
			break;
		case IP_ID_BEHAVIOR_RANDOM:
		case IP_ID_BEHAVIOR_ZERO:
		default:
			break;
	}

	return 0;
}


#endif

/**
 * @brief Compress or not the IP-ID
 *
 * See RFC5225 page 70
 *
 * @param pmptr            The destination for the compressed value
 * @param behavior         The IP-ID behavior
 * @param context_ip_id    The context value of IP-ID
 * @param ip_id            The IP-ID value to compress
 * @param msn              The Master Sequence Number
 * @return                 Indicator : 0 if short, 1 if long
 */

int c_ip_id_sequential_variable( multi_ptr_t *pmptr, int behavior, WB_t context_ip_id, WB_t ip_id,
                                 u_int16_t msn )
{
	rohc_debugf(3, "behavior %Xh context_ip_id %Xh ip_id %Xh msn %Xh\n",behavior,
	            context_ip_id.uint16,ip_id.uint16,
	            msn);

	switch(behavior)
	{
		case IP_ID_BEHAVIOR_SEQUENTIAL:
			if( ( context_ip_id.uint16 & 0xFF00 ) == ( ip_id.uint16 & 0xFF00 ) )
			{
				// =:= ip_id_lsb(behavior,8);
				*(pmptr->uint8)++ = c_ip_id_lsb(behavior,8,3,context_ip_id,ip_id,msn);
				rohc_debugf(3, "write ip_id %Xh\n",*(pmptr->uint8 - 1));
				return 0;
			}
			else
			{
				// =:= irregular(16)
				WRITE16_TO_PMPTR(pmptr,htons(ip_id.uint16));
				rohc_debugf(3, "write ip_id %Xh\n",ip_id.uint16);
				return 1;
			}
			break;
		case IP_ID_BEHAVIOR_SEQUENTIAL_SWAPPED:
			if( ( context_ip_id.uint16 & 0x00FF ) == ( ip_id.uint16 & 0x00FF ) )
			{
				// =:= ip_id_lsb(behavior,8);
				*(pmptr->uint8)++ = c_ip_id_lsb(behavior,8,3,context_ip_id,ip_id,msn);
				rohc_debugf(3, "write ip_id %Xh\n",*(pmptr->uint8 - 1));
				return 0;
			}
			else
			{
				WB_t swapped_ip_id;
				swapped_ip_id.uint8[0] = ip_id.uint8[1];
				swapped_ip_id.uint8[1] = ip_id.uint8[0];
				// =:= irregular(16)
				WRITE16_TO_PMPTR(pmptr,htons(swapped_ip_id.uint16));
				rohc_debugf(3, "write ip_id %Xh\n",swapped_ip_id.uint16);
				return 1;
			}
			break;
		case IP_ID_BEHAVIOR_RANDOM:
		case IP_ID_BEHAVIOR_ZERO:
		default:
			return 0;
	}
}


#ifndef USE_ROHC_V2_MACROS

/**
 * @brief Calculate the ip_id irregular encodage
 *
 * See RFC5225 page 60
 *
 * @param mptr            The destination for the compressed value
 * @param behavior        The ip_id behavior
 * @param ip_id           The value of ip_id
 * @return                Pointer to the next location
 */

// See RFC5225 page 60

u_int8_t * c_ip_id_enc_irreg( multi_ptr_t mptr, int behavior, u_int16_t ip_id )
{
	switch(behavior)
	{
		case IP_ID_BEHAVIOR_SEQUENTIAL:
			// =:= ip_id_seq
			break;
		case IP_ID_BEHAVIOR_SEQUENTIAL_SWAPPED:
			// =:= ip_id_seq_swapped
			break;
		case IP_ID_BEHAVIOR_RANDOM:
			// =:= ip_id_rand
			WRITE16_TO_MPTR(mptr,ip_id);
			rohc_debugf(3, "Add ip_id %4.4Xh\n",ntohs(ip_id));
			break;
		case IP_ID_BEHAVIOR_ZERO:
			// =:= ip_id_zero
			break;
	}

	return mptr.uint8;
}


#endif

#ifdef CODE_ALREADY_DEFINED

/**
 * @brief Encode the DSCP field
 *
 * See RFC4996 page 75
 *
 * @param pmptr            The destination for the compressed value
 * @param context_value    The context value of DSCP
 * @param value            The DSCP value to compress
 * @return                 Indicator 1 if value compressed, 0 otherwise
 */


unsigned int dscp_encode( multi_ptr_t *pmptr, u_int8_t context_value, u_int8_t value )
{
	if(value == context_value)
	{
		return 0;
	}
	// 6 bits + 2 bits padding
	*(pmptr->uint8)++ = value & 0x3F;
	return 1;
}


#endif

/**
 * @brief Encode the Master Sequence Number
 *
 * See RFC5225 page 69
 *
 * @param k             The number of bits
 * @param v2_context    The specific TCP context
 * @return              The MSN compressed value
 */

//#pragma GCC diagnostic push
//#pragma GCC diagnostic error "-Wuninitialized"
//#pragma GCC diagnostic ignored "-Wuninitialized"

unsigned int c_msn_lsb( int k, struct sc_v2_context *v2_context )
{
	unsigned int master;

	switch(v2_context->reorder_ratio)
	{
		case REORDERING_NONE:
			master = c_lsb(k,1,v2_context->msn,v2_context->msn);
			break;
		case REORDERING_QUATER:
			master = c_lsb(k,((1 << k) / 4) - 1,v2_context->msn,v2_context->msn);
			break;
		case REORDERING_HALF:
			master = c_lsb(k,((1 << k) / 2) - 1,v2_context->msn,v2_context->msn);
			break;
		case REORDERING_THREEQUARTERS:
			master = c_lsb(k,(((1 << k) * 3) / 4) - 1,v2_context->msn,v2_context->msn);
			break;
//		default :
//			                    master = 0;
//			                    break;
	}

	rohc_debugf(3, "return %Xh\n",master);
	return master;
}


//#pragma GCC diagnostic pop

void c_optional_32( multi_ptr_t *pmptr, u_int32_t value, int flag )
{
	if(flag == 1)
	{
		WRITE32_TO_PMPTR(pmptr,value);
	}
}


// Send the entire value or keep previous value
// See RFC5225 page 48

int c_sdvl_or_static( multi_ptr_t *pmptr, u_int32_t static_value, u_int32_t value )
{
	#ifdef ROHCV2_DEBUG
	u_int8_t *ptr = pmptr->uint8;
	#endif

	rohc_debugf(3, "mptr %p static_value %d value %d\n",pmptr->uint8,static_value,value);

	do
	{

		if(value == static_value)
		{
			rohc_debugf(3, "value %Xh is static\n",value);
			return 0;
		}
		// If value < 2^7
		if(value < (1 << 7) )
		{
			rohc_debugf(3, "value %Xh < (2^7)\n",value);
			// Discriminator =:= '0'
			*(pmptr->uint8++) = value;
			break;
		}
		// If value < 2^14
		if(value < (1 << 14) )
		{
			rohc_debugf(3, "value %Xh < (2^14)\n",value);
			// Discriminator =:= '10'
			*(pmptr->uint8++) = 0x80 | ( value >> 8 );
			*(pmptr->uint8++) = value;
			break;
		}
		// If value < 2^21
		if(value < (1 << 21) )
		{
			rohc_debugf(3, "value %Xh < (2^21)\n",value);
			// Discriminator =:= '110'
			*(pmptr->uint8++) = 0xC0 | ( value >> 16 );
			*(pmptr->uint8++) = value >> 8;
			*(pmptr->uint8++) = value;
			break;
		}
		// If value < 2^28
		if(value < (1 << 28) )
		{
			rohc_debugf(3, "value %Xh < (2^28)\n",value);
			// Discriminator =:= '1110'
			*(pmptr->uint8++) = 0xE0 | ( value >> 24 );
			*(pmptr->uint8++) = value >> 16;
			*(pmptr->uint8++) = value >> 8;
			*(pmptr->uint8++) = value;
			break;
		}
		rohc_debugf(3, "value %Xh\n",value);
		// Discriminator =:= '11111111'
		*(pmptr->uint8++) = 0xFF;
		WRITE32_TO_PMPTR(pmptr,htonl(value));
		/*
		*(pmptr->uint8++) = value >> 24;
		*(pmptr->uint8++) = value >> 16;
		*(pmptr->uint8++) = value >> 8;
		*(pmptr->uint8++) = value;
		*/

	}
	while(0);

	#ifdef ROHCV2_DEBUG
	rohc_debugf(3, "c_sdvl_or_static(%Xh) return length %d\n",value,(int)(pmptr->uint8 - ptr));
	TraceData(ptr,pmptr->uint8 - ptr);
	#endif

	return 1;
}


// Send the entire value or revert to default value
// See RFC5225 page 49

int c_sdvl_or_default( multi_ptr_t *pmptr, u_int32_t default_value, u_int32_t value )
{
	#ifdef ROHCV2_DEBUG
	u_int8_t *ptr = pmptr->uint8;
	#endif

	rohc_debugf(3, "mptr %p default_value %d value %d\n",pmptr->uint8,default_value,value);

	do
	{

		if(value == default_value)
		{
			// Default value
			return 0;
		}
		// If value < (2^7)
		if(value < (1 << 7) )
		{
			// Discriminator =:= '0'
			*(pmptr->uint8++) = value;
			break;
		}
		// If value < (2^14)
		if(value < (1 << 14) )
		{
			// Discriminator =:= '10'
			*(pmptr->uint8++) = 0x80 | ( value >> 8 );
			*(pmptr->uint8++) = value;
			break;
		}
		// If value < (2^21)
		if(value < (1 << 21) )
		{
			// Discriminator =:= '110'
			*(pmptr->uint8++) = 0xC0 | ( value >> 16 );
			*(pmptr->uint8++) = value >> 8;
			*(pmptr->uint8++) = value;
			break;
		}
		// If value < (2^28)
		if(value < (1 << 28) )
		{
			// Discriminator =:= '1110'
			*(pmptr->uint8++) = 0xE0 | ( value >> 24 );
			*(pmptr->uint8++) = value >> 16;
			*(pmptr->uint8++) = value >> 8;
			*(pmptr->uint8++) = value;
			break;
		}
		// Discriminator =:= '11111111'
		*(pmptr->uint8++) = 0xFF;
		WRITE32_TO_PMPTR(pmptr,htonl(value));
		/*
		*(pmptr->uint8++) = value >> 24;
		*(pmptr->uint8++) = value >> 16;
		*(pmptr->uint8++) = value >> 8;
		*(pmptr->uint8++) = value;
		*/

	}
	while(0);

	#ifdef ROHCV2_DEBUG
	rohc_debugf(3, "c_sdvl_or_default(%Xh) return length %d\n",value,(int)(pmptr->uint8 - ptr));
	TraceData(ptr,pmptr->uint8 - ptr);
	#endif

	return 1;
}


// Self-describing variable length encoding with reordering offset
// See RFC5225 page 72

u_int8_t * c_sdvl_sn_lsb( multi_ptr_t mptr, u_int32_t value )
{
	#ifdef ROHCV2_DEBUG
	u_int8_t *ptr = mptr.uint8;
	#endif

	rohc_debugf(3, "mptr %p value %d %Xh\n",mptr.uint8,value,value);

	do
	{
		if(value < 0x80)
		{
			// Discriminator =:= '0'
			*(mptr.uint8++) = value;
			break;
		}
		if(value < 0x4000)
		{
			// Discriminator =:= '10'
			*(mptr.uint8++) = 0x80 | ( value >> 8 );
			*(mptr.uint8++) = value;
			break;
		}
		if(value < 0x200000)
		{
			// Discriminator =:= '110'
			*(mptr.uint8++) = 0xC0 | ( value >> 16 );
			WRITE16_TO_MPTR(mptr,(htons(value)) & 0xFFFF);
			break;
		}
		if(value < 0x10000000)
		{
			// Discriminator =:= '1110'
			*(mptr.uint8++) = 0xE0 | ( value >> 24 );
			WRITE16_TO_MPTR(mptr,(htons(value)) & 0xFFFF);
			break;
		}
		// Discriminator =:= '11111111'
		*(mptr.uint8++) = 0xFF;
		WRITE32_TO_MPTR(mptr,htonl(value));
		break;
	}
	while(0);

	#ifdef ROHCV2_DEBUG
	rohc_debugf(3, "c_sdvl_sn_lsb(%Xh) return length %d\n",value,(int)(mptr.uint8 - ptr));
	TraceData(ptr,mptr.uint8 - ptr);
	#endif

	rohc_debugf(3, "return mptr %p\n",mptr.uint8);
	return mptr.uint8;
}


void put_timestamp_in_window( rtp_context_t *rtp_context, u_int16_t sequence_number,
                              u_int32_t timestamp )
{

	rtp_context->timestamp_in_window[ rtp_context->put_in_window ] = timestamp;
	rtp_context->sequence_number_in_window[ rtp_context->put_in_window ] = sequence_number;
	rtp_context->put_in_window = (++rtp_context->put_in_window) & (TS_WINDOW_SIZE - 1);
	if(rtp_context->nb_element_in_window < TS_WINDOW_SIZE)
	{
		++rtp_context->nb_element_in_window;
	}
	else
	{
		rtp_context->get_in_window = (++rtp_context->get_in_window) & (TS_WINDOW_SIZE - 1);
	}
}


void put_rtp_frame_in_window( rtp_context_t *rtp_context, base_header_rtp_t *rtp )
{
	rtp_context->timestamp_in_window[ rtp_context->put_in_window ] = ntohl( rtp->timestamp );
	rtp_context->sequence_number_in_window[ rtp_context->put_in_window ] = ntohl(
	   rtp->sequence_number );
	ftime( &rtp_context->timeb_in_window[ rtp_context->put_in_window ] );

	rohc_debugf(3, "timestamp %Xh inserted at %d\n",
	            rtp_context->timestamp_in_window[ rtp_context->put_in_window ],
	            rtp_context->put_in_window);

	rtp_context->put_in_window = (++rtp_context->put_in_window) & (TS_WINDOW_SIZE - 1);
	if(rtp_context->nb_element_in_window < TS_WINDOW_SIZE)
	{
		++rtp_context->nb_element_in_window;
	}
	else
	{
		rtp_context->get_in_window = (++rtp_context->get_in_window) & (TS_WINDOW_SIZE - 1);
	}

}


// See RFC5225 section 6.6.9 page 30 and appendix B.3
u_int8_t * c_timer_based_lsb( multi_ptr_t mptr, rtp_context_t *rtp_context, int k, int p,
                              u_int32_t timestamp )
{
//	u_int32_t tsc_ref_advanced;
	u_int32_t value;
	u_int32_t mask;
	unsigned int min;
	unsigned int max;
	int i;
	int j;

	rohc_debugf(3, "mptr %p k %d p %d timestamp %Xh\n",mptr.uint8,k,p,timestamp);
	rohc_debugf(3, "rtp_context %p time_stride %d\n",rtp_context,rtp_context->time_stride);

	min = 0xFFFFFFFF;
	max = 0;

	for(i = rtp_context->nb_element_in_window, j = rtp_context->get_in_window; i != 0; --i)
	{
		if(rtp_context->timestamp_in_window[ i ] > max)
		{
			max = rtp_context->timestamp_in_window[ i ];
		}
		if(rtp_context->timestamp_in_window[ i ] < min)
		{
			min = rtp_context->timestamp_in_window[ i ];
		}
		j = (j + 1) & (TS_WINDOW_SIZE - 1);
	}

	rohc_debugf(3, "nb_element %d min %Xh max %Xh\n",rtp_context->nb_element_in_window,min,max);

	// Try to find min LSB bits
	value = timestamp;
	for(mask = 0xFFFFFFFF; mask != 0; mask <<= 1)
	{
		if( ( timestamp & mask ) == ( min & mask ) &&
		    ( timestamp & mask ) == (max & mask ) )
		{
			// Mask found
			value = timestamp & (~mask);
			break;
		}
	}

	// Lire l'heure
	// tsc_ref_advanced = tsc_ref + (a_n - a_ref) / time_stride
	// c_lsb(k,p);
	rohc_debugf(3, "return mptr %p\n",mptr.uint8);
	return mptr.uint8;
}


// See RFC5225 page 71
u_int32_t c_scaled_ts_lsb( multi_ptr_t mptr, rtp_context_t *rtp_context, int k, u_int32_t timestamp )
{
	u_int32_t value;

	rohc_debugf(3, "mptr %p k %d timestamp %Xh\n",mptr.uint8,k,timestamp);
	rohc_debugf(3, "rtp_context %p time_stride %d last_timestamp %Xh\n",rtp_context,
	            rtp_context->time_stride,
	            rtp_context->last_timestamp);

	if(rtp_context->time_stride != 0)
	{
		// timestamp =:= timer_based_lsb(time_stride_value, k, ((2^k) /2 -1)
		mptr.uint8 = c_timer_based_lsb(mptr,rtp_context,k,((1 << k) / 2) - 1,timestamp);
	}
	else
	{
		// timestamp =:= lsb(k, ((2^k) / 4) -1)
		value = c_lsb(k,((1 << k) / 4) - 1,rtp_context->last_timestamp,timestamp);

	}

	rohc_debugf(3, "return value %d %Xh\n",value,value);

	return value;
}


#define LSBX_MASK_NUMBER 4
static u_int32_t lsbx_masks[LSBX_MASK_NUMBER] = { 0xFFFFFF80, 0xFFFFC000, 0xFFE00000, 0xF0000000 };

// Self-describing variable length encoding
// RFC5225 page 73
u_int8_t * c_sdvl_lsb( multi_ptr_t mptr, int field_width, u_int32_t context_value, u_int32_t value )
{
	#ifdef ROHCV2_DEBUG
	u_int8_t *ptr = mptr.uint8;
	#endif
	u_int32_t field;
	int i;

	rohc_debugf(3, "mptr %p field_width %d context_value %Xh value %Xh\n",mptr.uint8,field_width,
	            context_value,
	            value);

	for(i = 0; i < LSBX_MASK_NUMBER; ++i)
	{
		field = value & lsbx_masks[i];
		rohc_debugf(3, "Try i %d mask %Xh field %Xh <-> %Xh\n",i,lsbx_masks[i],field,
		            context_value & lsbx_masks[i]);
		// If same higher bits
		if( ( context_value & lsbx_masks[i] ) == field)
		{
			field = value & (~lsbx_masks[i]);
			rohc_debugf(3, "Choose i %d field %Xh\n",i,field);
			switch(i)
			{
				case 0:
					assert( field < 0x80 );
					// discriminator =:= '0'
					*(mptr.uint8)++ = field;
					goto end_c_sdvl_lsb;
				case 1:
					assert( field < 0x4000 );
					// discriminator =:= '10'
					*(mptr.uint8)++ = 0x80 | ( field >> 8 );
					*(mptr.uint8)++ = field;
					goto end_c_sdvl_lsb;
				case 2:
					assert( field < 0x200000 );
					// discriminator =:= '110'
					*(mptr.uint8)++ = 0xC0 | ( field >> 16 );
					*(mptr.uint8)++ = field >> 8;
					*(mptr.uint8)++ = field;
					goto end_c_sdvl_lsb;
				case 4:
					assert( field < 0x10000000 );
					// discriminator =:= '1110'
					*(mptr.uint8)++ = 0xE0 | ( field >> 24 );
					*(mptr.uint8)++ = field >> 16;
					*(mptr.uint8)++ = field >> 8;
					*(mptr.uint8)++ = field;
					goto end_c_sdvl_lsb;
			}
		}
	}

	rohc_debugf(3, "Discriminator 0xFF\n");
	// discriminator =:= '11111111'
	*(mptr.uint8)++ = 0xFF;
	WRITE32_TO_MPTR(mptr,htonl(value));
	/*
	*(mptr.uint8)++ = value >> 24 );
	*(mptr.uint8)++ = value >> 16;
	*(mptr.uint8)++ = value >> 8;
	*(mptr.uint8)++ = value;
	*/

end_c_sdvl_lsb:

	#ifdef ROHCV2_DEBUG
	rohc_debugf(3, "Value coded:\n");
	TraceData(ptr,mptr.uint8 - ptr);
	#endif

	rohc_debugf(3, "return mptr %p\n",mptr.uint8);
	return mptr.uint8;
}


// RFC5225 page 73
u_int8_t * c_sdvl_scaled_ts_lsb( multi_ptr_t mptr, rtp_context_t *rtp_context, u_int32_t timestamp )
{
	u_int32_t value;

	rohc_debugf(3, "mptr %p timestamp %Xh\n",mptr.uint8,timestamp);

	// field =:= scaled_ts_lsb(time_stride,X)

//u_int32_t c_scaled_ts_lsb( multi_ptr_t mptr , u_int16_t time_stride_value , int k , u_int32_t context_timestamp , u_int32_t timestamp )

	int k = 7; // 7, 14, 21, 28
	value = c_scaled_ts_lsb(mptr,rtp_context,k,timestamp);

	do
	{
		if(value < 0x80)
		{
			// Discriminator '0'
			*(mptr.uint8++) = value;
			break;
		}
		if(value < 0x4000)
		{
			// Discriminator '10'
			*(mptr.uint8++) = 0x80 | ( value >> 8 );
			*(mptr.uint8++) = value;
			break;
		}
		if(value < 0x200000)
		{
			// Discriminator '110'
			*(mptr.uint8++) = 0xC0 | ( value >> 16 );
			*(mptr.uint8++) = value >> 8;
			*(mptr.uint8++) = value;
			break;
		}
		if(value < 0x10000000)
		{
			// Discriminator '1110'
			*(mptr.uint8++) = 0xE0 | ( value >> 24 );
			*(mptr.uint8++) = value >> 16;
			*(mptr.uint8++) = value >> 8;
			*(mptr.uint8++) = value;
			break;
		}
		// Discriminator '11111111'
		*(mptr.uint8++) = 0xFF;
		WRITE32_TO_MPTR(mptr,htonl(value));
		/*
		*(mptr.uint8++) = value >> 24;
		*(mptr.uint8++) = value >> 16;
		*(mptr.uint8++) = value >> 8;
		*(mptr.uint8++) = value;
		*/

	}
	while(0);

	rohc_debugf(3, "return mptr %p\n",mptr.uint8);
	return mptr.uint8;
}


// RFC5225 page 74
u_int8_t * c_variable_scaled_timestamp( multi_ptr_t mptr, rtp_context_t *rtp_context, int tss_flag,
                                        int tsc_flag,
                                        u_int32_t timestamp )
{
	rohc_debugf(3, "mptr %p tss_flag %d tsc_flag %d timestamp %Xh\n",mptr.uint8,tss_flag,tsc_flag,
	            timestamp);
	rohc_debugf(3, "rtp_context %p ts_stride %d\n",rtp_context,rtp_context->ts_stride);

	if(tsc_flag != 0 && tss_flag == 0 && rtp_context->ts_stride != 0)
	{
		// Present
		// ts_scaled à calculer depuis la fenetre ts_sc_comp
		// scaled_value =:= sdvl_scaled_ts_lsb(time_stride)
		mptr.uint8 = c_sdvl_scaled_ts_lsb(mptr,rtp_context,timestamp);
	}
	else
	{
		assert( tsc_flag == 0 );
		// Not present
		rohc_debugf(3, "Not present\n");
	}

	rohc_debugf(3, "return mptr %p\n",mptr.uint8);
	return mptr.uint8;
}


// RFC5225 page 74
u_int8_t * c_variable_unscaled_timestamp( multi_ptr_t mptr, int tss_flag, int tsc_flag,
                                          u_int32_t context_timestamp,
                                          u_int32_t timestamp )
{
	rohc_debugf(3, "mptr %p tss_flag %d tsc_flag %d context_timestamp %Xh timestamp %Xh\n",
	            mptr.uint8,tss_flag,tsc_flag,context_timestamp,
	            timestamp);

	if(tsc_flag == 0 /* ( tss_flag != 0 && tsc_flag == 0 ) || ( tss_flag == 0 && tsc_flag == 0 ) */)
	{
		// Present
		// timestamp =:= sdvl_lsb(32)
		mptr.uint8 = c_sdvl_lsb(mptr,32,context_timestamp,timestamp);
	}
	else
	{
		assert( tss_flag == 0 && tsc_flag != 0 );
		// Not present
		rohc_debugf(3, "Not present\n");
	}
	rohc_debugf(3, "return mptr %p\n",mptr.uint8);
	return mptr.uint8;
}


#endif // SUPPORT_ROHC_V2

