/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/**
 * @file d_rohc_v2.h
 * @brief ROHC decompression context for the RoHCv2 profile.
 * @author FWX <rohc_team@dialine.fr>
 */

#ifndef D_ROHC_V2_H
#define D_ROHC_V2_H

#include <netinet/ip.h>
#include <string.h>

#include "rohc_v2.h"
#include "d_generic.h"
#include "ts_sc_decomp.h"
#include "d_ip.h"

/**
 * @brief Define the IPv6 option context for Destination, Hop-by-Hop and Routing option
 *
 */

typedef struct __attribute__((packed)) ipv6_option_context
{
	u_int8_t context_length;
	u_int8_t option_length;

	u_int8_t next_header;
	u_int8_t length;

	u_int8_t value[6];

} ipv6_option_context_t;

/**
 * @brief Define the IPv6 option context for GRE option
 *
 */

typedef struct __attribute__((packed)) ipv6_gre_option_context
{
	u_int8_t context_length;
	u_int8_t option_length;

	u_int8_t next_header;

	u_int8_t c_flag : 1;
	u_int8_t k_flag : 1;
	u_int8_t s_flag : 1;
	u_int8_t protocol : 1;
	u_int8_t padding : 4;

	u_int32_t key;               // if k_flag set
	u_int32_t sequence_number;   // if s_flag set

} ipv6_gre_option_context_t;

/**
 * @brief Define the IPv6 option context for MIME option
 *
 */

typedef struct __attribute__((packed)) ipv6_mime_option_context
{
	u_int8_t context_length;
	u_int8_t option_length;

	u_int8_t next_header;

	u_int8_t s_bit : 1;
	u_int8_t res_bits : 7;
	u_int32_t orig_dest;
	u_int32_t orig_src;         // if s_bit set

} ipv6_mime_option_context_t;

/**
 * @brief Define the IPv6 option context for AH option
 *
 */

typedef struct __attribute__((packed)) ipv6_ah_option_context
{
	u_int8_t context_length;
	u_int8_t option_length;

	u_int8_t next_header;
	u_int8_t length;
	u_int32_t spi;
	u_int32_t sequence_number;
	u_int32_t icv[1];
} ipv6_ah_option_context_t;

/**
 * @brief Define the common IP header context to IPv4 and IPv6.
 *
 */

typedef struct __attribute__((packed)) ipvx_context
{
	u_int8_t context_length;

	u_int8_t version : 4;
	u_int8_t unused : 1;
	u_int8_t inner : 3;

	u_int8_t dscp : 6;
	u_int8_t ip_ecn_flags : 2;

	u_int8_t next_header;

	u_int8_t tos_tc;
	u_int8_t ttl_hopl;

	u_int8_t ip_id_behavior;
	u_int16_t ip_id_offset;

} ipvx_context_t;

/**
 * @brief Define the IPv4 header context.
 *
 */

typedef struct __attribute__((packed)) ipv4_context
{
	u_int8_t context_length;

	u_int8_t version : 4;
	u_int8_t df : 1;
	u_int8_t inner : 3;

	u_int8_t dscp : 6;
	u_int8_t ip_ecn_flags : 2;

	u_int8_t protocol;

	u_int8_t tos_tc;
	u_int8_t ttl_hopl;

	u_int8_t ip_id_behavior;
	u_int16_t ip_id_offset;
	WB_t last_ip_id;

	u_int32_t src_addr;
	u_int32_t dst_addr;

} ipv4_context_t;

/**
 * @brief Define the IPv6 header context.
 *
 */

typedef struct __attribute__((packed)) ipv6_context
{
	u_int8_t context_length;

	u_int8_t version : 4;
	u_int8_t unused : 1;
	u_int8_t inner : 3;

	u_int8_t dscp : 6;
	u_int8_t ip_ecn_flags : 2;

	u_int8_t next_header;

#if __BYTE_ORDER == __LITTLE_ENDIAN
	u_int8_t tos_tc1 : 4;
	u_int8_t tos_tc2 : 4;
#elif __BYTE_ORDER == __BIG_ENDIAN
	u_int8_t tos_tc;
#endif
	u_int8_t ttl_hopl;

	u_int8_t ip_id_behavior;
	u_int16_t ip_id_offset;

	u_int8_t flow_label1 : 4;
	u_int16_t flow_label2;

	u_int32_t src_addr[4];
	u_int32_t dest_addr[4];

} ipv6_context_t;

/**
 * @brief Define the UDP context
 *
 */

typedef struct __attribute__((packed)) udp_context
{
	u_int8_t protocol;

	u_int16_t src_port;
	u_int16_t dst_port;

	u_int8_t checksum_used;

} udp_context_t;

/**
  * @brief Define the ESP context
  *
  */

typedef struct __attribute__((packed)) esp_context
{
	u_int8_t protocol;

	u_int32_t spi;

	u_int32_t sequence_number;

} esp_context_t;

/**
  * @brief Define the UDP-Lite context
  *
  */

typedef struct __attribute__((packed)) udp_lite_context
{
	u_int8_t protocol;

	u_int16_t src_port;
	u_int16_t dst_port;

	u_int8_t checksum_used;

	u_int8_t coverage_behavior;

	u_int16_t checksum_coverage;
} udp_lite_context_t;

/**
  * @brief Define the RTP context
  *
  */

typedef struct __attribute__((packed)) rtp_context
{
	u_int8_t protocol;

	u_int32_t ssrc;

	// The following are only used in RTP-based profiles
	u_int32_t ts_stride;
	u_int32_t time_stride;
	u_int32_t ts_scaled;
	u_int32_t ts_offset;

	u_int8_t marker : 1;
	u_int8_t payload_type : 7;
	u_int8_t pad_bit : 1;
	u_int8_t extension : 1;
	u_int8_t reserved : 6;

	u_int16_t sequence_number;
	u_int32_t timestamp;

	u_int16_t csrv_received;
	u_int32_t csrc_list[16];

//u_int8_t new_reorder_ratio;
//u_int32_t new_ts_stride;
//u_int32_t new_time_stride;
//u_int8_t new_ip_id_behavior;  // For all IP header !

//	u_int8_t	ts_decomp_initialized;
//	struct ts_sc_decomp ts_decomp;
} rtp_context_t;

/**
 * @brief Define union of IP contexts pointers.
 *
 */

typedef union
{
	u_int8_t *uint8;
	ipvx_context_t *vx;
	ipv4_context_t *v4;
	ipv6_context_t *v6;
	ipv6_option_context_t *v6_option;
	ipv6_gre_option_context_t *v6_gre_option;
	ipv6_mime_option_context_t *v6_mime_option;
	ipv6_ah_option_context_t *v6_ah_option;
	udp_context_t *udp;
	esp_context_t *esp;
	udp_lite_context_t *udp_lite;
	rtp_context_t *rtp;
} ip_context_ptr_t;

#define MAX_IP_CONTEXT_SIZE  (((sizeof(ipv4_context_t) + sizeof(ipv6_context_t) + \
                                sizeof(ipv6_option_context_t)) * 2) + sizeof(rtp_context_t))

/**
 * @brief Define the RoHCv2 part of the decompression profile context.
 *
 * This object must be used with the generic part of the decompression
 * context d_generic_context.
 *
 * @see d_generic_context
 */
struct d_v2_context
{

	// The Master Sequence Number
	u_int16_t msn;

	u_int8_t reorder_ratio;

	u_int8_t ip_context[MAX_IP_CONTEXT_SIZE];
};

typedef struct
{
	struct d_profile profile;
	u_int8_t protocol;
	u_int8_t static_part_size;
} d_v2_profile_t;

/*
 * Public function prototypes.
 */

void d_v2_destroy(void *context);

int d_v2_decode_ir(struct rohc_decomp *decomp,
                   struct d_context *context,
                   const unsigned char *rohc_packet,
                   const unsigned int rohc_length,
                   int large_cid_len,
                   int is_addcid_used,
                   unsigned char *dest);

unsigned int v2_detect_ir_size(struct d_context *context,
                               unsigned char *packet,
                               unsigned int plen,
                               unsigned int large_cid_len);

int v2_get_static_size(void);

#endif

