/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/**
 * @file rfc5225_decoding.c
 * @brief Library of decoding methods from RFC4997, RFC4996 and RFC5225
 * @author FWX <rohc_team@dialine.fr>
 */

#include "d_generic.h"
#include "d_rtp.h"
#include "config.h" /* for RTP_BIT_TYPE definition */

#if SUPPORT_ROHC_V2

#include "rohc_traces.h"
#include "rohc_time.h"
#include "rohc_debug.h"
#include "rohc_packets.h"
#include "rohc_bit_ops.h"
#include "wlsb.h"
#include "sdvl.h"
#include "crc.h"

#include <assert.h>

#include "trace.h" //FWX2
//#include "tcp.h" //FWX2
#include "rohc_v2.h"
//#include "d_tcp.h" // FWX2
#include "d_rohc_v2.h"
#include "rfc5225_decoding.h"
#include "../comp/rfc5225_encoding.h"


#ifdef ALREADY_DEFINED_IN_RFC4996_C

/**
 * @brief Table of the mask for lsb()
 */

unsigned int lsb_xor_masks[] =
{
	0xFFFFFFFF,
	0xFFFFFFFE, 0xFFFFFFFC, 0xFFFFFFF8, 0xFFFFFFF0,
	0xFFFFFFE0, 0xFFFFFFC0, 0xFFFFFF80, 0xFFFFFF00,
	0xFFFFFE00, 0xFFFFFC00, 0xFFFFF800, 0xFFFFF000,
	0xFFFFE000, 0xFFFFC000, 0xFFFF8000, 0xFFFF0000,
	0xFFFE0000, 0xFFFC0000, 0xFFF80000, 0xFFF00000
};

/**
 * @brief Decompress the lower bits from the given value and the context value.
 *
 * See RFC4997 page 27
 *
 * @param num_lsbs_param   The number of bits
 * @param offset_param     The offset
 * @param context_value    The value of the context
 * @param value            The compressed value
 * @return                 The uncompressed value
 */

u_int32_t d_lsb( int num_lsbs_param, int offset_param, unsigned int context_value,
                 unsigned int value )
{
	assert( num_lsbs_param < 20 );
	rohc_debugf(3, "d_lsb() num_lsbs_param %d context_value %Xh mask %Xh value %Xh -> %Xh\n",
	            num_lsbs_param,context_value,lsb_xor_masks[num_lsbs_param],value,
	            (context_value & lsb_xor_masks[num_lsbs_param]) | value);
	return ( context_value & lsb_xor_masks[num_lsbs_param] ) | value;
}


#endif

/**
 * @brief Decompress the 8 bits given value, according to the indicator
 *
 * See RFC4996 page 46
 *
 * @param pmptr            Pointer to the compressed value
 * @param context_value    The value of the context
 * @param indicator        Indicator of compression
 * @return                 The uncompressed value
 */
/*
u_int8_t d_static_or_irreg8( multi_ptr_t * pmptr , u_int8_t context_value , int indicator )
{
   if( indicator == 0 )
   {
      return context_value;
   }
   else
   {
      return *(pmptr->uint8)++;
   }
}
*/

#ifdef ALREADY_DEFINED_IN_RFC4996_C

/**
 * @brief Decompress the 16 bits given value, according to the indicator
 *
 * @param pmptr            Pointer to the compressed value
 * @param context_value    The value of the context
 * @param indicator        Indicator of compression
 * @return                 The uncompressed value
 */

u_int16_t d_static_or_irreg16( multi_ptr_t *pmptr, u_int16_t context_value, int indicator )
{
	if(indicator == 0)
	{
		return context_value;
	}
	else
	{
		return READ16_FROM_PMPTR(pmptr);
	}
}


#endif

#ifdef ALREADY_DEFINED_IN_RFC4996_C

/**
 * @brief Table of the size of the variable_length_32 encode value.
 */

unsigned int variable_length_32_size[4] =
{
	0,1,2,4
};

/**
 * @brief Decode the 32 bits value, according to the indicator
 *
 * See RFC4996 page 46
 *
 * @param pmptr            Pointer to the compressed value
 * @param indicator        Indicator of compression
 * @return                 The uncompressed value
 */

u_int32_t variable_length_32_dec( multi_ptr_t *pmptr, int indicator )
{
	multi_ptr_t mptr;
	u_int32_t value;

	switch(indicator)
	{
		case 0:
			value = 0;
			break;
		case 1:
			mptr.uint32 = (u_int32_t UNALIGNED *) &value;
			WRITE16_TO_MPTR(mptr,0);
			*(mptr.uint8++) = 0;
			*mptr.uint8 = *(pmptr->uint8)++;
			break;
		case 2:
			mptr.uint32 = (u_int32_t UNALIGNED *) &value;
			WRITE16_TO_MPTR(mptr,0);
			WRITE16_TO_MPTR(mptr,READ16_FROM_PMPTR(pmptr));
			break;
		case 3:
			value = READ32_FROM_PMPTR(pmptr);
			break;
		default:
			value = 0;
			break;
	}

	rohc_debugf(3, "indicator %d return value %d (%Xh)\n",indicator,value,value);
	fflush(stdout);

	return value;
}


#endif

#ifdef ALREADY_DEFINED_IN_RFC4996_C

/**
 * @brief Decode the 32 bits value, according to the indicator
 *
 * See RFC4996 page 47
 *
 * @param pmptr            Pointer to the compressed value
 * @param flag             Flag of compression
 * @param context_value    The context value
 * @return                 The uncompressed value
 */

u_int32_t d_optional32( multi_ptr_t *pmptr, int flag, u_int32_t context_value )
{
	if(flag == 1)
	{
		return READ32_FROM_PMPTR(pmptr);
	}
	return context_value;
}


#endif

#ifdef ALREADY_DEFINED_IN_RFC4996_C

/**
 * @brief Decompress the 7 or 31 bits of a 32 bits value
 *
 * See RFC4996 page 47
 *
 * @param pmptr            Pointer to the compressed value
 * @return                 The uncompressed value
 */

u_int32_t d_lsb_7_31( multi_ptr_t *pmptr )
{
	if( (*pmptr->uint8) & 0x80)
	{
		return ntohl( READ32_FROM_PMPTR(pmptr) ) & 0x7FFFFFFF;
	}
	else
	{
		return *(pmptr->uint8++);
	}
}


#endif

#ifndef USE_ROHCV2_MACROS

/**
 * @brief Calculate the value from the scaling factor, scaled value and residue
 *
 * See RFC4996 page 49
 *
 * @param scaling_factor   The scaling factor
 * @param scaled_value     The scaled value
 * @param residue_field    The residue value
 * @return                 The unscaled value
 */

u_int32_t d_field_scaling( u_int32_t scaling_factor, u_int32_t scaled_value,
                           u_int32_t residue_field )
{
	u_int32_t unscaled_value;

	if(scaling_factor == 0)
	{
		unscaled_value = residue_field;
	}
	else
	{
		unscaled_value = ( scaled_value * scaling_factor ) + residue_field;
	}
	return unscaled_value;
}


#endif

#ifdef TCP_H

/**
 * @brief Calculate the rsf flags from the rsf index
 *
 * See RFC4996 page 71
 *
 * @param rsf_index    The rsf index
 * @return             The rsf flags
 */
unsigned int rsf_index_dec( unsigned int rsf_index )
{
	switch(rsf_index)
	{
		case 0:
			return 0;
		case 1:
			return RSF_RST_ONLY;
		case 2:
			return RSF_SYN_ONLY;
		case 3:
			return RSF_FIN_ONLY;
		default:
			return 0;
	}
}


#endif

#ifdef ALREADY_DEFINED_IN_RFC4996_C

/**
 * @brief Decompress the lower bits of IP-ID
 *
 * See RFC4996 page 75
 *
 * @param behavior       The IP-ID behavior
 * @param k              The num_lsbs_param parameter for d_lsb()
 * @param p              The offset parameter for d_lsb()
 * @param context_ip_id  The context IP-ID value
 * @param value          The value to decompress
 * @param msn            The Master Sequence Number
 * @return               The IP-ID
 */

u_int16_t d_ip_id_lsb( int behavior, unsigned int k, unsigned int p, WB_t context_ip_id,
                       u_int16_t value,
                       u_int16_t msn )
{
	u_int16_t ip_id_offset;
	WB_t ip_id;


	rohc_debugf(3, "behavior %d k=%d p=%d context_ip_id %4.4Xh value %4.4Xh msn %4.4Xh\n",behavior,k,
	            p,context_ip_id.uint16,value,
	            msn);

	switch(behavior)
	{
		case IP_ID_BEHAVIOR_SEQUENTIAL:
			ip_id.uint16 = context_ip_id.uint16 + 1;
			ip_id_offset = ip_id.uint16 - msn;
			ip_id_offset = c_lsb( k, p, context_ip_id.uint16 - msn, ip_id_offset );
			rohc_debugf(3, "new ip_id %4.4Xh ip_id_offset %Xh value %Xh\n",ip_id.uint16,ip_id_offset,
			            value);
// ??? why not		                         assert( ip_id_offset == value );
			return ip_id.uint16;
		case IP_ID_BEHAVIOR_SEQUENTIAL_SWAPPED:
			ip_id.uint8[0] = context_ip_id.uint8[1];
			ip_id.uint8[1] = context_ip_id.uint8[0];
			++ip_id.uint16;
			ip_id_offset = ip_id.uint16 - msn;
			ip_id_offset = c_lsb( k, p, ip_id.uint16 - 1 - msn, ip_id_offset );
			rohc_debugf(3, "new ip_id %4.4Xh ip_id_offset %Xh value %Xh\n",ip_id.uint16,ip_id_offset,
			            value);
			assert( ip_id_offset == value );
			return ip_id.uint16;
	}

	return 0;
}


#endif

#ifdef ALREADY_DEFINED_IN_RFC4996_C

/**
 * @brief Decompress the IP-ID
 *
 * See RFC4996 page 76
 *
 * @param pmptr          Pointer to the compressed value
 * @param behavior       The IP-ID behavior
 * @param indicator      The compression indicator
 * @param context_ip_id  The context IP-ID value
 * @param ip_id          Pointer to the uncompressed IP-ID
 * @param msn            The Master Sequence Number
 * @return               Nothing
 */

void d_optional_ip_id_lsb( multi_ptr_t *pmptr, int behavior, int indicator, WB_t context_ip_id,
                           u_int16_t *ip_id,
                           u_int16_t msn )
{
	rohc_debugf(3, "behavior %d indicator %d context_ip_id %Xh msn %Xh\n",behavior,indicator,
	            context_ip_id.uint16,
	            msn);

	switch(behavior)
	{
		case IP_ID_BEHAVIOR_SEQUENTIAL:
			if(indicator == 0)
			{
				*ip_id =
				   ( context_ip_id.uint16 &
				     0xFF00 ) | d_ip_id_lsb(behavior,8,3,context_ip_id,*(pmptr->uint8)++,
				                            msn);
				rohc_debugf(3, "read ip_id %Xh -> %Xh\n",*(pmptr->uint8 - 1),*ip_id);
			}
			else
			{
				*ip_id = ntohs( READ16_FROM_PMPTR(pmptr) );
				rohc_debugf(3, "read ip_id %Xh\n",*ip_id);
			}
			break;
		case IP_ID_BEHAVIOR_SEQUENTIAL_SWAPPED:
		{
			WB_t swapped_context_ip_id;
			swapped_context_ip_id.uint8[0] = context_ip_id.uint8[1];
			swapped_context_ip_id.uint8[1] = context_ip_id.uint8[0];
			if(indicator == 0)
			{
				*ip_id =
				   ( swapped_context_ip_id.uint16 &
				     0xFF00 ) | d_ip_id_lsb(behavior,8,3,context_ip_id,*(pmptr->uint8)++,
				                            msn);
				rohc_debugf(3, "read ip_id %Xh -> %Xh\n",*(pmptr->uint8 - 1),*ip_id);
			}
			else
			{
				*ip_id = ntohs( READ16_FROM_PMPTR(pmptr) );
				rohc_debugf(3, "read ip_id %Xh\n",*ip_id);
			}
		}
		break;
		case IP_ID_BEHAVIOR_RANDOM:
			break;
		case IP_ID_BEHAVIOR_ZERO:
			*ip_id = 0;
		default:
			break;
	}
}


#endif

/**
 * @brief Decompress the IP-ID
 *
 * See RFC5225 page 70
 *
 * @param pmptr          Pointer to the compressed value
 * @param behavior       The IP-ID behavior
 * @param indicator      The compression indicator
 * @param context_ip_id  The context IP-ID value
 * @param ip_id          Pointer to the uncompressed IP-ID
 * @param msn            The Master Sequence Number
 * @return               Nothing
 */

void d_ip_id_sequential_variable( multi_ptr_t *pmptr, int behavior, int indicator,
                                  WB_t context_ip_id, u_int16_t *ip_id,
                                  u_int16_t msn )
{
	switch(behavior)
	{
		case IP_ID_BEHAVIOR_SEQUENTIAL:
			if(indicator == 0)
			{
				*ip_id =
				   ( context_ip_id.uint16 &
				     0xFF00 ) | d_ip_id_lsb(behavior,8,3,context_ip_id,*(pmptr->uint8)++,
				                            msn);
				rohc_debugf(3, "read ip_id %Xh -> %Xh\n",*(pmptr->uint8 - 1),*ip_id);
			}
			else
			{
				*ip_id = ntohs( READ16_FROM_PMPTR(pmptr) );
				rohc_debugf(3, "read ip_id %Xh\n",*ip_id);
			}
			break;
		case IP_ID_BEHAVIOR_SEQUENTIAL_SWAPPED:
		{
			WB_t swapped_context_ip_id;
			swapped_context_ip_id.uint8[0] = context_ip_id.uint8[1];
			swapped_context_ip_id.uint8[1] = context_ip_id.uint8[0];
			if(indicator == 0)
			{
				*ip_id =
				   ( swapped_context_ip_id.uint16 &
				     0xFF00 ) | d_ip_id_lsb(behavior,8,3,context_ip_id,*(pmptr->uint8)++,
				                            msn);
				rohc_debugf(3, "read ip_id %Xh -> %Xh\n",*(pmptr->uint8 - 1),*ip_id);
			}
			else
			{
				*ip_id = ntohs( READ16_FROM_PMPTR(pmptr) );
				rohc_debugf(3, "read ip_id %Xh\n",*ip_id);
			}
		}
		break;
		case IP_ID_BEHAVIOR_RANDOM:
			break;
		case IP_ID_BEHAVIOR_ZERO:
			*ip_id = 0;
		default:
			break;
	}
}


/**
 * @brief Calculate the IP-ID compressed size
 *
 * See RFC5225 page 70
 *
 * @param behavior       The IP-ID behavior
 * @param indicator      The compression indicator
 * @return               The size
 */

int d_ip_id_sequential_variable_size( int behavior, int indicator )
{
	int size;

	switch(behavior)
	{
		case IP_ID_BEHAVIOR_SEQUENTIAL:
		case IP_ID_BEHAVIOR_SEQUENTIAL_SWAPPED:
			if(indicator == 0)
			{
				size = 1;
				break;
			}
			size = 2;
			break;
		case IP_ID_BEHAVIOR_RANDOM:
		case IP_ID_BEHAVIOR_ZERO:
		default:
			size = 0;
			break;
	}
	rohc_debugf(3, "return size %d\n",size);
	return size;
}


#ifdef ALREADY_DEFINED_IN_RFC4996_C

/**
 * @brief Decode the DSCP field
 *
 * @param pmptr          Pointer to the compressed value
 * @param context_value  The context DSCP value
 * @param indicator      Indicator of the compression
 * @return               The DSCP decoded
 */

u_int8_t dscp_decode( multi_ptr_t *pmptr, u_int8_t context_value, int indicator )
{
	if(indicator == 0)
	{
		return context_value;
	}
	return (*(pmptr->uint8)++) & 0x3F;
}


#endif

/**
 * @brief Decode the Master Sequence Number
 *
 * @param v2_context     The specific RoHCv2 context
 * @param k              Number of bits in compressed value
 * @param value          The compressed value
 * @return               The MSN decoded
 */

u_int16_t d_msn_lsb( struct d_v2_context *v2_context, int k, unsigned int value )
{
	u_int16_t msn;

	switch(v2_context->reorder_ratio)
	{
		case REORDERING_NONE:
			msn = d_lsb(k,1,v2_context->msn + 1,value);
			break;
		case REORDERING_QUATER:
			msn = d_lsb(k,((1 << k) / 4) - 1,v2_context->msn + 1,value);
			break;
		case REORDERING_HALF:
			msn = d_lsb(k,((1 << k) / 2) - 1,v2_context->msn + 1,value);
			break;
		case REORDERING_THREEQUARTERS:
			msn = d_lsb(k,(((1 << k) * 3) / 4) - 1,v2_context->msn + 1,value);
			break;
	}

	rohc_debugf(3, "msn %4.4Xh\n",msn);
	return msn;
}


void d_optional_32( multi_ptr_t *pmptr, u_int32_t *value, int flag )
{
	if(flag == 1)
	{
		*value = READ32_FROM_PMPTR(pmptr);
	}
	else
	{
		*value = 0;
	}
}


// Send the entire value or keep previous value
// See RFC5225 page 48

void d_sdvl_or_static( multi_ptr_t *pmptr, int flag, u_int32_t context_value, u_int32_t *value )
{
	u_int32_t field;

	rohc_debugf(3, "flag %d context_value %d %Xh : %2.2Xh %2.2Xh %2.2Xh %2.2Xh\n",flag,context_value,
	            context_value,*pmptr->uint8,*(pmptr->uint8 + 1),*(pmptr->uint8 + 2),
	            *(pmptr->uint8 + 3));

	do
	{
		if(flag == 0)
		{
			*value = context_value;
			break;
		}
		if( ( (*pmptr->uint8) & 0x80 ) == 0)
		{
			// Discriminator =:= '0'
			*value = *(pmptr->uint8++);
			break;
		}
		if( ( (*pmptr->uint8) & 0x40 ) == 0)
		{
			// Discriminator =:= '10'
			field = *(pmptr->uint8++) & 0x3F;
			field <<= 8;
			field |= *(pmptr->uint8++);
			*value = field;
			break;
		}
		if( ( (*pmptr->uint8) & 0x20 ) == 0)
		{
			// Discriminator =:= '110'
			field = *(pmptr->uint8++) & 0x1F;
			field <<= 8;
			field |= *(pmptr->uint8++);
			field <<= 8;
			field |= *(pmptr->uint8++);
			*value = field;
			break;
		}
		if( ( (*pmptr->uint8) & 0x10 ) == 0)
		{
			// Discriminator =:= '1110'
			field = *(pmptr->uint8++) & 0x0F;
			field <<= 8;
			field |= *(pmptr->uint8++);
			field <<= 8;
			field |= *(pmptr->uint8++);
			field <<= 8;
			field |= *(pmptr->uint8++);
			*value = field;
			break;
		}
		// Discriminator =:= '11111111'
		rohc_debugf(3, "*pmptr %2.2Xh %2.2Xh %2.2Xh %2.2Xh %2.2Xh\n",*(pmptr->uint8),
		            *(pmptr->uint8 + 1),*(pmptr->uint8 + 2),*(pmptr->uint8 + 3),*(pmptr->uint8 + 4));
		assert( *pmptr->uint8 == 0xFF );
		++pmptr->uint8;
		field = *(pmptr->uint8++);
		field <<= 8;
		field |= *(pmptr->uint8++);
		field <<= 8;
		field |= *(pmptr->uint8++);
		field <<= 8;
		field |= *(pmptr->uint8++);
		*value = field;

	}
	while(0);
}


// Send the entire value or revert to default value
// See RFC5225 page 49

u_int32_t d_sdvl_or_default( multi_ptr_t *pmptr, int flag, u_int32_t default_value )
{
	u_int32_t value;

	rohc_debugf(3, "flag %d default value %d %Xh\n",flag,default_value,default_value);

	do
	{
		if(flag == 0)
		{
			// Default value
			value = default_value;
			break;
		}
		if( ( (*pmptr->uint8) & 0x80 ) == 0)
		{
			// Discriminator =:= '0'
			value = *(pmptr->uint8++);
			break;
		}
		if( ( (*pmptr->uint8) & 0x40 ) == 0)
		{
			// Discriminator =:= '10'
			value = *(pmptr->uint8++) & 0x3F;
			value <<= 8;
			value |= *(pmptr->uint8++);
			break;
		}
		if( ( (*pmptr->uint8) & 0x20 ) == 0)
		{
			// Discriminator =:= '110'
			value = *(pmptr->uint8++) & 0x1F;
			value <<= 8;
			value |= *(pmptr->uint8++);
			value <<= 8;
			value |= *(pmptr->uint8++);
			break;
		}
		if( ( (*pmptr->uint8) & 0x10 ) == 0)
		{
			// Discriminator =:= '1110'
			value = *(pmptr->uint8++) & 0x0F;
			value <<= 8;
			value |= *(pmptr->uint8++);
			value <<= 8;
			value |= *(pmptr->uint8++);
			value <<= 8;
			value |= *(pmptr->uint8++);
			break;
		}
		// Discriminator =:= '11111111'
		assert( *pmptr->uint8 == 0xFF );
		++pmptr->uint8;
		value = *(pmptr->uint8++);
		value <<= 8;
		value |= *(pmptr->uint8++);
		value <<= 8;
		value |= *(pmptr->uint8++);
		value <<= 8;
		value |= *(pmptr->uint8++);

	}
	while(0);

	return value;
}


int d_sdvl_or_default_size( multi_ptr_t *pmptr, int flag )
{
	int size;

	rohc_debugf(3, "flag %d\n",flag);

	do
	{
		if(flag == 0)
		{
			// Default value
			size = 0;
			break;
		}
		if( ( (*pmptr->uint8) & 0x80 ) == 0)
		{
			// Discriminator =:= '0'
			size = 1;
			break;
		}
		if( ( (*pmptr->uint8) & 0x40 ) == 0)
		{
			// Discriminator =:= '10'
			size = 2;
			break;
		}
		if( ( (*pmptr->uint8) & 0x20 ) == 0)
		{
			// Discriminator =:= '110'
			size = 3;
			break;
		}
		if( ( (*pmptr->uint8) & 0x10 ) == 0)
		{
			// Discriminator =:= '1110'
			size = 4;
			break;
		}
		// Discriminator =:= '11111111'
		assert( *pmptr->uint8 == 0xFF );
		size = 5;

	}
	while(0);

	pmptr->uint8 += size;

	rohc_debugf(3, "return size %d\n",size);
	return size;
}


// Self-describing variable length encoding with reordering offset
// See RFC5225 page 72

u_int32_t d_sdvl_sn_lsb( multi_ptr_t *pmptr )
{
	u_int32_t value;
	do
	{
		if( ( (*pmptr->uint8) & 0x80 ) == 0)
		{
			// Discriminator '0'
			value = *(pmptr->uint8++);
			break;
		}
		if( ( (*pmptr->uint8) & 0x40 ) == 0)
		{
			// Discriminator '10'
			value = *(pmptr->uint8++) & 0x3F;
			value <<= 8;
			value |= *(pmptr->uint8++);
			break;
		}
		if( ( (*pmptr->uint8) & 0x20 ) == 0)
		{
			// Discriminator '110'
			value = *(pmptr->uint8++) & 0x1F;
			value <<= 8;
			value |= *(pmptr->uint8++);
			value <<= 8;
			value |= *(pmptr->uint8++);
			break;
		}
		if( ( (*pmptr->uint8) & 0x10 ) == 0)
		{
			// Discriminator '1110'
			value = *(pmptr->uint8++) & 0x0F;
			value <<= 8;
			value |= *(pmptr->uint8++);
			value <<= 8;
			value |= *(pmptr->uint8++);
			value <<= 8;
			value |= *(pmptr->uint8++);
			break;
		}
		if( (*pmptr->uint8) == 0xFF)
		{
			// Discriminator '11111111'
			++pmptr->uint8;
			value = ntohl( READ32_FROM_PMPTR(pmptr) );
			break;
		}
		assert( 1 == 0 );
	}
	while(0);

	rohc_debugf(3, "return %Xh\n",value);

	return value;
}


// Return the size of the compressed value

int d_sdvl_size( u_int8_t discriminator )
{
	rohc_debugf(3, "discriminator %Xh\n",discriminator);

	if( ( discriminator & 0x80 ) == 0)
	{
		// Discriminator =:= '0'
		rohc_debugf(3, "return size 1\n");
		return 1;
	}
	if( ( discriminator & 0x40 ) == 0)
	{
		// Discriminator =:= '10'
		rohc_debugf(3, "return size 2\n");
		return 2;
	}
	if( ( discriminator & 0x20 ) == 0)
	{
		// Discriminator =:= '110'
		rohc_debugf(3, "return size 3\n");
		return 3;
	}
	if( ( discriminator & 0x10 ) == 0)
	{
		// Discriminator =:= '1110'
		rohc_debugf(3, "return size 4\n");
		return 4;
	}
	if(discriminator == 0xFF)
	{
		// Discriminator =:= '11111111'
		rohc_debugf(3, "return size 5\n");
		return 5;
	}
	// Error
	return -1;
}


// See RFC5225 section 6.6.9 page 30 and appendix B.3
u_int32_t d_timer_based_lsb( multi_ptr_t *pmptr, u_int16_t time_stride_param, int k, int p )
{
	u_int32_t timestamp;

	rohc_debugf(3, "mptr %p time_stride_param %Xh k %d p %d\n",pmptr->uint8,time_stride_param,k,p);

	// tsc_ref_advanced = tsc_ref + (a_n - a_ref) / time_stride

	return timestamp;
}


// See RFC5225 page 71
u_int32_t d_scaled_ts_lsb( multi_ptr_t *pmptr, u_int16_t time_stride_value, int k )
{
	u_int32_t timestamp;

	rohc_debugf(3, "mptr %p time_stride_value %d k %d\n",pmptr->uint8,time_stride_value,k);

	if(time_stride_value != 0)
	{
		timestamp = d_timer_based_lsb(pmptr,time_stride_value,k,((1 << k) / 2) - 1);
	}
	else
	{
		// u_int32_t d_lsb( int num_lsbs_param , int offset_param , unsigned int context_value , unsigned int value )

// A REVOIR	timestamp = d_lsb(k,((1<<k)/4)-1,context_value,value);
		timestamp = 0;
	}

	return timestamp;
}


// Self-describing variable length encoding
// RFC5225 page 73
u_int32_t d_sdvl_lsb( multi_ptr_t *pmptr, int field_width, u_int32_t context_value )
{
	u_int32_t field;
	u_int32_t value;

	rohc_debugf(3, "mptr %p field_width %d context_value %Xh\n",pmptr->uint8,field_width,
	            context_value);
	rohc_debugf(3, "mptr : %2.2Xh %2.2Xh %2.2Xh %2.2Xh %2.2Xh\n",*(pmptr->uint8),*(pmptr->uint8 + 1),
	            *(pmptr->uint8 + 2),*(pmptr->uint8 + 3),
	            *(pmptr->uint8 + 4));

	assert( field_width == 32 );

	do
	{

		if( ( (*pmptr->uint8) & 0x80 ) == 0)
		{
			// Discriminator '0'
			field = *(pmptr->uint8++);
			value = ( context_value & 0xFFFFFF80 ) | field;
			break;
		}
		if( ( (*pmptr->uint8) & 0x40 ) == 0)
		{
			// Discriminator '10'
			field = *(pmptr->uint8++) & 0x3F;
			field <<= 8;
			field |= *(pmptr->uint8++);
			value = ( context_value & 0xFFFFC000 ) | field;
			break;
		}
		if( ( (*pmptr->uint8) & 0x20 ) == 0)
		{
			// Discriminator '110'
			field = *(pmptr->uint8++) & 0x1F;
			field <<= 8;
			field |= *(pmptr->uint8++);
			field <<= 8;
			field |= *(pmptr->uint8++);
			value = ( context_value & 0xFFE00000 ) | field;
			break;
		}
		if( ( (*pmptr->uint8) & 0x10 ) == 0)
		{
			// Discriminator '1110'
			field = *(pmptr->uint8++) & 0x0F;
			field <<= 8;
			field |= *(pmptr->uint8++);
			field <<= 8;
			field |= *(pmptr->uint8++);
			field <<= 8;
			field |= *(pmptr->uint8++);
			value = ( context_value & 0xF0000000 ) | field;
			break;
		}
		assert( (*pmptr->uint8) == 0xFF );
		++pmptr->uint8;
		value = ntohl( READ32_FROM_PMPTR(pmptr) );

	}
	while(0);

	rohc_debugf(3, "return %Xh\n",value);

	return value;
}


// RFC5225 page 73
u_int32_t d_sdvl_scaled_ts_lsb( multi_ptr_t *pmptr, u_int32_t time_stride, u_int32_t value )
{
//	u_int32_t value;

	rohc_debugf(3, "mptr %p time_stride %Xh\n",pmptr->uint8,time_stride);

	do
	{
		if( ( *(pmptr->uint8) & 0x80 ) == 0)
		{
			// Discriminator '0'
			value = *(pmptr->uint8++);
			break;
		}
		if( ( *(pmptr->uint8) & 0x40 ) == 0)
		{
			// Discriminator '10'
			value = *(pmptr->uint8++) & 0x3F;
			value <<= 8;
			value |= *(pmptr->uint8++);
			break;
		}
		if( ( *(pmptr->uint8) & 0x20 ) == 0)
		{
			// Discriminator '110'
			value = *(pmptr->uint8++) & 0x1F;
			value <<= 8;
			value |= *(pmptr->uint8++);
			value <<= 8;
			value |= *(pmptr->uint8++);
			break;
		}
		if( ( *(pmptr->uint8) & 0x10 ) == 0)
		{
			// Discriminator '1110'
			value = *(pmptr->uint8++) & 0x0F;
			value <<= 8;
			value |= *(pmptr->uint8++);
			value <<= 8;
			value |= *(pmptr->uint8++);
			value <<= 8;
			value |= *(pmptr->uint8++);
			break;
		}
		// Discriminator '11111111'
		assert( (*pmptr->uint8) == 0xFF );
		++pmptr->uint8;
		value = *(pmptr->uint8++);
		value <<= 8;
		value |= *(pmptr->uint8++);
		value <<= 8;
		value |= *(pmptr->uint8++);
		value <<= 8;
		value |= *(pmptr->uint8++);

	}
	while(0);

//	return d_scaled_ts_lsb(pmptr,time_stride);

	rohc_debugf(3, "return %d %Xh\n",value,value);

	return value;
}


// RFC5225 page 74
void d_variable_scaled_timestamp( multi_ptr_t *pmptr, rtp_context_t *rtp_context, int tss_flag,
                                  int tsc_flag,
                                  u_int32_t *scaled_value )
{
	rohc_debugf(3, "mptr %p tss_flag %d tsc_flag %d\n",pmptr->uint8,tss_flag,tsc_flag);
	rohc_debugf(3, "rtp_context %p ts_stride %d time_stride %d\n",rtp_context,rtp_context->ts_stride,
	            rtp_context->time_stride);

	if(tss_flag == 0 && tsc_flag != 0 && rtp_context->ts_stride != 0)
	{
		u_int32_t value; // ???
		// Present
		*scaled_value = d_sdvl_scaled_ts_lsb(pmptr,rtp_context->time_stride,value);
		rohc_debugf(3, "scaled_timestamp %d %Xh\n",*scaled_value,*scaled_value);
	}
	else
	{
		// Not present
		assert( tsc_flag == 0 );
		rohc_debugf(3, "Not present\n");
	}
}


// RFC5225 page 74
int d_variable_scaled_timestamp_size( multi_ptr_t mptr, rtp_context_t *rtp_context, int tss_flag,
                                      int tsc_flag )
{
	int size;

	rohc_debugf(3, "tss_flag %d tsc_flag %d\n",tss_flag,tsc_flag);
	rohc_debugf(3, "rtp_context %p ts_stride %d time_stride %d\n",rtp_context,rtp_context->ts_stride,
	            rtp_context->time_stride);

	if(tss_flag == 0 && tsc_flag != 0 && rtp_context->ts_stride != 0)
	{
		// Present
		size = d_sdvl_size(*mptr.uint8);
	}
	else
	{
		// Not present
		assert( tsc_flag == 0 );
		size = 0;
		rohc_debugf(3, "Not present\n");
	}

	return size;
}


// RFC5225 page 74
void d_variable_unscaled_timestamp( multi_ptr_t *pmptr, int tss_flag, int tsc_flag,
                                    u_int32_t context_timestamp,
                                    u_int32_t *timestamp )
{
	rohc_debugf(3, "mptr %p tss_flag %d tsc_flag %d\n",pmptr->uint8,tss_flag,tsc_flag);

	if(tsc_flag == 0)
	{
		// Present
		rohc_debugf(3, "mptr %p\n",pmptr->uint8);
		*timestamp = d_sdvl_lsb(pmptr,32,context_timestamp);
		rohc_debugf(3, "mptr %p\n",pmptr->uint8);
	}
	else
	{
		// Not present
		assert( tss_flag == 0 );
	}
}


int d_variable_unscaled_timestamp_size( multi_ptr_t mptr, int tss_flag, int tsc_flag )
{
	int size;

	rohc_debugf(3, "mptr %p tss_flag %d tsc_flag %d\n",mptr.uint8,tss_flag,tsc_flag);

	if(tsc_flag == 0)
	{
		// Present
		size = d_sdvl_size(*mptr.uint8);
	}
	else
	{
		// Not present
		assert( tss_flag == 0 );
		size = 0;
	}

	rohc_debugf(3, "return size %d\n",size);
	return size;
}


#endif // SUPPORT_ROHC_V2

