/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/**
 * @file c_rohc_v2.c
 * @brief ROHC compression context for the RoHCv2 profile.
 * @author FWX <rohc_team@dialine.fr>
 */
#include "c_generic.h"
#include "config.h" /* for RTP_BIT_TYPE and ROHC_DEBUG_LEVEL definitions */

#if SUPPORT_ROHC_V2

#include "rohc_v2.h"
#include "c_rohc_v2.h"
#include "rohc_traces.h"
#include "rohc_packets.h"
#include "rfc5225_encoding.h"
#include "cid.h"
#include "crc.h"
#include "protocols/ipproto.h"

#include "trace.h"

#include <assert.h>

#include <stdlib.h>
#include <string.h>

/*
 * Private datas.
 */


/*
 * Private function prototypes.
 */
u_int8_t * rohc_v2_code_static_ipv6_option_part(struct sc_v2_context *v2_context,
                                                ip_context_ptr_t ip_context,
                                                multi_ptr_t mptr,
                                                u_int8_t protocol,
                                                base_header_ip_t base_header,
                                                const int packet_size);
u_int8_t * rohc_v2_code_dynamic_ipv6_option_part(struct sc_v2_context *v2_context,
                                                 ip_context_ptr_t ip_context,
                                                 multi_ptr_t mptr,
                                                 u_int8_t protocol,
                                                 base_header_ip_t base_header,
                                                 const int packet_size);
u_int8_t * rohc_v2_code_irregular_ipv6_option_part(struct sc_v2_context *v2_context,
                                                   ip_context_ptr_t ip_context,
                                                   multi_ptr_t mptr,
                                                   u_int8_t protocol,
                                                   base_header_ip_t base_header,
                                                   const int packet_size);
u_int8_t * rohc_v2_code_static_ip_part(struct sc_v2_context *v2_context,
                                       ip_context_ptr_t ip_context,
                                       multi_ptr_t mptr,
                                       base_header_ip_t base_header,
                                       const int packet_size,
                                       int is_innermost);
u_int8_t * rohc_v2_code_dynamic_ip_part(struct sc_v2_context *v2_context,
                                        ip_context_ptr_t ip_inner_context,
                                        ip_context_ptr_t ip_context,
                                        multi_ptr_t mptr,
                                        base_header_ip_t base_header,
                                        const int packet_size,
                                        int profile_value);
u_int8_t * rohc_v2_code_irregular_ip_part(ip_context_ptr_t ip_context,
                                          multi_ptr_t mptr,
                                          base_header_ip_t base_header,
                                          const int packet_size,
                                          int is_innermost,
                                          int outer_ip_flag);
u_int8_t * c_ip_baseheader(struct c_context *const context,
                           struct sc_v2_context *v2_context,
                           ip_context_ptr_t ip_inner_context,
                           multi_ptr_t c_base_header,
                           base_header_ip_t base_header_inner,
                           int outer_ip_flag);
u_int8_t * c_rohc_v2_code_udp_static_part(udp_context_t *udp_context,
                                          multi_ptr_t mptr,
                                          base_header_udp_t *udp);
u_int8_t * c_rohc_v2_code_udp_dynamic_part(struct sc_v2_context *v2_context,
                                           int profile,
                                           udp_context_t *udp_context,
                                           multi_ptr_t mptr,
                                           base_header_udp_t *udp);
u_int8_t * c_rohc_v2_code_udp_irregular_part(udp_context_t *udp_context,
                                             multi_ptr_t mptr,
                                             base_header_udp_t *udp);
u_int8_t * c_rohc_v2_code_rtp_static_part(rtp_context_t *rtp_context,
                                          multi_ptr_t mptr,
                                          base_header_rtp_t *rtp);
u_int8_t * c_rohc_v2_code_rtp_dynamic_part(struct sc_v2_context *v2_context,
                                           rtp_context_t *rtp_context,
                                           multi_ptr_t mptr,
                                           base_header_rtp_with_csrc_t *rtp);
u_int8_t * c_rohc_v2_code_udp_lite_static_part(udp_lite_context_t *udp_lite_context,
                                               multi_ptr_t mptr,
                                               base_header_udp_lite_t *udp_lite);
u_int8_t * c_rohc_v2_code_udp_lite_dynamic_part(struct sc_v2_context *v2_context,
                                                udp_lite_context_t *udp_lite_context,
                                                multi_ptr_t mptr,
                                                base_header_udp_lite_t *udp_lite,
                                                int profile_value);
u_int8_t * c_rohc_v2_code_udp_lite_irregular_part(udp_lite_context_t *udp_lite_context,
                                                  multi_ptr_t mptr,
                                                  base_header_udp_lite_t *udp_lite);
u_int8_t * c_udp_lite_baseheader(struct c_context *const context,
                                 struct sc_v2_context *v2_context,
                                 ip_context_ptr_t ip_inner_context,
                                 udp_lite_context_t *udp_lite_context,
                                 multi_ptr_t c_base_header,
                                 base_header_ip_t base_header_inner,
                                 int profile_value, int outer_ip_flag );
u_int8_t * c_rohc_v2_code_esp_static_part(esp_context_t *esp_context,
                                          multi_ptr_t mptr,
                                          base_header_esp_t *esp);
u_int8_t * c_rohc_v2_code_esp_dynamic_part(struct sc_v2_context *v2_context,
                                           esp_context_t *esp_context,
                                           multi_ptr_t mptr,
                                           base_header_esp_t *esp);
u_int8_t * c_udp_baseheader(struct c_context *const context,
                            struct sc_v2_context *v2_context,
                            ip_context_ptr_t ip_inner_context,
                            multi_ptr_t c_base_header,
                            base_header_ip_t base_header_inner,
                            int profile_value, int outer_ip_flag );
u_int8_t * c_esp_baseheader(struct c_context *const context,
                            struct sc_v2_context *v2_context,
                            ip_context_ptr_t ip_inner_context,
                            esp_context_t *esp_context,
                            multi_ptr_t c_base_header,
                            base_header_ip_t base_header_inner,
                            base_header_esp_t *esp,
                            int profile_value, int outer_ip_flag);
u_int8_t * c_rtp_baseheader(struct c_context *const context,
                            struct sc_v2_context *v2_context,
                            ip_context_ptr_t ip_inner_context,
                            rtp_context_t *rtp_context,
                            multi_ptr_t c_base_header,
                            base_header_ip_t base_header_inner,
                            base_header_rtp_with_csrc_t *rtp,
                            int profile_value, int outer_ip_flag);


//int tcp_changed_tcp_dynamic(const struct c_context *context,
//                            const tcphdr_t *tcp);


/**
 * @brief Create a new UDP/RTP/ESP context and initialize it thanks to the given IP
 *        packet.
 *
 * This function is one of the functions that must exist in one profile for the
 * framework to work.
 *
 * @param context The compression context
 * @param ip      The IP packet given to initialize the new context
 * @return        1 if successful, 0 otherwise
 */
int c_rohc_v2_create(struct c_context *const context, const struct ip_packet *ip )
{

	struct c_generic_context *g_context;
	struct sc_v2_context *v2_context;
	const c_v2_profile_t *profile;
	ip_context_ptr_t ip_context;
	base_header_ip_t base_header_inner;   // Source innermost
	base_header_ip_t base_header;   // Source
	u_int8_t protocol;
	int size_option;
	int size_context;
	int size;

	profile = (const c_v2_profile_t *) context->profile;
	rohc_debugf(3, "context %p ip %p profil %p profil id %Xh protocol %d\n",context,ip,profile,
	            profile->profile.id,
	            profile->profile.protocol);

	/* create and initialize the generic part of the profile context */
	if(!c_generic_create(context, 0, ip))
	{
		rohc_debugf(0, "generic context creation failed\n");
		goto quit;
	}
	g_context = (struct c_generic_context *) context->specific;

	// Init pointer to the initial packet
	base_header.ipvx = (base_header_ipvx_t *)ip->data;

	size = 0;
	size_context = 0;

	do
	{

		rohc_debugf(3, "base_header %p IP version %d\n",base_header.uint8,base_header.ipvx->version);

		base_header_inner.ipvx = base_header.ipvx;
		TraceIp(base_header.ipvx);

		switch(base_header.ipvx->version)
		{
			case IPV4:
				// If options present
				if(base_header.ipv4->hdr_length != 5)
				{
					return -1;
				}
				// No fragmentation
				if(base_header.ipv4->mf != 0 || base_header.ipv4->rf != 0)
				{
					return -1;
				}
				/* get the transport protocol */
				protocol = base_header.ipv4->protocol;
				size += sizeof(base_header_ipv4_t);
				size_context += sizeof(ipv4_context_t);
				++base_header.ipv4;
				break;
			case IPV6:
				protocol = base_header.ipv6->next_header;
				size += sizeof(base_header_ipv6_t);
				size_context += sizeof(ipv6_context_t);
				++base_header.ipv6;
				while( ( ipproto_specifications[protocol] & IPV6_OPTION ) != 0)
				{
					TraceIpV6option(protocol,base_header);
					switch(protocol)
					{
						case IPPROTO_HOPOPTS:        // IPv6 Hop-by-Hop options
							size_option = ( base_header.ipv6_opt->length + 1 ) << 3;
							size_context += MAX_IPV6_CONTEXT_OPTION_SIZE;
							break;
						case IPPROTO_ROUTING:        // IPv6 routing header
							size_option = ( base_header.ipv6_opt->length + 1 ) << 3;
							size_context += MAX_IPV6_CONTEXT_OPTION_SIZE;
							break;
						case IPPROTO_GRE:
							size_option = base_header.ip_gre_opt->c_flag +
							              base_header.ip_gre_opt->k_flag +
							              base_header.ip_gre_opt->s_flag + 1;
							size_option <<= 3;
							size_context = sizeof(ipv6_gre_option_context_t);
							break;
						case IPPROTO_DSTOPTS:        // IPv6 destination options
							size_option = ( base_header.ipv6_opt->length + 1 ) << 3;
							size_context += MAX_IPV6_CONTEXT_OPTION_SIZE;
							break;
						case IPPROTO_MIME:
							size_option = ( 2 + base_header.ip_mime_opt->s_bit ) << 3;
							size_context = sizeof(ipv6_mime_option_context_t);
							break;
						case IPPROTO_AH:
							size_option = sizeof(ip_ah_opt_t) - sizeof(u_int32_t) +
							              ( base_header.ip_ah_opt->length << 4 ) - sizeof(int32_t);
							size_context = sizeof(ipv6_ah_option_context_t);
							break;
						// case IPPROTO_ESP : ???
						default:
							return -1;
					}
					protocol = base_header.ipv6_opt->next_header;
					size += size_option;
					base_header.uint8 += size_option;
				}
				fflush(stdout);
				break;
			default:
				return -1;
		}

	}
	while( ( ipproto_specifications[protocol] & IP_TUNNELING ) != 0 && size < ip->size);

	if(size >= ip->size)
	{
		return -1;
	}

	switch(profile->profile.id)
	{
		case ROHCv2_PROFILE_RTP:
			++base_header.udp;
			if(base_header.rtp->rtp_version != 2)
			{
				rohc_debugf(3, "Not RTP version 2!\n");
				goto clean;
			}
			break;
		case ROHCv2_PROFILE_UDPLITE_RTP:
			++base_header.udp_lite;
			if(base_header.rtp->rtp_version != 2)
			{
				rohc_debugf(3, "Not RTP version 2!\n");
				goto clean;
			}
			break;
	}

	size_context += profile->context_size + 1;

	/* create the RoHCv2 part of the profile context */
	v2_context = malloc(sizeof(struct sc_v2_context) + size_context);
	if(v2_context == NULL)
	{
		rohc_debugf(0, "no memory for the specific part of the profile context\n");
		goto clean;
	}
	g_context->specific = v2_context;

	/* initialize the specific context of the profile context */
	memset(v2_context->ip_context,0,size_context);

	// Init pointer to the initial packet
	base_header.ipvx = (base_header_ipvx_t *)ip->data;
	ip_context.uint8 = v2_context->ip_context;

	do
	{

		rohc_debugf(3, "base_header %p IP version %d\n",base_header.uint8,base_header.ipvx->version);

		base_header_inner.ipvx = base_header.ipvx;

		ip_context.vx->version = base_header.ipvx->version;
		rohc_debugf(3, "ip_context %p version %d\n",ip_context.vx,ip_context.vx->version);
		ip_context.vx->inner = 0;

		switch(base_header.ipvx->version)
		{
			case IPV4:
				ip_context.v4->last_ip_id.uint16 = ntohs(base_header.ipv4->ip_id);
				rohc_debugf(3, "IP-ID %4.4Xh\n", ip_context.v4->last_ip_id.uint16);
				ip_context.v4->ip_id_behavior = IP_ID_BEHAVIOR_UNKNOWN;
				ip_context.v4->last_ip_id_behavior = IP_ID_BEHAVIOR_UNKNOWN;
				/* get the transport protocol */
				protocol = base_header.ipv4->protocol;
				ip_context.v4->protocol = protocol;
				ip_context.v4->df = base_header.ipv4->df;
				ip_context.v4->ttl_hopl = base_header.ipv4->ttl_hopl;
				ip_context.v4->src_addr = base_header.ipv4->src_addr;
				ip_context.v4->dst_addr = base_header.ipv4->dest_addr;
//rohc_debugf(3, "IPv4 context at %p size %d\n",ip_context.uint8,(int)sizeof(ipv4_context_t));
//TraceData(ip_context.uint8,(int)sizeof(ipv4_context_t));
				++base_header.ipv4;
				++ip_context.v4;
				break;
			case IPV6:
				ip_context.v6->ip_id_behavior = IP_ID_BEHAVIOR_RANDOM;
				/* get the transport protocol */
				protocol = base_header.ipv6->next_header;
				ip_context.v6->next_header = protocol;
				ip_context.v6->ttl_hopl = base_header.ipv6->ttl_hopl;
				ip_context.v6->flow_label1 = base_header.ipv6->flow_label1;
				ip_context.v6->flow_label2 = base_header.ipv6->flow_label2;
				memcpy(ip_context.v6->src_addr,base_header.ipv6->src_addr,sizeof(u_int32_t) * 4 * 2);
//rohc_debugf(3, "IPv6 context at %p size %d\n",ip_context.uint8,(int)sizeof(ipv6_context_t));
//TraceData(ip_context.uint8,(int)sizeof(ipv6_context_t));
				++base_header.ipv6;
				++ip_context.v6;
				while( ( ipproto_specifications[protocol] & IPV6_OPTION ) != 0)
				{
					switch(protocol)
					{
						case IPPROTO_HOPOPTS:             // IPv6 Hop-by-Hop options
						case IPPROTO_ROUTING:             // IPv6 routing header
						case IPPROTO_DSTOPTS:             // IPv6 destination options
							size_option = ( base_header.ipv6_opt->length + 1 ) << 3;
							ip_context.v6_option->context_length = 2 + size_option;
							memcpy(&ip_context.v6_option->next_header,&base_header.ipv6_opt->next_header,
							       size_option);
							break;
						case IPPROTO_GRE:
							ip_context.v6_gre_option->context_length = sizeof(ipv6_gre_option_context_t);
							size_option = base_header.ip_gre_opt->c_flag +
							              base_header.ip_gre_opt->k_flag +
							              base_header.ip_gre_opt->s_flag + 1;
							size_option <<= 3;
							ip_context.v6_gre_option->c_flag = base_header.ip_gre_opt->c_flag;
							ip_context.v6_gre_option->k_flag = base_header.ip_gre_opt->k_flag;
							ip_context.v6_gre_option->s_flag = base_header.ip_gre_opt->s_flag;
							ip_context.v6_gre_option->protocol = base_header.ip_gre_opt->protocol;
							ip_context.v6_gre_option->key =
							   base_header.ip_gre_opt->datas[base_header.ip_gre_opt->c_flag];
							ip_context.v6_gre_option->sequence_number =
							   base_header.ip_gre_opt->datas[base_header.ip_gre_opt->c_flag +
							                                 base_header.ip_gre_opt->k_flag];
							break;
						case IPPROTO_MIME:
							size_option = ( 2 + base_header.ip_mime_opt->s_bit ) << 3;
							ip_context.v6_mime_option->context_length = sizeof(ipv6_mime_option_context_t);
							ip_context.v6_mime_option->next_header = base_header.ipv6_opt->next_header;
							ip_context.v6_mime_option->s_bit = base_header.ip_mime_opt->s_bit;
							ip_context.v6_mime_option->res_bits = base_header.ip_mime_opt->res_bits;
							ip_context.v6_mime_option->checksum = base_header.ip_mime_opt->checksum;
							ip_context.v6_mime_option->orig_dest = base_header.ip_mime_opt->orig_dest;
							ip_context.v6_mime_option->orig_src = base_header.ip_mime_opt->orig_src;
							break;
						case IPPROTO_AH:
							size_option = sizeof(ip_ah_opt_t) - sizeof(u_int32_t) +
							              ( base_header.ip_ah_opt->length << 4 ) - sizeof(int32_t);
							ip_context.v6_ah_option->context_length = sizeof(ipv6_ah_option_context_t);
							ip_context.v6_ah_option->next_header = base_header.ipv6_opt->next_header;
							ip_context.v6_ah_option->length = base_header.ip_ah_opt->length;
							ip_context.v6_ah_option->spi = base_header.ip_ah_opt->spi;
							ip_context.v6_ah_option->sequence_number =
							   base_header.ip_ah_opt->sequence_number;
							break;
						// case IPPROTO_ESP : ???
						default:
							return -1;
					}
					rohc_debugf(3, "protocol %d ipv6_option_context %p next_header %d\n",protocol,
					            ip_context.uint8,
					            ip_context.v6_option->next_header);
					ip_context.v6_option->option_length = size_option;
//rohc_debugf(3, "IPv6 option context at %p size %d\n",ip_context.uint8,ip_context.v6_option->context_length);
//TraceData(ip_context.uint8,ip_context.v6_option->context_length);
					protocol = base_header.ipv6_opt->next_header;
					base_header.uint8 += size_option;
					ip_context.uint8 += ip_context.v6_option->context_length;
				}
				fflush(stdout);
				break;
			default:
				return -1;
		}

	}
	while( ( ipproto_specifications[protocol] & IP_TUNNELING ) != 0 && size < ip->size);

	rohc_debugf(3, "protocol %d XXX_context %p\n",protocol,ip_context.uint8);

	// Last IP context is the inner
	ip_context.vx->inner = 1;

	/* init the Master Sequence Number to a random value */
	v2_context->msn = random();
	fflush(stdout);

	switch(profile->profile.id)
	{
		case ROHCv2_PROFILE_RTP:
		case ROHCv2_PROFILE_UDP:
			ip_context.udp->protocol = IPPROTO_UDP;
			ip_context.udp->src_port = base_header.udp->src_port;
			ip_context.udp->dst_port = base_header.udp->dst_port;
			ip_context.udp->checksum_used = base_header.udp->checksum != 0;
			if(profile->profile.id == ROHCv2_PROFILE_RTP)
			{
				++base_header.udp;
				++ip_context.udp;
				goto idem_udplite_rtp;
			}
			break;
		case ROHCv2_PROFILE_ESP:
			ip_context.esp->protocol = IPPROTO_ESP;
			ip_context.esp->spi = base_header.esp->spi;
			ip_context.esp->sequence_number = ntohl(base_header.esp->sequence_number);
			v2_context->msn = (u_int16_t) ntohl(ip_context.esp->sequence_number);
			++ip_context.esp;
			break;
		case ROHCv2_PROFILE_IP:
			ip_context.ip->protocol = protocol;
			rohc_debugf(3, "init IP-Only context with protocol %d\n",protocol);
			break;
		case ROHCv2_PROFILE_UDPLITE_RTP:
		case ROHCv2_PROFILE_UDPLITE:
			ip_context.udp_lite->protocol = IPPROTO_UDPLITE;
			ip_context.udp_lite->src_port = base_header.udp_lite->src_port;
			ip_context.udp_lite->dst_port = base_header.udp_lite->dst_port;
			ip_context.udp_lite->coverage_behavior =
			   ip_context.udp_lite->last_coverage_behavior = UDP_LITE_COVERAGE_UNKNOWN;
			if(profile->profile.id == ROHCv2_PROFILE_UDPLITE_RTP)
			{
				++base_header.udp_lite;
				++ip_context.udp_lite;
idem_udplite_rtp:
				ip_context.rtp->protocol = IPPROTO_UDP;
				ip_context.rtp->ssrc = base_header.rtp->ssrc;
				ip_context.rtp->ts_stride = TS_STRIDE_DEFAULT;
				ip_context.rtp->last_ts_stride = TS_STRIDE_DEFAULT;
				ip_context.rtp->time_stride = TIME_STRIDE_DEFAULT;
				ip_context.rtp->last_time_stride = TIME_STRIDE_DEFAULT;
				rohc_debugf(3, "init RTP context %p\n",ip_context.rtp);
				// Create the ts_sc_comp object
				// c_create_sc(&ip_context.rtp->ts_comp);
				v2_context->msn = ntohs(base_header.rtp->sequence_number);
			}
			break;
	}

	v2_context->v2_seq_number_change_count = 0;
	v2_context->v2_last_seq_number = -1;
	v2_context->reorder_ratio = REORDERING_NONE;

	/* init the RoHCv2-specific temporary variables DBX */
//	v2_context->tmp.send_tcp_dynamic = -1;

	rohc_debugf(3, "msn %4.4Xh\n", v2_context->msn);
	fflush(stdout);
	/* init the RoHCv2-specific variables and functions */
	g_context->next_header_proto = profile->profile.protocol;
//	g_context->next_header_len = sizeof(tcphdr_t); // + options ???
//	g_context->decide_state = tcp_decide_state;
	g_context->decide_state = NULL;
	g_context->init_at_IR = NULL;
	g_context->code_static_part = NULL;
//	g_context->code_dynamic_part = rohc_v2_code_dynamic_tcp_part;
	g_context->code_UO_packet_head = NULL;
//	g_context->code_UO_packet_tail = NULL;
// A REVOIR	g_context->compute_crc_static = tcp_compute_crc_static;
// A REVOIR	g_context->compute_crc_dynamic = tcp_compute_crc_dynamic;

	return 1;

clean:
	c_generic_destroy(context);
quit:
	return 0;
}


/**
 * @brief Destroy a UDP/RTP/ESP context.
 *
 * This function is one of the functions that must exist in one profile for the
 * framework to work.
 *
 * @param context The compression context
 * @return        None
 */
void c_rohc_v2_destroy(struct c_context *const context)
{
	struct c_generic_context *g_context;
	struct sc_v2_context *v2_context;
	ip_context_ptr_t ip_context;
	u_int8_t protocol;

	rohc_debugf(3, "context %p\n",context);

	g_context = (struct c_generic_context *) context->specific;
	v2_context = (struct sc_v2_context *) g_context->specific;

	// Init pointer to the initial context
	ip_context.uint8 = v2_context->ip_context;

	do
	{

		protocol = ip_context.vx->next_header;

		switch(ip_context.vx->version)
		{
			case IPV4:
				++ip_context.v4;
				break;
			case IPV6:
				++ip_context.v6;
				while( ( ipproto_specifications[protocol] & IPV6_OPTION ) != 0)
				{
					protocol = ip_context.v6_option->next_header;
					ip_context.uint8 += ip_context.v6_option->context_length;
				}
				break;
		}

	}
	while( ( ipproto_specifications[protocol] & IP_TUNNELING ) != 0);

	switch(protocol)
	{
		case IPPROTO_UDP:
			++ip_context.udp;
			protocol = *ip_context.uint8;
			break;
		case IPPROTO_UDPLITE:
			++ip_context.udp_lite;
			protocol = *ip_context.uint8;
			break;
	}
	// If RTP
	if(protocol == IPPROTO_UDP)
	{
		// Destroy the ts_sc_comp object
		// c_destroy_sc(&ip_context.rtp->ts_comp);
	}
	else
	{
		rohc_debugf(3, "WARNING protocol %d!\n",protocol);
	}

	// Destroy context
	c_generic_destroy(context);
}


/**
 * @brief Check if the IP/RoHCv2 packet belongs to the context
 *
 * Conditions are:
 *  - the number of IP headers must be the same as in context
 *  - IP version of the two IP headers must be the same as in context
 *  - IP packets must not be fragmented
 *  - the source and destination addresses of the two IP headers must match the
 *    ones in the context
 *  - the transport protocol must be the same
 *  - the source and destination ports of the UDP/RTP header must match the ones in
 *    the context
 *  - IPv6 only: the Flow Label of the two IP headers must match the ones the
 *    context
 *
 * This function is one of the functions that must exist in one profile for the
 * framework to work.
 *
 * @param context The compression context
 * @param ip      The IP packet to check
 * @return        1 if the IP packet belongs to the context,
 *                0 if it does not belong to the context and
 *                -1 if the profile cannot compress it or an error occurs
 */
int c_rohc_v2_check_context(const struct c_context *context,
                            const struct ip_packet *ip)
{
	struct c_generic_context *g_context;
	struct sc_v2_context *v2_context;
	const struct c_profile *profile;
	ip_context_ptr_t ip_context;
	base_header_ip_t base_header;   // Source
	u_int8_t protocol;
	int size;

	rohc_debugf(3, "context %p ip %p\n",context,ip);

	g_context = (struct c_generic_context *) context->specific;
	v2_context = (struct sc_v2_context *) g_context->specific;
	profile = context->profile;

	// Init pointer to the initial packet
	base_header.ipvx = (base_header_ipvx_t *)ip->data;
	ip_context.uint8 = v2_context->ip_context;
	size = ip->size;

	do
	{

		rohc_debugf(3, "  base_header %p IP version %d\n",base_header.uint8,base_header.ipvx->version);

		if(base_header.ipvx->version != ip_context.vx->version)
		{
			rohc_debugf(3, "  Not same IP version\n");
			goto bad_context;
		}

		switch(base_header.ipvx->version)
		{
			case IPV4:
				// No option
				if(base_header.ipv4->hdr_length != 5)
				{
					goto bad_context;
				}
				// No fragmentation
				if(base_header.ipv4->mf != 0 || base_header.ipv4->rf != 0)
				{
					goto bad_context;
				}
				if(base_header.ipv4->src_addr != ip_context.v4->src_addr ||
				   base_header.ipv4->dest_addr != ip_context.v4->dst_addr)
				{
					rohc_debugf(3, "  Not same IPV4 addresses\n");
					goto bad_context;
				}
				rohc_debugf(3, "  Same IPV4 addresses\n");
				/* get the transport protocol */
				protocol = base_header.ipv4->protocol;
				if(base_header.ipv4->protocol != ip_context.v4->protocol)
				{
					rohc_debugf(3, "  IPV4 Not same protocol\n");
					goto bad_context;
				}
				rohc_debugf(3, "  IPV4 same protocol %d\n",protocol);
				++base_header.ipv4;
				++ip_context.v4;
				size -= sizeof(base_header_ipv4_t);
				break;
			case IPV6:
				if(memcmp(base_header.ipv6->src_addr,ip_context.v6->src_addr,sizeof(u_int32_t) * 4 *
				          2) != 0)
				{
					rohc_debugf(3, "  Not same IPV6 addresses\n");
					goto bad_context;
				}
				rohc_debugf(3, "  Same IPV6 addresses\n");
				if(base_header.ipv6->flow_label1 != ip_context.v6->flow_label1 ||
				   base_header.ipv6->flow_label2 != ip_context.v6->flow_label2)
				{
					rohc_debugf(3, "  Not same IPV6 flow label\n");
					goto bad_context;
				}
				protocol = base_header.ipv6->next_header;
//			                 rohc_debugf(3, "  IPV6 protocol %d  context %p next_header %d\n",protocol,ip_context.uint8,ip_context.v6->next_header);
				if(protocol != ip_context.v6->next_header)
				{
					rohc_debugf(3, "  IPV6 Not same protocol %d\n",protocol);
					goto bad_context;
				}
				++base_header.ipv6;
				++ip_context.v6;
				size -= sizeof(base_header_ipv6_t);
				while( ( ipproto_specifications[protocol] & IPV6_OPTION ) != 0 && size < ip->size)
				{
///rohc_debugf(3, "IPv6 option context at %p size %d\n",ip_context.uint8,ip_context.v6_option->context_length);
//TraceData(ip_context.uint8,ip_context.v6_option->context_length);
					protocol = base_header.ipv6_opt->next_header;
					rohc_debugf(3, "  IPv6 protocol %d context %p option %d\n",protocol,ip_context.uint8,
					            ip_context.v6_option->next_header);
					//	rohc_debugf(3, "IPv6 option %d at %p\n",protocol,base_header.uint8);
					if(protocol != ip_context.v6_option->next_header)
					{
						rohc_debugf(3, " not same IPv6 option (%d != %d)\n",protocol,
						            ip_context.v6_option->next_header);
						goto bad_context;
					}
					rohc_debugf(3, "  Same IPv6 option %d\n",protocol);
//			                 rohc_debugf(3, "      option_length %d context_length %d\n",ip_context.v6_option->option_length,ip_context.v6_option->context_length);
					base_header.uint8 += ip_context.v6_option->option_length;
					ip_context.uint8 += ip_context.v6_option->context_length;
				}
				fflush(stdout);
				break;
			default:
				return -1;
		}

	}
	while( ( ipproto_specifications[protocol] & IP_TUNNELING ) != 0 && size < ip->size);

	if(size >= ip->size)
	{
		return -1;
	}

	if(protocol != *ip_context.uint8)
	{
		rohc_debugf(3, "  not same protocol %d != %d\n",protocol,*ip_context.uint8);
		return -1;
	}

#ifdef LMJSMLQJKS
	switch(protocol)
	{
		case IPPROTO_UDP:
			if(ip_context.udp->src_port != base_header.udp->src_port ||
			   ip_context.udp->dst_port != base_header.udp->dst_port)
			{
				rohc_debugf(3, "  UDP not same src or dst port\n");
				goto bad_context;
			}
			break;
		case IPPROTO_ESP:
			if(ip_context.esp->spi != base_header.esp->spi)
			{
				rohc_debugf(3, "  ESP not same spi\n");
				goto bad_context;
			}
			break;
		case IPPROTO_UDPLITE:
			if(ip_context.udp_lite->src_port != base_header.udp_lite->src_port ||
			   ip_context.udp_lite->dst_port != base_header.udp_lite->dst_port)
			{
				rohc_debugf(3, "  UDP-Lite not same src or dst port\n");
				goto bad_context;
			}
			break;
	}
#endif

	switch(context->profile->id)
	{
		case ROHCv2_PROFILE_RTP:
			if(ip_context.udp->src_port != base_header.udp->src_port ||
			   ip_context.udp->dst_port != base_header.udp->dst_port)
			{
				rohc_debugf(3, "  RTP not same src or dst port\n");
				goto bad_context;
			}
			rohc_debugf(3, "  RTP same src or dst port\n");
			break;
		case ROHCv2_PROFILE_UDP:
			if(ip_context.udp->src_port != base_header.udp->src_port ||
			   ip_context.udp->dst_port != base_header.udp->dst_port)
			{
				rohc_debugf(3, "  UDP not same src or dst port\n");
				goto bad_context;
			}
			rohc_debugf(3, "  UDP same src or dst port\n");
			break;
		case ROHCv2_PROFILE_ESP:
			if(ip_context.esp->spi != base_header.esp->spi)
			{
				rohc_debugf(3, "  ESP not same spi\n");
				goto bad_context;
			}
			rohc_debugf(3, "  ESP same spi\n");
			break;
		case ROHCv2_PROFILE_IP:
			break;
		case ROHCv2_PROFILE_UDPLITE_RTP:
			if(ip_context.udp_lite->src_port != base_header.udp_lite->src_port ||
			   ip_context.udp_lite->dst_port != base_header.udp_lite->dst_port)
			{
				rohc_debugf(3, "  UDP-Lite RTP not same src or dst port\n");
				goto bad_context;
			}
			rohc_debugf(3, "  UDP-Lite RTP same src or dst port\n");
			break;
		case ROHCv2_PROFILE_UDPLITE:
			if(ip_context.udp_lite->src_port != base_header.udp_lite->src_port ||
			   ip_context.udp_lite->dst_port != base_header.udp_lite->dst_port)
			{
				rohc_debugf(3, "  UDP-Lite not same src or dst port\n");
				goto bad_context;
			}
			rohc_debugf(3, "  UDP-Lite same src or dst port\n");
			break;
	}

	return 1;

bad_context:
	return 0;
//error:
//	return -1;
}


/**
 * @brief Encode an IP packet according to a pattern decided by several
 *        different factors.
 *
 * @param context        The compression context
 * @param ip             The IP packet to encode
 * @param packet_size    The length of the IP packet to encode
 * @param dest           The rohc-packet-under-build buffer
 * @param dest_size      The length of the rohc-packet-under-build buffer
 * @param packet_type    OUT: The type of ROHC packet that is created
 * @param payload_offset The offset for the payload in the IP packet
 * @return               The length of the created ROHC packet
 *                       or -1 in case of failure
 */
int c_rohc_v2_encode(struct c_context *const context,
                     const struct ip_packet *ip,
                     const int packet_size,
                     unsigned char *const dest,
                     const int dest_size,
                     rohc_packet_t *const packet_type,
                     int *const payload_offset)
{
	struct c_generic_context *g_context;
	struct sc_v2_context *v2_context;
	const c_v2_profile_t *profile;
	ip_context_ptr_t ip_inner_context;
	ip_context_ptr_t ip_context;
	base_header_ip_t base_header_inner;   // Source innermost
	base_header_ip_t base_header;   // Source
	u_int8_t new_context_state;
	u_int8_t ip_id_behaviors[MAX_IP_HEADERS];
	u_int8_t ip_id_behaviors_count = 0;
	multi_ptr_t mptr;
	u_int8_t protocol;
	u_int8_t packet_id;
	int profile_value;
	int first_position;
	int crc_position;
	int counter;
	int size;

	rohc_debugf(3, "context %p ip %p packet_size %d dest %p dest_size %d\n",context,ip,packet_size,
	            dest,
	            dest_size);

	g_context = (struct c_generic_context *) context->specific;
	v2_context = (struct sc_v2_context *) g_context->specific;
	profile = (const c_v2_profile_t *) context->profile;
	profile_value = profile->profile.id;

	rohc_debugf(3, "Packet to compress, length %d\n",ip->size);
	TraceData((unsigned char*)ip->data,ip->size);
	fflush(stdout);

	// Init pointer to the initial packet
	base_header.ipvx = (base_header_ipvx_t *)ip->data;
	ip_context.uint8 = v2_context->ip_context;

	new_context_state = context->state;
#ifdef lQJLSJKSQs
	switch(context->state)
	{
		case IR:  /* The Initialization and Refresh (IR) state */
			new_context_state = FO;
			break;
		case FO:  /* The First Order (FO) state */
			new_context_state = SO;
			break;
		case SO:  /* The Second Order (SO) state */
		default:
			new_context_state = SO;
			break;
	}
#endif

	size = 0;

	do
	{

		rohc_debugf(3, "base_header %p IP version %d\n",base_header.uint8,base_header.ipvx->version);

		base_header_inner.ipvx = base_header.ipvx;
		ip_inner_context.uint8 = ip_context.uint8;

		switch(base_header.ipvx->version)
		{
			case IPV4:
				TraceIpV4(base_header.ipv4);
				// If innermost
				if(ip_context.vx->inner != 0)
				{
					WB_t swapped_ip_id;
					WB_t ip_id;

					/* Try to determine the IP_ID behavior */
					ip_id.uint16 = ntohs(base_header_inner.ipv4->ip_id);
					rohc_debugf(3, "ip_id_behavior %d last_ip_id %Xh ip_id %Xh\n",
					            ip_inner_context.v4->ip_id_behavior,
					            ip_inner_context.v4->last_ip_id.uint16,
					            ip_id.uint16);

					switch(ip_inner_context.v4->ip_id_behavior)
					{
						case IP_ID_BEHAVIOR_SEQUENTIAL:
							if( (ip_inner_context.v4->last_ip_id.uint16 + 1) != ip_id.uint16)
							{
								// Problem
								rohc_debugf(3, "ip_id_behavior not SEQUENTIAL %Xh +1  != %Xh\n",
								            ip_inner_context.v4->last_ip_id.uint16,ip_id.uint16);
								ip_inner_context.v4->ip_id_behavior = IP_ID_BEHAVIOR_RANDOM;
								break;
							}
							break;
						case IP_ID_BEHAVIOR_SEQUENTIAL_SWAPPED:
							swapped_ip_id.uint8[0] = ip_inner_context.v4->last_ip_id.uint8[1];
							swapped_ip_id.uint8[1] = ip_inner_context.v4->last_ip_id.uint8[0];
							rohc_debugf(3, " swapped_ip_id %4.4Xh + 1 %4.4Xh ip_id %4.4Xh\n",
							            swapped_ip_id.uint16,swapped_ip_id.uint16 + 1,ip_id.uint16);
							++swapped_ip_id.uint16;
							if(swapped_ip_id.uint8[0] != ip_id.uint8[1] || swapped_ip_id.uint8[1] !=
							   ip_id.uint8[0])
							{
								// Problem
								rohc_debugf(3, "ip_id_behavior not SEQUENTIAL_SWAPPED %Xh +1 != %Xh\n",
								            ip_inner_context.v4->last_ip_id.uint16,swapped_ip_id.uint16);
								ip_inner_context.v4->ip_id_behavior = IP_ID_BEHAVIOR_RANDOM;
								break;
							}
							break;
						case IP_ID_BEHAVIOR_RANDOM:
							if( (ip_inner_context.v4->last_ip_id.uint16 + 1) == ip_id.uint16)
							{
								rohc_debugf(3, "ip_id_behavior SEQUENTIAL\n");
								ip_inner_context.v4->ip_id_behavior = IP_ID_BEHAVIOR_SEQUENTIAL;
								break;
							}
							swapped_ip_id.uint8[0] = ip_inner_context.v4->last_ip_id.uint8[1];
							swapped_ip_id.uint8[1] = ip_inner_context.v4->last_ip_id.uint8[0];
							rohc_debugf(3, " swapped_ip_id %4.4Xh + 1 %4.4Xh ip_id %4.4Xh\n",
							            swapped_ip_id.uint16,swapped_ip_id.uint16 + 1,ip_id.uint16);
							++swapped_ip_id.uint16;
							if(swapped_ip_id.uint8[0] == ip_id.uint8[1] && swapped_ip_id.uint8[1] ==
							   ip_id.uint8[0])
							{
								ip_inner_context.v4->ip_id_behavior = IP_ID_BEHAVIOR_SEQUENTIAL_SWAPPED;
								rohc_debugf(3, "ip_id_behavior SEQUENTIAL SWAPPED\n");
								break;
							}
							if(ip_id.uint16 == 0)
							{
								ip_inner_context.v4->ip_id_behavior = IP_ID_BEHAVIOR_ZERO;
								rohc_debugf(3, "ip_id_behavior SEQUENTIAL ZERO\n");
								break;
							}
							break;
						case IP_ID_BEHAVIOR_ZERO:
							if(ip_id.uint16 != 0)
							{
								rohc_debugf(3, "ip_id_behavior not ZERO %Xh != 0\n",ip_id.uint16);
								if(ip_id.uint16 == 0x0001)
								{
									rohc_debugf(3, "ip_id_behavior SEQUENTIAL\n");
									ip_inner_context.v4->ip_id_behavior = IP_ID_BEHAVIOR_SEQUENTIAL;
									break;
								}
								if(ip_id.uint16 == 0x0100)
								{
									ip_inner_context.v4->ip_id_behavior = IP_ID_BEHAVIOR_SEQUENTIAL_SWAPPED;
									rohc_debugf(3, "ip_id_behavior SEQUENTIAL SWAPPED\n");
									break;
								}
								// Problem
								ip_inner_context.v4->ip_id_behavior = IP_ID_BEHAVIOR_RANDOM;
								break;
							}
							break;
						case IP_ID_BEHAVIOR_UNKNOWN:
							if(ip_id.uint16 == 0)
							{
								ip_inner_context.v4->ip_id_behavior = IP_ID_BEHAVIOR_ZERO;
								rohc_debugf(3, "ip_id_behavior ZERO\n");
								break;
							}
							if( (ip_inner_context.v4->last_ip_id.uint16 + 1) == ip_id.uint16)
							{
								ip_inner_context.v4->ip_id_behavior = IP_ID_BEHAVIOR_SEQUENTIAL;
								rohc_debugf(3, "ip_id_behavior SEQUENTIAL\n");
								break;
							}
							if(ip_inner_context.v4->last_ip_id.uint16 == ip_id.uint16)
							{
								break;
							}
							swapped_ip_id.uint8[0] = ip_id.uint8[1];
							swapped_ip_id.uint8[1] = ip_id.uint8[0];
							if( (ip_inner_context.v4->last_ip_id.uint16 + 1) == swapped_ip_id.uint16)
							{
								ip_inner_context.v4->ip_id_behavior = IP_ID_BEHAVIOR_SEQUENTIAL_SWAPPED;
								rohc_debugf(3, "ip_id_behavior SEQUENTIAL_SWAPPED\n");
								break;
							}
							if(ip_inner_context.v4->last_ip_id.uint16 == swapped_ip_id.uint16)
							{
								break;
							}
							ip_inner_context.v4->ip_id_behavior = IP_ID_BEHAVIOR_RANDOM;
							rohc_debugf(3, "ip_id_behavior RANDOM\n");
							break;
						default:
							break;
					}

				}
				else
				{
					if(base_header_inner.ipv4->ip_id == 0)
					{
						ip_context.v4->ip_id_behavior = IP_ID_BEHAVIOR_ZERO;
					}
					else
					{
						ip_context.v4->ip_id_behavior = IP_ID_BEHAVIOR_RANDOM;
					}
				}
				// Copy ip_id_behavior of this header for control_crc3_encoding (see 6.6.11 page 32)
				ip_id_behaviors[ ip_id_behaviors_count++ ] = ip_context.v4->ip_id_behavior;
				rohc_debugf(3, "Add ip_id_behavior %d to control_data_crc\n",
				            ip_context.v4->ip_id_behavior);
				/* get the transport protocol */
				protocol = base_header.ipv4->protocol;
				size += sizeof(base_header_ipv4_t);
				++base_header.ipv4;
				++ip_context.v4;
				break;
			case IPV6:
				TraceIpV6(base_header.ipv6);
				// Copy ip_id_behavior of this header for control_crc3_encoding (see 6.6.11 page 32)
				ip_id_behaviors[ ip_id_behaviors_count++ ] = ip_context.v4->ip_id_behavior;
				rohc_debugf(3, "Add ip_id_behavior %d to control_data_crc\n",
				            ip_context.v4->ip_id_behavior);
				protocol = base_header.ipv6->next_header;
				size += sizeof(base_header_ipv6_t);
				++base_header.ipv6;
				++ip_context.v6;
				while( ( ipproto_specifications[protocol] & IPV6_OPTION ) != 0)
				{
					TraceIpV6option(protocol,base_header);
					switch(protocol)
					{
						case IPPROTO_HOPOPTS:     // IPv6 Hop-by-Hop options
						case IPPROTO_ROUTING:     // IPv6 routing header
						case IPPROTO_DSTOPTS:     // IPv6 destination options
						case IPPROTO_AH:
							if(base_header.ipv6_opt->length != ip_context.v6_option->length)
							{
								rohc_debugf(3, "IPv6 option %d length changed (%d -> %d)\n",protocol,
								            ip_context.v6_option->length,
								            base_header.ipv6_opt->length);
								assert( base_header.ipv6_opt->length < MAX_IPV6_OPTION_LENGTH );
								ip_context.v6_option->option_length =
								   (base_header.ipv6_opt->length + 1) << 3;
								ip_context.v6_option->length = base_header.ipv6_opt->length;
								memcpy(ip_context.v6_option->value,base_header.ipv6_opt->value,
								       ip_context.v6_option->option_length - 2);
								new_context_state = IR;
								break;
							}
							if(memcmp(base_header.ipv6_opt->value,ip_context.v6_option->value,
							          ip_context.v6_option->option_length - 2) != 0)
							{
								rohc_debugf(3, "IPv6 option %d value changed (%d -> %d)\n",protocol,
								            ip_context.v6_option->length,
								            base_header.ipv6_opt->length);
								memcpy(ip_context.v6_option->value,base_header.ipv6_opt->value,
								       ip_context.v6_option->option_length - 2);
								new_context_state = IR;
								break;
							}
							break;
						case IPPROTO_GRE:
							if(base_header.ip_gre_opt->c_flag != ip_context.v6_gre_option->c_flag)
							{
								rohc_debugf(3, "IPv6 option %d c_flag changed (%d -> %d)\n",protocol,
								            ip_context.v6_gre_option->c_flag,
								            base_header.ip_gre_opt->c_flag);
								new_context_state = IR;
								break;
							}
							break;
						case IPPROTO_MIME:
							if(base_header.ip_mime_opt->s_bit != ip_context.v6_mime_option->s_bit)
							{
								rohc_debugf(3, "IPv6 option %d s_bit changed (%Xh -> %Xh)\n",protocol,
								            ip_context.v6_mime_option->s_bit,
								            base_header.ip_mime_opt->s_bit);
								ip_context.v6_option->option_length =
								   (2 + base_header.ip_mime_opt->s_bit) << 3;
								new_context_state = IR;
								break;
							}
							if(base_header.ip_mime_opt->checksum != ip_context.v6_mime_option->checksum)
							{
								rohc_debugf(3, "IPv6 option %d checksum changed (%Xh -> %Xh)\n",protocol,
								            ip_context.v6_mime_option->checksum,
								            base_header.ip_mime_opt->checksum);
								new_context_state = IR;
								break;
							}
							break;
					}
					protocol = base_header.ipv6_opt->next_header;
					base_header.uint8 += ip_context.v6_option->option_length;
					ip_context.uint8 += ip_context.v6_option->context_length;
					size += ip_context.v6_option->option_length;
				}
				break;
			default:
				return -1;
		}
		fflush(stdout);
	}
	while( ( ipproto_specifications[protocol] & IP_TUNNELING ) != 0 && size < ip->size);

	switch(profile_value)
	{
		case ROHCv2_PROFILE_RTP:
		case ROHCv2_PROFILE_UDP:
			TraceUdp(base_header.udp);
			size += sizeof(base_header_udp_t);
			if(profile->profile.id == ROHCv2_PROFILE_RTP)
			{
				++base_header.udp;
				++ip_context.udp;
				goto idem_rtp_udplite;
			}
			break;
		case ROHCv2_PROFILE_ESP:
			TraceEsp(base_header.esp);
			size += sizeof(base_header_esp_t);
			break;
		case ROHCv2_PROFILE_IP:
			break;
		case ROHCv2_PROFILE_UDPLITE_RTP:
		case ROHCv2_PROFILE_UDPLITE:
			TraceUdpLite(base_header.udp_lite);

			rohc_debugf(3, "UDP-Lite ip->size %d size %d checksum_coverage %d\n",ip->size,size,
			            ntohs(base_header.udp_lite->checksum_coverage));
			rohc_debugf(3, "    last_coverage_behavior %d last_checksum_coverage %d\n",
			            ip_context.udp_lite->last_coverage_behavior,
			            ip_context.udp_lite->last_checksum_coverage);

			switch(ip_context.udp_lite->coverage_behavior)
			{
				case UDP_LITE_COVERAGE_INFERRED:
					if( (ip->size - size) != ntohs(base_header.udp_lite->checksum_coverage) )
					{
						ip_context.udp_lite->coverage_behavior = UDP_LITE_COVERAGE_IRREGULAR;
						rohc_debugf(3, "UDP-Lite COVERAGE Set to IRREGULAR\n");
					}
					break;
				case UDP_LITE_COVERAGE_STATIC:
					if(ip_context.udp_lite->last_checksum_coverage !=
					   base_header.udp_lite->checksum_coverage)
					{
						ip_context.udp_lite->coverage_behavior = UDP_LITE_COVERAGE_IRREGULAR;
						rohc_debugf(3, "UDP-Lite COVERAGE Set to IRREGULAR\n");
					}
					break;
				case UDP_LITE_COVERAGE_IRREGULAR:
					if(ip_context.udp_lite->last_checksum_coverage ==
					   base_header.udp_lite->checksum_coverage)
					{
						ip_context.udp_lite->coverage_behavior = UDP_LITE_COVERAGE_STATIC;
						rohc_debugf(3, "UDP-Lite COVERAGE Set to STATIC\n");
					}
					break;
				case UDP_LITE_COVERAGE_UNKNOWN:
					if( (ip->size - size) == ntohs(base_header.udp_lite->checksum_coverage) )
					{
						ip_context.udp_lite->coverage_behavior = UDP_LITE_COVERAGE_INFERRED;
						rohc_debugf(3, "UDP-Lite COVERAGE Set to INFERRED\n");
					}
					else
					{
						ip_context.udp_lite->coverage_behavior = UDP_LITE_COVERAGE_IRREGULAR;
						rohc_debugf(3, "UDP-Lite COVERAGE Set to IRREGULAR\n");
					}
					break;
				default:
					rohc_debugf(3, "UDP-Lite COVERAGE Set to ???\n");
					break;
			}
			rohc_debugf(3, "UDP-Lite COVERAGE Set to %d (last_coverage_behavior %d)\n",
			            ip_context.udp_lite->coverage_behavior,
			            ip_context.udp_lite->last_coverage_behavior);
			size += sizeof(base_header_udp_lite_t);

			if(profile->profile.id == ROHCv2_PROFILE_UDPLITE_RTP)
			{
				// u_int32_t timestamp;

				++base_header.udp_lite;
				++ip_context.udp_lite;
idem_rtp_udplite:
				TraceRtp(base_header.rtp_with_csrc);
				// Add timestamp in the window
				put_rtp_frame_in_window(ip_context.rtp,base_header.rtp);
										 #ifdef LMQJSMLQJs
				/* initialisation of SN with the SN field of the RTP packet */
				g_context->sn = ntohs(base_header.rtp->sequence_number);
				timestamp = ntohl(base_header.rtp_with_csrc->timestamp);
				c_add_ts(&ip_context.rtp->ts_comp, timestamp, g_context->sn);

				if(ip_context.rtp->ts_comp.state == SEND_SCALED)
				{
					int ret;

					/* TS_SCALED value will be send */
					ip_context.rtp->ts_send = get_ts_scaled(ip_context.rtp->ts_comp);
					ret = nb_bits_scaled(ip_context.rtp->ts_comp, &(ip_context.rtp->nr_ts_bits));
					if(ret != 1)
					{
						const uint32_t ts_send = ip_context.rtp->ts_send;
						size_t nr_bits;
						uint32_t mask;

						/* this is the first TS scaled to be sent, we cannot code it with
						  * W-LSB and we must find its size (in bits) */
						for(nr_bits = 1, mask = 1;
						    nr_bits <= 32 && (ts_send & mask) != ts_send;
						    nr_bits++, mask |= (1 << (nr_bits - 1)))
						{
							;
						}
						/*
						rohc_assert((ts_send & mask) == ts_send, error,
						             "size of TS scaled (0x%x) not found, this should "
						              "never happen!", ts_send);
						              */

						rohc_debugf(3, "first TS scaled to be sent: ts_send = %u, "
						            "mask = 0x%x, nr_bits = %zd\n", ts_send, mask, nr_bits);
						ip_context.rtp->nr_ts_bits = nr_bits;
					}

					/* save the new TS_SCALED value */
					add_scaled(&ip_context.rtp->ts_comp, g_context->sn);
					rohc_debugf(3, "ts_scaled = %u on %zd bits\n",
					            ip_context.rtp->ts_send,
					            ip_context.rtp->nr_ts_bits);
				}
				else if(ip_context.rtp->ts_comp.state == INIT_STRIDE)
				{
					/* TS and TS_STRIDE will be send */
					ip_context.rtp->ts_send = timestamp;
					ip_context.rtp->nr_ts_bits = 32;
				}
						#endif

			}
			break;
	}

	fflush(stdout);
	if(new_context_state != context->state)
	{
		if(new_context_state == IR)
		{
			change_state(context, FO);
			packet_id = PACKET_TYPE_IR;
		}
	}
	else
	{
		rohc_debugf(3, "old_context_state %d new_context_state %d\n",context->state,new_context_state);
//	switch( context->state )
		switch(new_context_state)
		{
			case IR: /* The Initialization and Refresh (IR) state */
				change_state(context, FO);
				packet_id = PACKET_TYPE_IR;
				break;
			case FO: /* The First Order (FO) state */
				change_state(context, SO);
//		          packet_id = PACKET_TYPE_IR_DYN;
//		          break;
			case SO: /* The Second Order (SO) state */
			default:
				packet_id = 0;
				break;
		}
	}


	/* encode the IP packet */

	rohc_debugf(3, "state %d\n",context->state);

	rohc_debugf(3, "msn %Xh\n",v2_context->msn);
	fflush(stdout);

	/* parts 1 and 3:
	 *  - part 2 will be placed at 'first_position'
	 *  - part 4 will start at 'counter'
	 */
	counter = code_cid_values(context->compressor->medium.cid_type, context->cid,
	                          dest, g_context->tmp.max_size,
	                          &first_position);
	rohc_debugf(3, "counter = %d first_position = %d dest[0]=%2.2Xh dest[1]=%2.2Xh\n",counter,
	            first_position,dest[0],
	            dest[1]);

	if(base_header_inner.ipvx->version == IPV4)
	{
		rohc_debugf(3, "CONTEXT %p msn %4.4Xh IP-ID %4.4Xh behavior %d\n",context,v2_context->msn,
		            ntohs(
		               base_header_inner.ipv4->ip_id),ip_inner_context.v4->ip_id_behavior);
	}

	if(packet_id == PACKET_TYPE_IR)
	{
		int udp_length;

		/* part 2: type of packet */
		dest[first_position] = packet_id;
		rohc_debugf(3, "packet type %2.2Xh\n",packet_id);

		/* part 4 */
		rohc_debugf(3, "profile ID = 0x%02x\n", profile->profile.id);
		dest[counter] = profile->profile.id;
		counter++;

		/* part 5: the CRC is computed later since it must be computed
		 * over the whole packet with an empty CRC field */
		rohc_debugf(3, "CRC = 0x00 for CRC calculation\n");
		crc_position = counter;
		dest[counter] = 0;
		counter++;

		mptr.uint8 = &dest[counter];

		/* part 6 : static chain */

		// Init pointer to the initial packet
		base_header.uint8 = (unsigned char*) ip->data;
		ip_context.uint8 = v2_context->ip_context;

		do
		{

			rohc_debugf(3, "base_header %p IP version %d\n",base_header.uint8,
			            base_header.ipvx->version);

			switch(base_header.ipvx->version)
			{
				case IPV4:
					mptr.uint8 =
					   rohc_v2_code_static_ip_part(v2_context,ip_context,mptr,base_header,packet_size,
					                               base_header.ipvx == base_header_inner.ipvx);
					/* get the transport protocol */
					protocol = base_header.ipv4->protocol;
					udp_length = ntohs(base_header.ipv4->length) - sizeof(base_header_ipv4_t);
					rohc_debugf(3, "udp_length %d %Xh\n",udp_length,udp_length);
					++base_header.ipv4;
					++ip_context.v4;
					break;
				case IPV6:
					mptr.uint8 =
					   rohc_v2_code_static_ip_part(v2_context,ip_context,mptr,base_header,packet_size,
					                               base_header.ipvx == base_header_inner.ipvx);
					udp_length = ntohs(base_header.ipv6->payload_length);
					rohc_debugf(3, "udp_length %d %Xh\n",udp_length,udp_length);
					/* get the transport protocol */
					protocol = base_header.ipv6->next_header;
					++base_header.ipv6;
					++ip_context.v6;
					while( ( ipproto_specifications[protocol] & IPV6_OPTION ) != 0)
					{
						rohc_debugf(3, "IPv6 option %d at %p\n",protocol,base_header.uint8);
						mptr.uint8 =
						   rohc_v2_code_static_ipv6_option_part(v2_context,ip_context,mptr,protocol,
						                                        base_header,
						                                        packet_size);
						protocol = base_header.ipv6_opt->next_header;
						base_header.uint8 += ip_context.v6_option->option_length;
						ip_context.uint8 += ip_context.v6_option->context_length;
					}
					break;
				default:
					return -1;
			}
			if(mptr.uint8 == NULL)
			{
				goto quit;
			}
			rohc_debugf(3, "counter %d protocol %d\n",counter,protocol);
			fflush(stdout);
		}
		while( ( ipproto_specifications[protocol] & IP_TUNNELING ) != 0);

		switch(profile_value)
		{
			case ROHCv2_PROFILE_RTP:
				mptr.uint8 = c_rohc_v2_code_udp_static_part(ip_context.udp,mptr,base_header.udp);
				++ip_context.udp;
				++base_header.udp;
				mptr.uint8 = c_rohc_v2_code_rtp_static_part(ip_context.rtp,mptr,base_header.rtp);
				break;
			case ROHCv2_PROFILE_UDP:
				mptr.uint8 = c_rohc_v2_code_udp_static_part(ip_context.udp,mptr,base_header.udp);
				break;
			case ROHCv2_PROFILE_ESP:
				mptr.uint8 = c_rohc_v2_code_esp_static_part(ip_context.esp,mptr,base_header.esp);
				break;
			case ROHCv2_PROFILE_IP:
				break;
			case ROHCv2_PROFILE_UDPLITE_RTP:
				mptr.uint8 = c_rohc_v2_code_udp_lite_static_part(ip_context.udp_lite,mptr,
				                                                 base_header.udp_lite);
				++ip_context.udp_lite;
				++base_header.udp_lite;
				mptr.uint8 = c_rohc_v2_code_rtp_static_part(ip_context.rtp,mptr,base_header.rtp);
				break;
			case ROHCv2_PROFILE_UDPLITE:
				mptr.uint8 = c_rohc_v2_code_udp_lite_static_part(ip_context.udp_lite,mptr,
				                                                 base_header.udp_lite);
				break;
		}

		/* Part 7 : add dynamic chain */

		// Reinit pointer to the initial packet
		base_header.ipvx = (base_header_ipvx_t *)ip->data;
		ip_context.uint8 = v2_context->ip_context;

		do
		{

			rohc_debugf(3, "base_header %p IP version %d\n",base_header.uint8,
			            base_header.ipvx->version);

			switch(base_header.ipvx->version)
			{
				case IPV4:
					mptr.uint8 =
					   rohc_v2_code_dynamic_ip_part(v2_context,ip_inner_context,ip_context,mptr,
					                                base_header,
					                                packet_size,
					                                profile->profile.id);
					/* get the transport protocol */
					protocol = base_header.ipv4->protocol;
					++base_header.ipv4;
					++ip_context.v4;
					break;
				case IPV6:
					mptr.uint8 =
					   rohc_v2_code_dynamic_ip_part(v2_context,ip_inner_context,ip_context,mptr,
					                                base_header,
					                                packet_size,
					                                profile->profile.id);
					protocol = base_header.ipv6->next_header;
					++base_header.ipv6;
					++ip_context.v6;
					fflush(stdout);
					while( ( ipproto_specifications[protocol] & IPV6_OPTION ) != 0)
					{
						rohc_debugf(3, "IPv6 option %d at %p\n",protocol,base_header.uint8);
						mptr.uint8 =
						   rohc_v2_code_dynamic_ipv6_option_part(v2_context,ip_context,mptr,protocol,
						                                         base_header,
						                                         packet_size);
						protocol = base_header.ipv6_opt->next_header;
						base_header.uint8 += ip_context.v6_option->option_length;
						ip_context.uint8 += ip_context.v6_option->context_length;
					}
					break;
				default:
					return -1;
			}
			if(mptr.uint8 == NULL)
			{
				goto quit;
			}
			rohc_debugf(3, "counter %d protocol %d\n",counter,protocol);
			fflush(stdout);
		}
		while( ( ipproto_specifications[protocol] & IP_TUNNELING ) != 0);

		switch(profile_value)
		{
			case ROHCv2_PROFILE_RTP:
				mptr.uint8 =
				   c_rohc_v2_code_udp_dynamic_part(v2_context,profile->profile.id,ip_context.udp,mptr,
				                                   base_header.udp);
				++base_header.udp;
				++ip_context.udp;
				mptr.uint8 = c_rohc_v2_code_rtp_dynamic_part(v2_context,ip_context.rtp,mptr,
				                                             base_header.rtp_with_csrc);
				++base_header.rtp;
				break;
			case ROHCv2_PROFILE_UDP:
				mptr.uint8 =
				   c_rohc_v2_code_udp_dynamic_part(v2_context,profile->profile.id,ip_context.udp,mptr,
				                                   base_header.udp);
				++base_header.udp;
				break;
			case ROHCv2_PROFILE_ESP:
				mptr.uint8 = c_rohc_v2_code_esp_dynamic_part(v2_context,ip_context.esp,mptr,
				                                             base_header.esp);
				++base_header.esp;
				break;
			case ROHCv2_PROFILE_IP:
				break;
			case ROHCv2_PROFILE_UDPLITE_RTP:
				mptr.uint8 =
				   c_rohc_v2_code_udp_lite_dynamic_part(v2_context,ip_context.udp_lite,mptr,
				                                        base_header.udp_lite,
				                                        profile->profile.id);
				++base_header.udp_lite;
				++ip_context.udp_lite;
				mptr.uint8 = c_rohc_v2_code_rtp_dynamic_part(v2_context,ip_context.rtp,mptr,
				                                             base_header.rtp_with_csrc);
				++base_header.rtp;
				break;
			case ROHCv2_PROFILE_UDPLITE:
				mptr.uint8 =
				   c_rohc_v2_code_udp_lite_dynamic_part(v2_context,ip_context.udp_lite,mptr,
				                                        base_header.udp_lite,
				                                        profile->profile.id);
				++base_header.udp_lite;
				break;
		}
		fflush(stdout);
		counter = mptr.uint8 - dest;

		/* part 5 */
		dest[crc_position] = crc_calculate(CRC_TYPE_8, dest, counter, CRC_INIT_8,
		                                   context->compressor->crc_table_8);
		rohc_debugf(3, "CRC (header length = %d, crc = 0x%x)\n", counter, dest[crc_position]);

		*packet_type = PACKET_IR;
		fflush(stdout);
	}
	else
	{
		u_int8_t ctrl_data_value[11 + MAX_IP_HEADERS];
		multi_ptr_t ctrl_data_ptr;
		u_int8_t save_first_byte;
		int outer_ip_flag = 0;
		multi_ptr_t pheader;

		// If SMALL_CID
		// If CID = 0         counter = 1   first_position = 0  no ADD-CID
		// If CID = 1-15      counter = 2   first_position = 1  0xEx
		// else
		//               1 <= counter <= 5  first_position = 0

		/* save the last CID octet */
		save_first_byte = dest[counter - 1];

		pheader.uint8 =
		   mptr.uint8 = &dest[counter - 1];

		// If outer IP header
		if(base_header_inner.uint8 != ip->data)
		{
			// Reinit pointer to the initial packet
			base_header.uint8 = (unsigned char *) ip->data;
			ip_context.uint8 = v2_context->ip_context;

			do
			{

				rohc_debugf(3, "base_header %p IP version %d\n",base_header.uint8,
				            base_header.ipvx->version);

				switch(base_header.ipvx->version)
				{
					case IPV4:
						// See RFC5225 page 39
						if(base_header.ipv4->tos_tc != ip_context.v4->tos_tc ||
						   base_header.ipv4->ttl_hopl != ip_context.v4->ttl_hopl)
						{
							rohc_debugf(3, "IPv4 tos_tc %Xh %Xh or ttl_hopl %Xh %Xh changed\n",
							            base_header.ipv4->tos_tc,ip_context.v4->tos_tc,
							            base_header.ipv4->ttl_hopl,ip_context.v4->ttl_hopl);
							++outer_ip_flag;
						}
						else
						{
							rohc_debugf(3, "IPv4 tos_tc %Xh %Xh and ttl_hopl %Xh %Xh same\n",
							            base_header.ipv4->tos_tc,ip_context.v4->tos_tc,
							            base_header.ipv4->ttl_hopl,ip_context.v4->ttl_hopl);
						}
						++base_header.ipv4;
						++ip_context.v4;
						break;
					case IPV6:
						// See RFC5225 page 39
						#if __BYTE_ORDER == __LITTLE_ENDIAN
						if(base_header.ipv6->tos_tc1 != ip_context.v6->tos_tc1 ||
						   base_header.ipv6->tos_tc2 != ip_context.v6->tos_tc2 ||
						#elif __BYTE_ORDER == __BIG_ENDIAN
						if(base_header.ipv6->tos_tc != ip_context.v6->tos_tc ||
						#endif
						   base_header.ipv6->ttl_hopl != ip_context.v6->ttl_hopl)
						{
							rohc_debugf(3, "IPv6 tos_tc %Xh %Xh or ttl_hopl %Xh %Xh changed\n",
											#if __BYTE_ORDER == __LITTLE_ENDIAN
							            (base_header.ipv6->tos_tc1 << 4) | base_header.ipv6->tos_tc2,
							            (ip_context.v6->tos_tc1 << 4) | ip_context.v6->tos_tc2,
											#elif __BYTE_ORDER == __BIG_ENDIAN
							            base_header.ipv6->tos_tc,ip_context.v6->tos_tc,
											#endif
							            base_header.ipv6->ttl_hopl,ip_context.v6->ttl_hopl);
							++outer_ip_flag;
						}
						/* get the transport protocol */
						protocol = base_header.ipv6->next_header;
						++base_header.ipv6;
						++ip_context.v6;
						fflush(stdout);
						while( ( ipproto_specifications[protocol] & IPV6_OPTION ) != 0)
						{
							rohc_debugf(3, "IPv6 option %d at %p\n",protocol,base_header.uint8);
							protocol = base_header.ipv6_opt->next_header;
							base_header.uint8 += ip_context.v6_option->option_length;
							ip_context.uint8 += ip_context.v6_option->context_length;
						}
						break;
					default:
						return -1;
				}
				fflush(stdout);
			}
			while(base_header.ipvx != base_header_inner.ipvx);

			switch(profile_value)
			{
				case ROHCv2_PROFILE_RTP:
					++base_header.udp;
					++ip_context.udp;
					break;
				case ROHCv2_PROFILE_UDPLITE_RTP:
					++base_header.udp_lite;
					++ip_context.udp_lite;
					break;
			}
		}

		rohc_debugf(3, "profile_value %Xh outer_ip_flag %d\n",profile_value,outer_ip_flag);
		fflush(stdout);

		// Init control datas for control_crc3_encoding (see 6.6.11 page 32)
		ctrl_data_ptr.uint8 = ctrl_data_value;

		// Init control datas for control_crc3_encoding (see 6.6.11 page 32)
		*(ctrl_data_ptr.uint8++) = v2_context->reorder_ratio;
		rohc_debugf(3, "Add reorder_ratio %d to control_data_crc\n",v2_context->reorder_ratio);

		switch(profile_value)
		{
			case ROHCv2_PROFILE_RTP:
				rohc_debugf(3, "RTP context %p ts_stride %d time_stride %d\n",ip_context.rtp,
				            ip_context.rtp->ts_stride,
				            ip_context.rtp->time_stride);
				WRITE32_TO_MPTR(ctrl_data_ptr,htonl(ip_context.rtp->ts_stride));
				rohc_debugf(3, "Add ts_stride %d to control_data_crc\n",ip_context.rtp->ts_stride);
				WRITE32_TO_MPTR(ctrl_data_ptr,htonl(ip_context.rtp->time_stride));
				rohc_debugf(3, "Add time_stride %d to control_data_crc\n",ip_context.rtp->time_stride);
				mptr.uint8 =
				   c_rtp_baseheader(context,v2_context,ip_inner_context,ip_context.rtp,mptr,
				                    base_header_inner,
				                    base_header.rtp_with_csrc,profile_value,
				                    outer_ip_flag);
				break;
			case ROHCv2_PROFILE_UDP:
				WRITE16_TO_MPTR(ctrl_data_ptr,htons(v2_context->msn));
				rohc_debugf(3, "Add msn %Xh to control_data_crc\n",htons(v2_context->msn));
				mptr.uint8 =
				   c_udp_baseheader(context,v2_context,ip_inner_context,mptr,base_header_inner,
				                    profile_value,
				                    outer_ip_flag);
				break;
			case ROHCv2_PROFILE_ESP:
				mptr.uint8 =
				   c_esp_baseheader(context,v2_context,ip_inner_context,ip_context.esp,mptr,
				                    base_header_inner,
				                    base_header.esp,profile_value,
				                    outer_ip_flag);
				break;
			case ROHCv2_PROFILE_IP:
				WRITE16_TO_MPTR(ctrl_data_ptr,htons(v2_context->msn));
				rohc_debugf(3, "Add msn %Xh to control_data_crc\n",v2_context->msn);
				mptr.uint8 = c_ip_baseheader(context,v2_context,ip_inner_context,mptr,base_header_inner,
				                             outer_ip_flag);
				break;
			case ROHCv2_PROFILE_UDPLITE_RTP:
				*(ctrl_data_ptr.uint8++) = ip_context.udp_lite->coverage_behavior;
				rohc_debugf(3, "Add coverage_behavior %d to control_data_crc\n",
				            ip_context.udp_lite->coverage_behavior);
				WRITE32_TO_MPTR(ctrl_data_ptr,htonl(ip_context.rtp->ts_stride));
				rohc_debugf(3, "Add ts_stride %d to control_data_crc\n",ip_context.rtp->ts_stride);
				WRITE32_TO_MPTR(ctrl_data_ptr,htonl(ip_context.rtp->time_stride));
				rohc_debugf(3, "Add time_stride %d to control_data_crc\n",ip_context.rtp->time_stride);
				break;
			case ROHCv2_PROFILE_UDPLITE:
				WRITE16_TO_MPTR(ctrl_data_ptr,htons(v2_context->msn));
				rohc_debugf(3, "Add msn %Xh to control_data_crc\n",v2_context->msn);
				*(ctrl_data_ptr.uint8++) = ip_context.udp_lite->coverage_behavior;
				rohc_debugf(3, "Add coverage_behavior %d to control_data_crc\n",
				            ip_context.udp_lite->coverage_behavior);
				mptr.uint8 =
				   c_udp_lite_baseheader(context,v2_context,ip_inner_context,ip_context.udp_lite,mptr,
				                         base_header_inner,profile_value,
				                         outer_ip_flag);
				//      c_udp_lite_baseheader(context,v2_context,ip_inner_context,ip_context.udp_lite,mptr,base_header_inner,base_header.udp_lite,profile_value,outer_ip_flag);
				break;
		}

		// If co_common compressed format used ( discriminator =:= '11111010' )
		if(pheader.xxx_co_common->discriminator == 0xFA)
		{
			// Add all ip_id_behavior of headers
			memcpy(ctrl_data_ptr.uint8,ip_id_behaviors,ip_id_behaviors_count);
			ctrl_data_ptr.uint8 += ip_id_behaviors_count;
			// Calculate CRC over a number of control fields
			pheader.xxx_co_common->control_crc3 =
			   crc_calculate(CRC_TYPE_3,ctrl_data_value,ctrl_data_ptr.uint8 - ctrl_data_value,
			                 CRC_INIT_3,
			                 context->compressor->crc_table_3);
			rohc_debugf(3, "control_crc3 %d for size %d\n",pheader.xxx_co_common->control_crc3,
			            (int)(ctrl_data_ptr.uint8 - ctrl_data_value));
			TraceData(ctrl_data_value,ctrl_data_ptr.uint8 - ctrl_data_value);
			// Calculate CRC over the header
			// =:= crc7(THIS.UVALUE,THIS.ULENGTH)
			pheader.xxx_co_common->header_crc = 0;
			pheader.xxx_co_common->header_crc =
			   crc_calculate(CRC_TYPE_7,pheader.uint8,mptr.uint8 - pheader.uint8,CRC_INIT_7,
			                 context->compressor->crc_table_7);
			rohc_debugf(3, "header_crc %2.2Xh length %d\n",pheader.xxx_co_common->header_crc,
			            (int)(mptr.uint8 - pheader.uint8));
		}

		// Now add irregular chain

		// Reinit pointer to the initial packet
		base_header.ipvx = (base_header_ipvx_t *)ip->data;
		ip_context.uint8 = v2_context->ip_context;

		do
		{

			rohc_debugf(3, "base_header %p IP version %d\n",base_header.uint8,
			            base_header.ipvx->version);

			mptr.uint8 =
			   rohc_v2_code_irregular_ip_part(ip_context,mptr,base_header,packet_size,
			                                  base_header.ipvx ==
			                                  base_header_inner.ipvx,
			                                  outer_ip_flag);

			switch(base_header.ipvx->version)
			{
				case IPV4:
					/* get the transport protocol */
					protocol = base_header.ipv4->protocol;
					++base_header.ipv4;
					++ip_context.v4;
					break;
				case IPV6:
					protocol = base_header.ipv6->next_header;
					++base_header.ipv6;
					++ip_context.v6;
					fflush(stdout);
					while( ( ipproto_specifications[protocol] & IPV6_OPTION ) != 0)
					{
						rohc_debugf(3, "IPv6 option %d at %p\n",protocol,base_header.uint8);
						mptr.uint8 =
						   rohc_v2_code_irregular_ipv6_option_part(v2_context,ip_context,mptr,protocol,
						                                           base_header,
						                                           packet_size);
						protocol = base_header.ipv6_opt->next_header;
						base_header.uint8 += ip_context.v6_option->option_length;
						ip_context.uint8 += ip_context.v6_option->context_length;
					}
					break;
				default:
					return -1;
			}
			if(mptr.uint8 == NULL)
			{
				goto quit;
			}
			rohc_debugf(3, "counter %d protocol %d\n",counter,protocol);
			fflush(stdout);
		}
		while( ( ipproto_specifications[protocol] & IP_TUNNELING ) != 0);

		switch(profile_value)
		{
			case ROHCv2_PROFILE_RTP:
				mptr.uint8 = c_rohc_v2_code_udp_irregular_part(ip_context.udp,mptr,base_header.udp);
				++base_header.udp;
				++ip_context.udp;
				++base_header.rtp;
				break;
			case ROHCv2_PROFILE_UDP:
				mptr.uint8 = c_rohc_v2_code_udp_irregular_part(ip_context.udp,mptr,base_header.udp);
				++base_header.udp;
				break;
			case ROHCv2_PROFILE_ESP:
				//  mptr.uint8 = c_rohc_v2_code_esp_irregular_part(v2_context,ip_context.esp,mptr,base_header.esp);
				++base_header.esp;
				break;
			case ROHCv2_PROFILE_IP:
				break;
			case ROHCv2_PROFILE_UDPLITE_RTP:
				mptr.uint8 = c_rohc_v2_code_udp_lite_irregular_part(ip_context.udp_lite,mptr,
				                                                    base_header.udp_lite);
				++base_header.udp_lite;
				++ip_context.udp_lite;
				++base_header.rtp;
				break;
			case ROHCv2_PROFILE_UDPLITE:
				mptr.uint8 = c_rohc_v2_code_udp_lite_irregular_part(ip_context.udp_lite,mptr,
				                                                    base_header.udp_lite);
				++base_header.udp_lite;
				break;
		}
		fflush(stdout);

		if(context->compressor->medium.cid_type != ROHC_SMALL_CID)
		{
			rohc_debugf(3, "counter %d dest[counter-1]=%2.2Xh save_first_byte=%2.2Xh\n",counter,
			            dest[counter - 1],save_first_byte);
			// Restore last CID octet saved
			dest[first_position] = dest[counter - 1];
			dest[counter - 1] = save_first_byte;
		}

		counter = mptr.uint8 - dest;

		*packet_type = PACKET_IR;
	}

	*payload_offset = base_header.uint8 - (u_int8_t*) ip->data;

	rohc_debugf(3, "payload_offset %d\n",*payload_offset);

	++v2_context->msn;

	rohc_debugf(3, "Compressed packet length %d\n",counter);
	TraceData(dest,counter);
	fflush(stdout);
	assert( *payload_offset <= packet_size );
quit:
	return counter;
}


// Page 63

u_int8_t * c_rohc_v2_code_udp_static_part( udp_context_t *udp_context, multi_ptr_t mptr,
                                           base_header_udp_t *udp )
{
	// Copy static part
	mptr.udp_static->src_port = udp->src_port;
	mptr.udp_static->dst_port = udp->dst_port;

	#if ROHCV2_DEBUG
	rohc_debugf(3, "UDP static part length %d\n",(int)sizeof(udp_static_t));
	TraceData(mptr.uint8,(int)sizeof(udp_static_t));
	#endif

	++mptr.udp_static;

	return mptr.uint8;
}


u_int8_t * c_rohc_v2_code_udp_dynamic_part( struct sc_v2_context *v2_context, int profile_value,
                                            udp_context_t *udp_context, multi_ptr_t mptr,
                                            base_header_udp_t *udp )
{
	int size;

	rohc_debugf(3, "mptr %p profile %Xh udp_context %p base_header_udp %p\n",mptr.uint8,
	            profile_value,udp_context,
	            udp);

	assert( udp_context->checksum_used == ( udp->checksum != 0 ) );

	if(profile_value == ROHCv2_PROFILE_UDP)
	{
		mptr.udp_endpoint_dynamic->checksum = udp->checksum;
		mptr.udp_endpoint_dynamic->msn = htons(v2_context->msn);
		mptr.udp_endpoint_dynamic->reserved = 0;
		mptr.udp_endpoint_dynamic->reorder_ratio = v2_context->reorder_ratio;
		size = sizeof(udp_endpoint_dynamic_t);
	}
	else
	{
		mptr.udp_regular_dynamic->checksum = udp->checksum;
		size = sizeof(udp_regular_dynamic_t);
	}

	#if ROHCV2_DEBUG
	rohc_debugf(3, "UDP dynamic part length %d\n",size);
	TraceData(mptr.uint8,size);
	#endif

	return mptr.uint8 + size;
}


u_int8_t * c_rohc_v2_code_udp_regular_dynamic_part( udp_context_t *udp_context, multi_ptr_t mptr,
                                                    base_header_udp_t *udp )
{
	if(udp_context->checksum_used != 0)
	{
		assert( udp->checksum != 0 );
		WRITE16_TO_MPTR(mptr,udp->checksum);
	}
	else
	{
		assert( udp->checksum == 0 );
	}
	return mptr.uint8;
}


// See RFC5225 page 63

u_int8_t * c_rohc_v2_code_udp_irregular_part( udp_context_t *udp_context, multi_ptr_t mptr,
                                              base_header_udp_t *udp )
{
	if(udp_context->checksum_used != 0)
	{
		WRITE16_TO_MPTR(mptr,udp->checksum);
		#if ROHCV2_DEBUG
		rohc_debugf(3, "UDP irregular part length 2\n");
		TraceData(mptr.uint8 - 2,2);
		#endif
	}
	#if ROHCV2_DEBUG
	else
	{
		rohc_debugf(3, "UDP irregular part length 0\n");
	}
	#endif

	return mptr.uint8;
}


// Page 64

u_int8_t * c_rohc_v2_code_rtp_static_part(rtp_context_t *rtp_context,
                                          multi_ptr_t mptr,
                                          base_header_rtp_t *rtp)
{
	mptr.rtp_static->ssrc = rtp->ssrc;

	#if ROHCV2_DEBUG
	rohc_debugf(3, "RTP static part length %d\n",(int)sizeof(rtp_static_t));
	TraceData(mptr.uint8,(int)sizeof(rtp_static_t));
	#endif

	return mptr.uint8 + sizeof(rtp_static_t);
}


u_int8_t * c_csrc_list_dynchain(rtp_context_t *rtp_context,multi_ptr_t mptr,
                                base_header_rtp_with_csrc_t *rtp)
{
	#if ROHCV2_DEBUG
	u_int8_t *ptr = mptr.uint8;
	#endif
	multi_ptr_t pNewItem;
	u_int8_t j;
	int i;

	// 8-bit XI fields, cc items
	*(mptr.uint8++) = 0x10 | rtp->cc;

	// Init pointer to first storage for new items
	pNewItem.uint8 = mptr.uint8 + rtp->cc;

	// For each RTP CSRC identifiers of the header
	for(i = 0; i < rtp->cc; ++i)
	{
		// Look if present in context list
		for(j = 0; j < rtp_context->max_csrc_index; ++j)
		{
			// If found in the list
			if(rtp->csrc_list[i] == rtp_context->csrc_list[j])
			{
				// Value already present in list of context, at index j
				rohc_debugf(3, "RTP csrc[%d] = %Xh present at index %d\n",i,rtp->csrc_list[i],j);
				*(mptr.uint8)++ = j;
				break;
			}
		}
		// If value not found
		if(j == rtp_context->max_csrc_index)
		{
			if(j > 15)
			{
				rohc_debugf(3, "RTP csrc[%d] = %Xh can't be added in the context list at index %d!\n",i,
				            rtp->csrc_list[i],
				            j);
				continue;
			}
			// Value not present in list of context
			rohc_debugf(3, "RTP csrc[%d] = %Xh added at index %d\n",i,rtp->csrc_list[i],j);
			// Save it on context
			rtp_context->csrc_list[j] = rtp->csrc_list[i];
			// Item present in list for index j
			*(mptr.uint8)++ = 0x80 | j;
			// Add item in the list
			WRITE32_TO_MPTR(pNewItem,rtp->csrc_list[i]);
			// One more CSRC value in the context list
			++rtp_context->max_csrc_index;
		}
	}

	#if ROHCV2_DEBUG
	rohc_debugf(3, "RTP CSRV list, size %d\n",(int)(pNewItem.uint8 - ptr));
	TraceData(ptr,(int)(pNewItem.uint8 - ptr));
	#endif

	return pNewItem.uint8;
}


u_int8_t * c_rohc_v2_code_rtp_dynamic_part(struct sc_v2_context *v2_context,
                                           rtp_context_t *rtp_context,
                                           multi_ptr_t c_base_header,
                                           base_header_rtp_with_csrc_t *rtp)
{
	multi_ptr_t mptr;
	int size;

	rohc_debugf(3, "v2_context %p rtp_context %p c_base_header %p rtp %p\n",v2_context,rtp_context,
	            c_base_header.uint8,
	            rtp);

	mptr.rtp_dynamic = c_base_header.rtp_dynamic + 1;

	c_base_header.rtp_dynamic->reserved = 0;
	c_base_header.rtp_dynamic->reorder_ratio = v2_context->reorder_ratio;
	c_base_header.rtp_dynamic->pad_bit = rtp->pad_bit;
	c_base_header.rtp_dynamic->extension = rtp->extension;
	c_base_header.rtp_dynamic->marker = rtp->marker;
	c_base_header.rtp_dynamic->payload_type = rtp->payload_type;
	c_base_header.rtp_dynamic->sequence_number = rtp->sequence_number;
	c_base_header.rtp_dynamic->timestamp = rtp->timestamp;
	// =:= sdvl_or_default(tss_indicator.CVALUE, TS_STRIDE_DEFAULT)
	c_base_header.rtp_dynamic->tss_indicator = c_sdvl_or_default(&mptr,0,TS_STRIDE_DEFAULT);
	// =:= sdvl_or_default(tis_indicator.CVALUE, TIME_STRIDE_DEFAULT)
	c_base_header.rtp_dynamic->tis_indicator = c_sdvl_or_default(&mptr,0,TIME_STRIDE_DEFAULT);

	rtp_context->marker = rtp->marker;
	rtp_context->pad_bit = rtp->pad_bit;
	rtp_context->extension = rtp->extension;
	rtp_context->payload_type = rtp->payload_type;
//	rtp_context->sequence_number = rtp->sequence_number;
	rtp_context->last_timestamp = ntohl(rtp->timestamp);

	// =:= csrc_list_dynchain(list_present.CVALUE, cc.UVALUE)
	if(rtp->cc == 0)
	{
		c_base_header.rtp_dynamic->list_present = 0;
	}
	else
	{
		c_base_header.rtp_dynamic->list_present = 1;
		mptr.uint8 = c_csrc_list_dynchain(rtp_context,mptr,rtp);
	}

	size = mptr.uint8 - c_base_header.uint8;

	#if ROHCV2_DEBUG
	rohc_debugf(3, "RTP dynamic part length %d\n",size);
	TraceData(c_base_header.uint8,size);
	#endif

	return mptr.uint8;
}


// Page 67

u_int8_t * c_rohc_v2_code_udp_lite_static_part( udp_lite_context_t *udp_lite_context,
                                                multi_ptr_t mptr,
                                                base_header_udp_lite_t *udp_lite )
{
	// Copy static part
	mptr.udp_lite_static->src_port = udp_lite->src_port;
	mptr.udp_lite_static->dst_port = udp_lite->dst_port;

	#if ROHCV2_DEBUG
	rohc_debugf(3, "UDP-Lite static part length %d\n",(int)sizeof(udp_lite_static_t));
	TraceData(mptr.uint8,(int)sizeof(udp_lite_static_t));
	#endif

	++mptr.udp_lite_static;

	return mptr.uint8;
}


/*

  UDP-Lite Coverage Behavior

       The control field coverage_behavior specifies how the checksum
       coverage field of the UDP-Lite header is compressed with RoHCv2.
       It can indicate one of the following encoding methods: irregular,
       static, or inferred encoding.

       o  coverage_behavior_value:

          Set to a 2-bit integer value, using one of the constants whose
          name begins with the prefix UDP_LITE_COVERAGE_ and as defined
          in Section 6.8.2.4 :

          UDP_LITE_COVERAGE_INFERRED  = 0;
          UDP_LITE_COVERAGE_STATIC    = 1;
          UDP_LITE_COVERAGE_IRREGULAR = 2;

  UDP-Lite Header Fields

   Checksum Coverage

      The Checksum Coverage field may behave in different ways: it may
      have a value of zero, it may be equal to the datagram length, or
      it may have any value between eight octets and the length of the
      datagram.  From a compression perspective, this field is expected
      to either be entirely predictable (for the cases where it follows
      the same behavior as the UDP Length field or where it takes on a
      constant value) or to change randomly for each packet (making the
      value unpredictable from a header-compression perspective).  For
      all cases, the behavior itself is not expected to change for this
      field during the lifetime of a packet flow, or to change
      relatively seldom.

   Checksum

      The information used for the calculation of the UDP-Lite checksum
      is governed by the value of the checksum coverage and minimally
      includes the UDP-Lite header.  The checksum is a changing field
      that must always be sent as-is.

*/

// See RFC 5225 page 66

u_int16_t c_checksum_coverage_dynchain( udp_lite_context_t *udp_lite_context, u_int16_t value )
{
	switch(udp_lite_context->coverage_behavior)
	{
		case UDP_LITE_COVERAGE_INFERRED:
		// Ne doit pas se produire, ou erreur dans appel page 68 en haut
// A REVOIR			                           assert( 1 == 0 );
//			                           break;
		case UDP_LITE_COVERAGE_STATIC:
		case UDP_LITE_COVERAGE_IRREGULAR:
			return value;
	}
	return 0;
}


// See RFC 5225 page 66/67

u_int16_t c_checksum_coverage_irregular( udp_lite_context_t *udp_lite_context, u_int16_t value )
{
	switch(udp_lite_context->coverage_behavior)
	{
		case UDP_LITE_COVERAGE_INFERRED:  // =:= inferred_coverage
			return value;
		case UDP_LITE_COVERAGE_STATIC:  // =:= static_coverage
			return value;
		case UDP_LITE_COVERAGE_IRREGULAR:  // =:= irregular_coverage
			return value;
	}
	return 0;
}


// Page 67/68

u_int8_t * c_rohc_v2_code_udp_lite_dynamic_part(struct sc_v2_context *v2_context,
                                                udp_lite_context_t *udp_lite_context,
                                                multi_ptr_t mptr,
                                                base_header_udp_lite_t *udp_lite,
                                                int profile_value)
{
	int size;

	rohc_debugf(3, "v2_context %p udp_lite_context %p mptr %p udp_lite %p profile %Xh\n",
	            v2_context,udp_lite_context,mptr.uint8,udp_lite,profile_value);

	if(profile_value == PROFILE_UDPLITE_0108)
	{
		mptr.udp_lite_endpoint_dynamic1->reserved = 0;
		mptr.udp_lite_endpoint_dynamic1->reorder_ratio = v2_context->reorder_ratio;
		if( ( mptr.udp_lite_endpoint_dynamic1->coverage_behavior =
		         udp_lite_context->coverage_behavior ) == UDP_LITE_COVERAGE_INFERRED)
		{
			mptr.udp_lite_endpoint_dynamic1->checksum = udp_lite->checksum;
			mptr.udp_lite_endpoint_dynamic1->msn = htons(v2_context->msn);
			size = sizeof(udp_lite_endpoint_dynamic1_t);
		}
		else
		{
			mptr.udp_lite_endpoint_dynamic2->checksum_coverage = udp_lite->checksum_coverage;
			mptr.udp_lite_endpoint_dynamic2->checksum = udp_lite->checksum;
			mptr.udp_lite_endpoint_dynamic2->msn = htons(v2_context->msn);
			size = sizeof(udp_lite_endpoint_dynamic2_t);
		}
	}
	else
	{
		assert( profile_value == PROFILE_RTP_0107 );
		mptr.udp_lite_regular_dynamic1->reserved = 0;
		if( ( mptr.udp_lite_regular_dynamic1->coverage_behavior =
		         udp_lite_context->coverage_behavior ) == UDP_LITE_COVERAGE_INFERRED)
		{
			mptr.udp_lite_regular_dynamic1->checksum = udp_lite->checksum;
			size = sizeof(udp_lite_regular_dynamic1_t);
		}
		else
		{
			mptr.udp_lite_regular_dynamic2->checksum_coverage = udp_lite->checksum_coverage;
			mptr.udp_lite_regular_dynamic2->checksum = udp_lite->checksum;
			size = sizeof(udp_lite_regular_dynamic2_t);
		}
	}

	udp_lite_context->last_coverage_behavior = udp_lite_context->coverage_behavior;
	udp_lite_context->last_checksum_coverage = udp_lite->checksum_coverage;

	#if ROHCV2_DEBUG
	rohc_debugf(3, "UDP-Lite dynamic part length %d\n",size);
	TraceData(mptr.uint8,size);
	#endif

	return mptr.uint8 + size;
}


// Page 68

u_int8_t * c_rohc_v2_code_udp_lite_irregular_part( udp_lite_context_t *udp_lite_context,
                                                   multi_ptr_t mptr,
                                                   base_header_udp_lite_t *udp_lite )
{
	int size;

	if(udp_lite_context->coverage_behavior == UDP_LITE_COVERAGE_IRREGULAR)
	{
		mptr.udp_lite_irregular2->checksum_coverage = udp_lite->checksum_coverage;
		mptr.udp_lite_irregular2->checksum = udp_lite->checksum;
		size = sizeof(udp_lite_irregular2_t);
	}
	else
	{
		mptr.udp_lite_irregular1->checksum = udp_lite->checksum;
		size = sizeof(udp_lite_irregular1_t);
	}

	#if ROHCV2_DEBUG
	rohc_debugf(3, "UDP-Lite irregular part length %d\n",size);
	TraceData(mptr.uint8,size);
	#endif

	return mptr.uint8 + size;
}


/**
 * @brief Compress the innermost IP header AND the UDP-Lite header
 *
 * See RFC5225 page 97
 *
 * @param context                   The compression context
 * @param v2_context                The specific RoHCv2 context
 * @param ip_inner_context          The specific IP innermost context
 * @param udp_lite_context          The specific UDP-Lite context
 * @param c_base_header             The compressed header
 * @param base_header_inner         The innermost IP header
 * @param profile_value             The profile value
 * @param outer_ip_flag             The IP flag of the outer IP header
 * @return                          The new pointer in the rohc-packet-under-build buffer
 */

u_int8_t * c_udp_lite_baseheader(struct c_context *const context,
                                 struct sc_v2_context *v2_context,
                                 ip_context_ptr_t ip_inner_context,
                                 udp_lite_context_t *udp_lite_context,
                                 multi_ptr_t c_base_header,
                                 base_header_ip_t base_header_inner,
                                 int profile_value, int outer_ip_flag )
{
	multi_ptr_t mptr;
	u_int16_t old_msn;
	WB_t ip_id;
	#if __BYTE_ORDER == __LITTLE_ENDIAN
	u_int8_t msn;
	u_int8_t ipid;
	#endif
	int size;
	int udp_length;

	rohc_debugf(
	   3,
	   "context %p v2_context %p ip_context %p c_base_header %p base_header_inner %p profile_value %Xh outer_ip_flag %d\n",
	   context,v2_context,ip_inner_context.uint8,c_base_header.uint8,
	   base_header_inner.uint8,profile_value,outer_ip_flag);

	if(base_header_inner.ipvx->version == IPV4)
	{
		udp_length = ntohs(base_header_inner.ipv4->length) - sizeof(base_header_ipv4_t);
		rohc_debugf(3, "udp_length %d %Xh\n",udp_length,udp_length);
		ip_id.uint16 = ntohs( base_header_inner.ipv4->ip_id );
		if(outer_ip_flag != 0)
		{
			goto code_udp_lite_co_common;
		}
		if(ip_inner_context.v4->df != base_header_inner.ipv4->df ||
		   ip_inner_context.v4->tos_tc != base_header_inner.ipv4->tos_tc ||
		   ip_inner_context.v4->ttl_hopl != base_header_inner.ipv4->ttl_hopl)
		{
			goto code_udp_lite_co_common;
		}
		if(ip_inner_context.v4->ip_id_behavior != ip_inner_context.v4->last_ip_id_behavior)
		{
			goto code_udp_lite_co_common;
		}
		switch(ip_inner_context.v4->ip_id_behavior)
		{
			case IP_ID_BEHAVIOR_SEQUENTIAL:
				// If only 6 bits change
				if( ( ip_inner_context.v4->last_ip_id.uint16 & 0xFFC0 ) == ( ip_id.uint16 & 0xFFC0 ) )
				{
					goto code_udp_lite_pt_1_seq_id;
				}
				break;
				// If only 8 bits change
				if( ( ip_inner_context.v4->last_ip_id.uint16 & 0xFF00 ) == ( ip_id.uint16 & 0xFF00 ) )
				{
					goto code_udp_lite_pt_2_seq_id;
				}
				break;
			case IP_ID_BEHAVIOR_SEQUENTIAL_SWAPPED:
				// If only 6 bits change
				if( ( ip_inner_context.v4->last_ip_id.uint16 & 0xC0FF ) == ( ip_id.uint16 & 0xC0FF ) )
				{
					goto code_udp_lite_pt_1_seq_id;
				}
				break;
				// If only 8 bits change
				if( ( ip_inner_context.v4->last_ip_id.uint16 & 0x00FF ) == ( ip_id.uint16 & 0x00FF ) )
				{
					goto code_udp_lite_pt_2_seq_id;
				}
				break;
			default:
				goto code_udp_lite_co_common;
		}
	}
	else
	{
		udp_length = ntohs(base_header_inner.ipv6->payload_length);
		rohc_debugf(3, "udp_length %d %Xh\n",udp_length,udp_length);
		if(outer_ip_flag != 0)
		{
			goto code_udp_lite_co_common;
		}
		#if __BYTE_ORDER == __LITTLE_ENDIAN
		if(ip_inner_context.v6->tos_tc1 != base_header_inner.ipv6->tos_tc1 ||
		   ip_inner_context.v6->tos_tc2 != base_header_inner.ipv6->tos_tc2)
		#elif __BYTE_ORDER == __BIG_ENDIAN
		if(ip_inner_context.v6->tos_tc != base_header_inner.ipv6->tos_tc)
		#endif
		{
			goto code_udp_lite_co_common;
		}
		if(ip_inner_context.v6->ttl_hopl != base_header_inner.ipv6->ttl_hopl)
		{
			goto code_udp_lite_co_common;
		}
	}

	old_msn = v2_context->msn - 1;
	rohc_debugf(3, "old msn %Xh -> %Xh\n",old_msn,v2_context->msn);

	if( ( old_msn & 0xFFF0 ) == ( v2_context->msn & 0xFFF0 ) )
	{
		goto code_udp_lite_pt_0_crc3;
	}

	if( ( old_msn & 0xFFC0 ) == ( v2_context->msn & 0xFFC0 ) )
	{
		goto code_udp_lite_pt_0_crc7;
	}

code_udp_lite_co_common:
	rohc_debugf(3, "code_udp_lite_co_common\n");
	mptr.udp_lite_co_common = c_base_header.udp_lite_co_common + 1;
	// =:= '11111010'
	c_base_header.udp_lite_co_common->discriminator = 0xFA;
	// =:= irregular(2)
	c_base_header.udp_lite_co_common->reorder_ratio = v2_context->reorder_ratio;
	// =:= control_crc3_encoding
	c_base_header.udp_lite_co_common->control_crc3 = 0;
	if(base_header_inner.ipvx->version == IPV4)
	{
		// =:= profile_8_flags_enc(flags_indicator.CVALUE, ip_version.UVALUE)
		if(outer_ip_flag != 0 ||
		   base_header_inner.ipv4->df != ip_inner_context.v4->df ||
		   ip_inner_context.v4->ip_id_behavior != ip_inner_context.v4->last_ip_id_behavior ||
		   udp_lite_context->coverage_behavior != udp_lite_context->last_coverage_behavior)
		{
			// =:= irregular(1)
			c_base_header.udp_lite_co_common->flags_indicator = 1;
			// See RFC5225 page 76/77
			// =:= compressed_value(2,0)
			mptr.profile_8_flags->reserved = 0;
			// =:= irregular(1)
			mptr.profile_8_flags->ip_outer_indicator = outer_ip_flag;
			// =:= dont_fragment(ip_version)
			mptr.profile_8_flags->df = base_header_inner.ipv4->df;
			// =:= irregular(2)
			mptr.profile_8_flags->ip_id_behavior = ip_inner_context.v4->ip_id_behavior;
			// =:= irregular(2)
			mptr.profile_8_flags->coverage_behavior = udp_lite_context->coverage_behavior;
			++mptr.profile_8_flags;

			ip_inner_context.v4->df = base_header_inner.ipv4->df;
			ip_inner_context.v4->last_ip_id_behavior = ip_inner_context.v4->ip_id_behavior;
			udp_lite_context->last_coverage_behavior = udp_lite_context->coverage_behavior;
		}
		else
		{
			// =:= irregular(1)
			c_base_header.udp_lite_co_common->flags_indicator = 0;
		}
		// =:= static_or_irreg(tos_tc_indicator.CVALUE, 8)
		c_base_header.udp_lite_co_common->tos_tc_indicator = c2_static_or_irreg8(
		   &mptr,&ip_inner_context.v4->tos_tc,base_header_inner.ipv4->tos_tc);
		// =:= static_or_irreg(ttl_hopl_indicator.CVALUE, 8)
		c_base_header.udp_lite_co_common->ttl_hopl_indicator = c2_static_or_irreg8(
		   &mptr,&ip_inner_context.v4->ttl_hopl,base_header_inner.ipv4->ttl_hopl);
		// =:= msn_lsb(8)
		*(mptr.uint8++) = c_msn_lsb(8,v2_context);
		// =:= ip_id_sequential_variable(ip_id_behavior_innermost.UVALUE,ip_id_indicator.CVALUE)
		c_base_header.udp_lite_co_common->ip_id_indicator = c_ip_id_sequential_variable(
		   &mptr,ip_inner_context.v4->ip_id_behavior,ip_inner_context.v4->last_ip_id,ip_id,
		   v2_context->msn);
	}
	else
	{
		// =:= irregular(1)
		if(outer_ip_flag != 0 ||
		   udp_lite_context->coverage_behavior != udp_lite_context->last_coverage_behavior)
		{
			// =:= irregular(1)
			c_base_header.udp_lite_co_common->flags_indicator = 1;
			// See RFC5225 page 76/77
			// =:= compressed_value(2,0)
			mptr.profile_8_flags->reserved = 0;
			// =:= irregular(1)
			mptr.profile_8_flags->ip_outer_indicator = outer_ip_flag;
			// =:= dont_fragment(ip_version)
			mptr.profile_8_flags->df = 0;
			// =:= irregular(2)
			mptr.profile_8_flags->ip_id_behavior = ip_inner_context.v6->ip_id_behavior;
			// =:= irregular(2)
			mptr.profile_8_flags->coverage_behavior = udp_lite_context->coverage_behavior;
			++mptr.profile_8_flags;

			udp_lite_context->last_coverage_behavior = udp_lite_context->coverage_behavior;
		}
		else
		{
			// =:= irregular(1)
			c_base_header.udp_lite_co_common->flags_indicator = 0;
		}
		// =:= static_or_irreg(tos_tc_indicator.CVALUE, 8)
		#if __BYTE_ORDER == __LITTLE_ENDIAN
		if(ip_inner_context.v6->tos_tc1 != base_header_inner.ipv6->tos_tc1 ||
		   ip_inner_context.v6->tos_tc2 != base_header_inner.ipv6->tos_tc2)
		{
			c_base_header.udp_lite_co_common->tos_tc_indicator = 1;
			*(mptr.uint8++) = (base_header_inner.ipv6->tos_tc1 << 4) | base_header_inner.ipv6->tos_tc2;
			ip_inner_context.v6->tos_tc1 = base_header_inner.ipv6->tos_tc1;
			ip_inner_context.v6->tos_tc2 = base_header_inner.ipv6->tos_tc2;
		}
		else
		{
			c_base_header.udp_lite_co_common->tos_tc_indicator = 0;
		}
		#elif __BYTE_ORDER == __BIG_ENDIAN
		c_base_header.udp_lite_co_common->tos_tc_indicator = c2_static_or_irreg8(
		   &mptr,&ip_inner_context.v6->tos_tc,base_header_inner.ipv6->tos_tc);
		#endif
		// =:= static_or_irreg(ttl_hopl_indicator.CVALUE, 8)
		c_base_header.udp_lite_co_common->ttl_hopl_indicator = c2_static_or_irreg8(
		   &mptr,&ip_inner_context.v6->ttl_hopl,base_header_inner.ipv6->ttl_hopl);
		// =:= ip_id_sequential_variable(ip_id_behavior_innermost.UVALUE,ip_id_indicator.CVALUE)
		c_base_header.udp_lite_co_common->ip_id_indicator = 0;
		// =:= msn_lsb(8)
		*(mptr.uint8++) = c_msn_lsb(8,v2_context);
	}
	rohc_debugf(3, "co_common indicators (%2.2Xh): ip_id %d flags %d ttl_hopl %d tos_tc %d\n",
	            *(c_base_header.uint8 + 2),c_base_header.udp_lite_co_common->ip_id_indicator,
	            c_base_header.udp_lite_co_common->flags_indicator,
	            c_base_header.udp_lite_co_common->ttl_hopl_indicator,
	            c_base_header.udp_lite_co_common->tos_tc_indicator);
	size = mptr.uint8 - c_base_header.uint8;
	// =:= crc7(THIS.UVALUE,THIS.ULENGTH)
	c_base_header.udp_lite_co_common->header_crc = 0;
	c_base_header.udp_lite_co_common->header_crc =
	   crc_calculate(CRC_TYPE_7,c_base_header.uint8,size,CRC_INIT_7,
	                 context->compressor->crc_table_7);
	goto end_code_udp_lite_baseheader;

code_udp_lite_pt_0_crc3:
	rohc_debugf(3, "code_udp_lite_pt_0_crc3\n");
	size = sizeof(udp_lite_pt_0_crc3_t);
	// =:= '0'
	c_base_header.udp_lite_pt_0_crc3->discriminator = 0;
	// =:= msn_lsb(4)
	c_base_header.udp_lite_pt_0_crc3->msn = c_msn_lsb(4,v2_context);
	// =:= crc3(THIS.UVALUE,THIS.ULENGTH)
	c_base_header.udp_lite_pt_0_crc3->header_crc = 0;
	c_base_header.udp_lite_pt_0_crc3->header_crc =
	   crc_calculate(CRC_TYPE_3,c_base_header.uint8,size,CRC_INIT_3,
	                 context->compressor->crc_table_3);
	goto end_code_udp_lite_baseheader;

code_udp_lite_pt_0_crc7:
	rohc_debugf(3, "code_udp_lite_pt_0_crc7\n");
	size = sizeof(udp_lite_pt_0_crc7_t);
	// =:= '100'
	c_base_header.udp_lite_pt_0_crc7->discriminator = 0x04;
	// =:= msn_lsb(6)
	#if __BYTE_ORDER == __LITTLE_ENDIAN
	msn = c_msn_lsb(6,v2_context);
	c_base_header.udp_lite_pt_0_crc7->msn1 = msn >> 1;
	c_base_header.udp_lite_pt_0_crc7->msn2 = msn;
	#elif __BYTE_ORDER == __BIG_ENDIAN
	c_base_header.udp_lite_pt_0_crc7->msn = c_msn_lsb(6,v2_context);
	#endif
	// =:= crc7(THIS.UVALUE,THIS.ULENGTH)
	c_base_header.udp_lite_pt_0_crc7->header_crc = 0;
	c_base_header.udp_lite_pt_0_crc7->header_crc =
	   crc_calculate(CRC_TYPE_7,c_base_header.uint8,size,CRC_INIT_7,
	                 context->compressor->crc_table_7);
	goto end_code_udp_lite_baseheader;

code_udp_lite_pt_1_seq_id:
	rohc_debugf(3, "code_udp_lite_pt_1_seq_id\n");
	size = sizeof(udp_lite_pt_1_seq_id_t);
	// =:= '101'
	c_base_header.udp_lite_pt_1_seq_id->discriminator = 0x05;
	// =:= msn_lsb(6)
	#if __BYTE_ORDER == __LITTLE_ENDIAN
	msn = c_msn_lsb(6,v2_context);
	c_base_header.udp_lite_pt_1_seq_id->msn1 = msn >> 4;
	c_base_header.udp_lite_pt_1_seq_id->msn2 = msn;
	#elif __BYTE_ORDER == __BIG_ENDIAN
	c_base_header.udp_lite_pt_1_seq_id->msn = msn_lsb(6,v2_context);
	#endif
	c_base_header.udp_lite_pt_1_seq_id->ip_id =
	   c_ip_id_lsb(ip_inner_context.vx->ip_id_behavior,4,3,ip_inner_context.v4->last_ip_id,ip_id,
	               v2_context->msn);
	// =:= crc3(THIS.UVALUE,THIS.ULENGTH)
	c_base_header.udp_lite_pt_1_seq_id->header_crc = 0;
	c_base_header.udp_lite_pt_1_seq_id->header_crc =
	   crc_calculate(CRC_TYPE_3,c_base_header.uint8,size,CRC_INIT_3,
	                 context->compressor->crc_table_3);
	goto end_code_udp_lite_baseheader;

code_udp_lite_pt_2_seq_id:
	rohc_debugf(3, "code_udp_lite_pt_2_seq_id\n");
	size = sizeof(udp_lite_pt_2_seq_id_t);
	// =:= '110'
	c_base_header.udp_lite_pt_2_seq_id->discriminator = 0x06;
	// =:= ip_id_lsb(ip_id_behavior_innermost.UVALUE, 6)
	#if __BYTE_ORDER == __LITTLE_ENDIAN
	ipid = c_ip_id_lsb(ip_inner_context.v4->ip_id_behavior,6,3,ip_inner_context.v4->last_ip_id,ip_id,
	                   v2_context->msn);
	c_base_header.udp_lite_pt_2_seq_id->ip_id1 = ipid >> 1;
	c_base_header.udp_lite_pt_2_seq_id->ip_id2 = ipid;
	#elif __BYTE_ORDER == __BIG_ENDIAN
	c_base_header.udp_lite_pt_2_seq_id->ip_id =
	   c_ip_id_lsb(ip_inner_context.v4->ip_id_behavior,6,3,ip_inner_context.v4->last_ip_id,ip_id,
	               v2_context->msn);
	#endif
	// =:= msn_lsb(8)
	c_base_header.udp_lite_pt_2_seq_id->msn = c_msn_lsb(8,v2_context);
	// =:= crc7(THIS.UVALUE,THIS.ULENGTH)
	c_base_header.udp_lite_pt_2_seq_id->header_crc = 0;
	c_base_header.udp_lite_pt_2_seq_id->header_crc =
	   crc_calculate(CRC_TYPE_7,c_base_header.uint8,size,CRC_INIT_7,
	                 context->compressor->crc_table_7);

end_code_udp_lite_baseheader:
	#if ROHCV2_DEBUG
	rohc_debugf(3, "UDP-Lite baseheader, length %d\n",size);
	TraceData(c_base_header.uint8,size);
	#endif

	return c_base_header.uint8 + size;
}


// Page 68

u_int8_t * c_rohc_v2_code_esp_static_part( esp_context_t *esp_context, multi_ptr_t mptr,
                                           base_header_esp_t *esp )
{
	// Copy static part
	mptr.esp_static->spi = esp->spi;

	#if ROHCV2_DEBUG
	rohc_debugf(3, "ESP static part length %d\n",(int)sizeof(esp_static_t));
	TraceData(mptr.uint8,(int)sizeof(esp_static_t));
	#endif

	++mptr.esp_static;

	return mptr.uint8;
}


u_int8_t * c_rohc_v2_code_esp_dynamic_part( struct sc_v2_context *v2_context,
                                            esp_context_t *esp_context, multi_ptr_t mptr,
                                            base_header_esp_t *esp )
{
	// Copy dynamic part
	mptr.esp_dynamic->sequence_number = esp->sequence_number;
	mptr.esp_dynamic->reserved = 0;
	mptr.esp_dynamic->reorder_ratio = v2_context->reorder_ratio;

	// Save values in context
	esp_context->sequence_number = ntohl(esp->sequence_number);
	v2_context->msn = esp_context->sequence_number;

	#if ROHCV2_DEBUG
	rohc_debugf(3, "ESP dynamic part length %d\n",(int)sizeof(esp_dynamic_t));
	TraceData(mptr.uint8,(int)sizeof(esp_dynamic_t));
	#endif

	++mptr.esp_dynamic;

	return mptr.uint8;
}


// Page 58

u_int8_t * c_ipv6( int profile_value, int is_innermost, int outer_ip_flag, int reorder_ratio_value )
{
	return NULL;
}


// Page 60
u_int8_t * c_ipv4( int profile_value, int is_innermost, int outer_ip_flag, int ip_id_behavior_value,
                   int reorder_ratio_value )
{
	return NULL;
}


// Page 63
u_int8_t * c_udp( int profile_value, int reorder_ratio_value )
{
	return NULL;
}


// Page 64
u_int8_t * c_rtp( int profile_value, int ts_stride_value, int time_stride_value,
                  int reorder_ratio_value )
{
	return NULL;
}


// Page 67
u_int8_t * c_udp_lite( int profile_value, int reorder_ratio_value, int coverage_behavior_value )
{
	return NULL;
}


// Page 68
u_int8_t * c_esp( int profile_value, int reorder_ratio_value )
{
	return NULL;
}


/**
 * @brief Compress the innermost IP header AND the UDP header
 *
 * See RFC5225 page 85
 *
 * @param context                   The compression context
 * @param v2_context                The specific RoHCv2 context
 * @param ip_inner_context          The specific IP innermost context
 * @param c_base_header             The compressed header
 * @param base_header_inner         The innermost IP header
 * @param profile_value             The profile value
 * @param outer_ip_flag             The IP flag of the outer IP header
 * @return                          The new pointer in the rohc-packet-under-build buffer
 */

u_int8_t * c_udp_baseheader(struct c_context *const context,
                            struct sc_v2_context *v2_context,
                            ip_context_ptr_t ip_inner_context,
                            multi_ptr_t c_base_header,
                            base_header_ip_t base_header_inner,
                            int profile_value, int outer_ip_flag )
{
	multi_ptr_t mptr;
	u_int16_t old_msn;
	WB_t ip_id;
	#if __BYTE_ORDER == __LITTLE_ENDIAN
	u_int8_t msn;
	u_int8_t ipid;
	#endif
	int size;

	rohc_debugf(
	   3,
	   "context %p v2_context %p ip_context %p c_base_header %p base_header_inner %p profile_value %Xh outer_ip_flag %d\n",
	   context,v2_context,ip_inner_context.uint8,c_base_header.uint8,
	   base_header_inner.uint8,profile_value,outer_ip_flag);

	if(base_header_inner.ipvx->version == IPV4)
	{
		ip_id.uint16 = ntohs( base_header_inner.ipv4->ip_id );
		if(outer_ip_flag != 0)
		{
			goto code_udp_co_common;
		}
		if(ip_inner_context.v4->df != base_header_inner.ipv4->df ||
		   ip_inner_context.v4->tos_tc != base_header_inner.ipv4->tos_tc ||
		   ip_inner_context.v4->ttl_hopl != base_header_inner.ipv4->ttl_hopl)
		{
			goto code_udp_co_common;
		}
		if(ip_inner_context.v4->ip_id_behavior != ip_inner_context.v4->last_ip_id_behavior)
		{
			goto code_udp_co_common;
		}
		switch(ip_inner_context.v4->ip_id_behavior)
		{
			case IP_ID_BEHAVIOR_SEQUENTIAL:
				// If only 6 bits change
				if( ( ip_inner_context.v4->last_ip_id.uint16 & 0xFFC0 ) == ( ip_id.uint16 & 0xFFC0 ) )
				{
					goto code_udp_pt_1_seq_id;
				}
				break;
				// If only 8 bits change
				if( ( ip_inner_context.v4->last_ip_id.uint16 & 0xFF00 ) == ( ip_id.uint16 & 0xFF00 ) )
				{
					goto code_udp_pt_2_seq_id;
				}
				break;
			case IP_ID_BEHAVIOR_SEQUENTIAL_SWAPPED:
				// If only 6 bits change
				if( ( ip_inner_context.v4->last_ip_id.uint16 & 0xC0FF ) == ( ip_id.uint16 & 0xC0FF ) )
				{
					goto code_udp_pt_1_seq_id;
				}
				break;
				// If only 8 bits change
				if( ( ip_inner_context.v4->last_ip_id.uint16 & 0x00FF ) == ( ip_id.uint16 & 0x00FF ) )
				{
					goto code_udp_pt_2_seq_id;
				}
				break;
			default:
				goto code_udp_co_common;
		}
	}
	else
	{
		if(outer_ip_flag != 0)
		{
			goto code_udp_co_common;
		}
		#if __BYTE_ORDER == __LITTLE_ENDIAN
		if(ip_inner_context.v6->tos_tc1 != base_header_inner.ipv6->tos_tc1 ||
		   ip_inner_context.v6->tos_tc2 != base_header_inner.ipv6->tos_tc2)
		#elif __BYTE_ORDER == __BIG_ENDIAN
		if(ip_inner_context.v6->tos_tc != base_header_inner.ipv6->tos_tc)
		#endif
		{
			goto code_udp_co_common;
		}
		if(ip_inner_context.v6->ttl_hopl != base_header_inner.ipv6->ttl_hopl)
		{
			goto code_udp_co_common;
		}
	}

	old_msn = v2_context->msn - 1;
	rohc_debugf(3, "old msn %Xh -> %Xh\n",old_msn,v2_context->msn);

	if( ( old_msn & 0xFFF0 ) == ( v2_context->msn & 0xFFF0 ) )
	{
		goto code_udp_pt_0_crc3;
	}

	if( ( old_msn & 0xFFC0 ) == ( v2_context->msn & 0xFFC0 ) )
	{
		goto code_udp_pt_0_crc7;
	}

code_udp_co_common:
	rohc_debugf(3, "code_udp_co_common\n");
	mptr.udp_co_common = c_base_header.udp_co_common + 1;
	// =:= '11111010'
	c_base_header.udp_co_common->discriminator = 0xFA;
	// =:= irregular(2)
	c_base_header.udp_co_common->reorder_ratio = v2_context->reorder_ratio;
	// =:= control_crc3_encoding
	c_base_header.udp_co_common->control_crc3 = 0;
	if(base_header_inner.ipvx->version == IPV4)
	{
		// =:= profile_2_3_4_flags_enc(flags_indicator.CVALUE, ip_version.UVALUE)
		if(outer_ip_flag != 0 ||
		   base_header_inner.ipv4->df != ip_inner_context.v4->df ||
		   ip_inner_context.v4->ip_id_behavior != ip_inner_context.v4->last_ip_id_behavior)
		{
			// =:= irregular(1)
			c_base_header.udp_co_common->flags_indicator = 1;
			// See RFC5225 page 76
			// =:= irregular(1)
			mptr.profile_2_3_4_flags->name.ip_outer_indicator = outer_ip_flag;
			// =:= dont_fragment(ip_version)
			mptr.profile_2_3_4_flags->name.df = base_header_inner.ipv4->df;
			// =:= irregular(2)
			mptr.profile_2_3_4_flags->name.ip_id_behavior = ip_inner_context.v4->ip_id_behavior;
			// =:= compressed_value(4,0)
			mptr.profile_2_3_4_flags->name.reserved = 0;
			++mptr.profile_2_3_4_flags;

			ip_inner_context.v4->df = base_header_inner.ipv4->df;
			ip_inner_context.v4->last_ip_id_behavior = ip_inner_context.v4->ip_id_behavior;
		}
		else
		{
			// =:= irregular(1)
			c_base_header.udp_co_common->flags_indicator = 0;
		}
		// =:= static_or_irreg(tos_tc_indicator.CVALUE, 8)
		c_base_header.udp_co_common->tos_tc_indicator = c2_static_or_irreg8(
		   &mptr,&ip_inner_context.v4->tos_tc,base_header_inner.ipv4->tos_tc);
		// =:= static_or_irreg(ttl_hopl_indicator.CVALUE, 8)
		c_base_header.udp_co_common->ttl_hopl_indicator = c2_static_or_irreg8(
		   &mptr,&ip_inner_context.v4->ttl_hopl,base_header_inner.ipv4->ttl_hopl);
		// =:= msn_lsb(8)
		*(mptr.uint8++) = c_msn_lsb(8,v2_context);
		// =:= ip_id_sequential_variable(ip_id_behavior_innermost.UVALUE,ip_id_indicator.CVALUE)
		c_base_header.udp_co_common->ip_id_indicator = c_ip_id_sequential_variable(
		   &mptr,ip_inner_context.v4->ip_id_behavior,ip_inner_context.v4->last_ip_id,ip_id,
		   v2_context->msn);
	}
	else
	{
		// =:= irregular(1)
		if( ( c_base_header.udp_co_common->flags_indicator = outer_ip_flag ) != 0)
		{
			// See RFC5225 page 76
			// =:= irregular(1)
			mptr.profile_2_3_4_flags->name.ip_outer_indicator = outer_ip_flag;
			// =:= dont_fragment(ip_version)
			mptr.profile_2_3_4_flags->name.df = 0;
			// =:= irregular(2)
			mptr.profile_2_3_4_flags->name.ip_id_behavior = ip_inner_context.v6->ip_id_behavior;
			// =:= compressed_value(4,0)
			mptr.profile_2_3_4_flags->name.reserved = 0;
			++mptr.profile_2_3_4_flags;
		}
		// =:= static_or_irreg(tos_tc_indicator.CVALUE, 8)
		#if __BYTE_ORDER == __LITTLE_ENDIAN
		if(ip_inner_context.v6->tos_tc1 != base_header_inner.ipv6->tos_tc1 ||
		   ip_inner_context.v6->tos_tc2 != base_header_inner.ipv6->tos_tc2)
		{
			c_base_header.udp_co_common->tos_tc_indicator = 1;
			*(mptr.uint8++) = (base_header_inner.ipv6->tos_tc1 << 4) | base_header_inner.ipv6->tos_tc2;
			ip_inner_context.v6->tos_tc1 = base_header_inner.ipv6->tos_tc1;
			ip_inner_context.v6->tos_tc2 = base_header_inner.ipv6->tos_tc2;
		}
		else
		{
			c_base_header.udp_co_common->tos_tc_indicator = 0;
		}
//		c_base_header.udp_co_common->tos_tc_indicator = c2_static_or_irreg8(&mptr,(ip_inner_context.v6->tos_tc1<<4)|ip_inner_context.v6->tos_tc2,(base_header_inner.ipv6->tos_tc1<<4)|base_header_inner.ipv6->tos_tc2);
		#elif __BYTE_ORDER == __BIG_ENDIAN
		c_base_header.udp_co_common->tos_tc_indicator = c2_static_or_irreg8(
		   &mptr,&ip_inner_context.v6->tos_tc,base_header_inner.ipv6->tos_tc);
		#endif
		// =:= static_or_irreg(ttl_hopl_indicator.CVALUE, 8)
		c_base_header.udp_co_common->ttl_hopl_indicator = c2_static_or_irreg8(
		   &mptr,&ip_inner_context.v6->ttl_hopl,base_header_inner.ipv6->ttl_hopl);
		// =:= ip_id_sequential_variable(ip_id_behavior_innermost.UVALUE,ip_id_indicator.CVALUE)
		c_base_header.udp_co_common->ip_id_indicator = 0;
		// =:= msn_lsb(8)
		*(mptr.uint8++) = c_msn_lsb(8,v2_context);
	}
	size = mptr.uint8 - c_base_header.uint8;
	// =:= crc7(THIS.UVALUE,THIS.ULENGTH)
	c_base_header.udp_co_common->header_crc = 0;
	c_base_header.udp_co_common->header_crc =
	   crc_calculate(CRC_TYPE_7,c_base_header.uint8,size,CRC_INIT_7,
	                 context->compressor->crc_table_7);
	goto end_code_udp_baseheader;

code_udp_pt_0_crc3:
	rohc_debugf(3, "code_udp_pt_0_crc3\n");
	size = sizeof(udp_pt_0_crc3_t);
	// =:= '0'
	c_base_header.udp_pt_0_crc3->discriminator = 0;
	// =:= msn_lsb(4)
	c_base_header.udp_pt_0_crc3->msn = c_msn_lsb(4,v2_context);
	// =:= crc3(THIS.UVALUE,THIS.ULENGTH)
	c_base_header.udp_pt_0_crc3->header_crc = 0;
	c_base_header.udp_pt_0_crc3->header_crc =
	   crc_calculate(CRC_TYPE_3,c_base_header.uint8,size,CRC_INIT_3,
	                 context->compressor->crc_table_3);
	goto end_code_udp_baseheader;

code_udp_pt_0_crc7:
	rohc_debugf(3, "code_udp_pt_0_crc7\n");
	size = sizeof(udp_pt_0_crc7_t);
	// =:= '100'
	c_base_header.udp_pt_0_crc7->discriminator = 0x04;
	// =:= msn_lsb(6)
	#if __BYTE_ORDER == __LITTLE_ENDIAN
	msn = c_msn_lsb(6,v2_context);
	c_base_header.udp_pt_0_crc7->msn1 = msn >> 1;
	c_base_header.udp_pt_0_crc7->msn2 = msn;
	#elif __BYTE_ORDER == __BIG_ENDIAN
	c_base_header.udp_pt_0_crc7->msn = c_msn_lsb(6,v2_context);
	#endif
	// =:= crc7(THIS.UVALUE,THIS.ULENGTH)
	c_base_header.udp_pt_0_crc7->header_crc = 0;
	c_base_header.udp_pt_0_crc7->header_crc =
	   crc_calculate(CRC_TYPE_7,c_base_header.uint8,size,CRC_INIT_7,
	                 context->compressor->crc_table_7);
	goto end_code_udp_baseheader;

code_udp_pt_1_seq_id:
	rohc_debugf(3, "code_udp_pt_1_seq_id\n");
	size = sizeof(udp_pt_1_seq_id_t);
	// =:= '101'
	c_base_header.udp_pt_1_seq_id->discriminator = 0x05;
	// =:= msn_lsb(6)
	#if __BYTE_ORDER == __LITTLE_ENDIAN
	msn = c_msn_lsb(6,v2_context);
	c_base_header.udp_pt_1_seq_id->msn1 = msn >> 4;
	c_base_header.udp_pt_1_seq_id->msn2 = msn;
	#elif __BYTE_ORDER == __BIG_ENDIAN
	c_base_header.udp_pt_1_seq_id->msn = msn_lsb(6,v2_context);
	#endif
	c_base_header.udp_pt_1_seq_id->ip_id =
	   c_ip_id_lsb(ip_inner_context.vx->ip_id_behavior,4,3,ip_inner_context.v4->last_ip_id,ip_id,
	               v2_context->msn);
	// =:= crc3(THIS.UVALUE,THIS.ULENGTH)
	c_base_header.udp_pt_1_seq_id->header_crc = 0;
	c_base_header.udp_pt_1_seq_id->header_crc =
	   crc_calculate(CRC_TYPE_3,c_base_header.uint8,size,CRC_INIT_3,
	                 context->compressor->crc_table_3);
	goto end_code_udp_baseheader;

code_udp_pt_2_seq_id:
	rohc_debugf(3, "code_udp_pt_2_seq_id\n");
	size = sizeof(udp_pt_2_seq_id_t);
	// =:= '110'
	c_base_header.udp_pt_2_seq_id->discriminator = 0x06;
	// =:=ip_id_lsb(ip_id_behavior_innermost.UVALUE, 6)
	#if __BYTE_ORDER == __LITTLE_ENDIAN
	ipid = c_ip_id_lsb(ip_inner_context.v4->ip_id_behavior,6,3,ip_inner_context.v4->last_ip_id,ip_id,
	                   v2_context->msn);
	c_base_header.udp_pt_2_seq_id->ip_id1 = ipid >> 1;
	c_base_header.udp_pt_2_seq_id->ip_id2 = ipid;
	#elif __BYTE_ORDER == __BIG_ENDIAN
	c_base_header.udp_pt_2_seq_id->ip_id =
	   c_ip_id_lsb(ip_inner_context.v4->ip_id_behavior,6,3,ip_inner_context.v4->last_ip_id,ip_id,
	               v2_context->msn);
	#endif
	// =:= msn_lsb(8)
	c_base_header.udp_pt_2_seq_id->msn = c_msn_lsb(8,v2_context);
	// =:= crc7(THIS.UVALUE,THIS.ULENGTH)
	c_base_header.udp_pt_2_seq_id->header_crc = 0;
	c_base_header.udp_pt_2_seq_id->header_crc =
	   crc_calculate(CRC_TYPE_7,c_base_header.uint8,size,CRC_INIT_7,
	                 context->compressor->crc_table_7);

end_code_udp_baseheader:
	#if ROHCV2_DEBUG
	rohc_debugf(3, "UDP baseheader, length %d\n",size);
	TraceData(c_base_header.uint8,size);
	#endif

	return c_base_header.uint8 + size;
}


/**
 * @brief Compress the innermost IP header AND the ESP header
 *
 * See RFC5225 page 88
 *
 * @param context                   The compression context
 * @param v2_context                The specific RoHCv2 context
 * @param ip_inner_context          The specific IP innermost context
 * @param esp_context               The specific ESP context
 * @param c_base_header             The compressed header
 * @param base_header_inner         The innermost IP header
 * @param esp                       The ESP header
 * @param profile_value             The profile ID
 * @param outer_ip_flag             The IP flag of the outer IP header
 * @return                          The new pointer in the rohc-packet-under-build buffer
 */

u_int8_t * c_esp_baseheader(struct c_context *const context,
                            struct sc_v2_context *v2_context,
                            ip_context_ptr_t ip_inner_context,
                            esp_context_t *esp_context,
                            multi_ptr_t c_base_header,
                            base_header_ip_t base_header_inner,
                            base_header_esp_t *esp,
                            int profile_value, int outer_ip_flag )
{
	multi_ptr_t mptr;
	u_int16_t old_msn;
	WB_t ip_id;
	#if __BYTE_ORDER == __LITTLE_ENDIAN
	u_int8_t sequence_number;
	u_int8_t ipid;
	#endif
	int size;

	rohc_debugf(3, "context %p v2_context %p ip_inner_context %p esp_context %p\n",
	            context,v2_context,ip_inner_context.uint8,esp_context);
	rohc_debugf(3,
	            "c_base_header %p base_header_inner %p esp %p profile_value %Xh outer_ip_flag %d\n",
	            c_base_header.uint8,base_header_inner.uint8,esp,profile_value,
	            outer_ip_flag);

	if(base_header_inner.ipvx->version == IPV4)
	{
		ip_id.uint16 = ntohs( base_header_inner.ipv4->ip_id );
		if(outer_ip_flag != 0)
		{
			goto code_esp_co_common;
		}
		if(ip_inner_context.v4->df != base_header_inner.ipv4->df ||
		   ip_inner_context.v4->tos_tc != base_header_inner.ipv4->tos_tc ||
		   ip_inner_context.v4->ttl_hopl != base_header_inner.ipv4->ttl_hopl)
		{
			goto code_esp_co_common;
		}
		if( ( esp_context->sequence_number & 0xFFFF0000 ) !=
		    ( ntohl(esp->sequence_number) & 0xFFFF0000 ) )
		{
			goto code_esp_co_common;
		}
		if(ip_inner_context.v4->ip_id_behavior != ip_inner_context.v4->last_ip_id_behavior)
		{
			goto code_esp_co_common;
		}
		switch(ip_inner_context.v4->ip_id_behavior)
		{
			case IP_ID_BEHAVIOR_SEQUENTIAL:
				// If only 6 bits change
				if( ( ip_inner_context.v4->last_ip_id.uint16 & 0xFFC0 ) == ( ip_id.uint16 & 0xFFC0 ) )
				{
					goto code_esp_pt_1_seq_id;
				}
				break;
				// If only 8 bits change
				if( ( ip_inner_context.v4->last_ip_id.uint16 & 0xFF00 ) == ( ip_id.uint16 & 0xFF00 ) )
				{
					goto code_esp_pt_2_seq_id;
				}
				break;
			case IP_ID_BEHAVIOR_SEQUENTIAL_SWAPPED:
				// If only 6 bits change
				if( ( ip_inner_context.v4->last_ip_id.uint16 & 0xC0FF ) == ( ip_id.uint16 & 0xC0FF ) )
				{
					goto code_esp_pt_1_seq_id;
				}
				break;
				// If only 8 bits change
				if( ( ip_inner_context.v4->last_ip_id.uint16 & 0x00FF ) == ( ip_id.uint16 & 0x00FF ) )
				{
					goto code_esp_pt_2_seq_id;
				}
				break;
			default:
				goto code_esp_co_common;
		}
	}
	else
	{
		if(outer_ip_flag != 0)
		{
			goto code_esp_co_common;
		}
		#if __BYTE_ORDER == __LITTLE_ENDIAN
		if(ip_inner_context.v6->tos_tc1 != base_header_inner.ipv6->tos_tc1 ||
		   ip_inner_context.v6->tos_tc2 != base_header_inner.ipv6->tos_tc2)
		#elif __BYTE_ORDER == __BIG_ENDIAN
		if(ip_inner_context.v6->tos_tc != base_header_inner.ipv6->tos_tc)
		#endif
		{
			goto code_esp_co_common;
		}
		if(ip_inner_context.v6->ttl_hopl != base_header_inner.ipv6->ttl_hopl)
		{
			goto code_esp_co_common;
		}
		if( ( esp_context->sequence_number & 0xFFFF0000 ) !=
		    ( ntohl(esp->sequence_number) & 0xFFFF0000 ) )
		{
			goto code_esp_co_common;
		}
	}

	old_msn = v2_context->msn - 1;
	rohc_debugf(3, "old msn %Xh -> %Xh\n",old_msn,v2_context->msn);

	if( ( old_msn & 0xFFF0 ) == ( v2_context->msn & 0xFFF0 ) )
	{
		goto code_esp_pt_0_crc3;
	}

	if( ( old_msn & 0xFFC0 ) == ( v2_context->msn & 0xFFC0 ) )
	{
		goto code_esp_pt_0_crc7;
	}

code_esp_co_common:
	rohc_debugf(3, "code_esp_co_common\n");
	mptr.esp_co_common = c_base_header.esp_co_common + 1;
	// =:= '11111010'
	c_base_header.esp_co_common->discriminator = 0xFA;
	// =:= irregular(2)
	c_base_header.esp_co_common->reorder_ratio = v2_context->reorder_ratio;
	rohc_debugf(3, "ip_id_behavior %d\n",ip_inner_context.v4->ip_id_behavior);
	// =:= control_crc3_encoding
	c_base_header.esp_co_common->control_crc3 = 0;
	if(base_header_inner.ipvx->version == IPV4)
	{
		// =:= profile_2_3_4_flags_enc(flags_indicator.CVALUE, ip_version.UVALUE)
		if(outer_ip_flag != 0 ||
		   base_header_inner.ipv4->df != ip_inner_context.v4->df ||
		   ip_inner_context.v4->ip_id_behavior != ip_inner_context.v4->last_ip_id_behavior)
		{
			// =:= irregular(1)
			c_base_header.esp_co_common->flags_indicator = 1;
			// See RFC5225 page 76
			// =:= irregular(1)
			mptr.profile_2_3_4_flags->name.ip_outer_indicator = outer_ip_flag;
			// =:= dont_fragment(ip_version)
			mptr.profile_2_3_4_flags->name.df = base_header_inner.ipv4->df;
			// =:= irregular(2)
			mptr.profile_2_3_4_flags->name.ip_id_behavior = ip_inner_context.v4->ip_id_behavior;
			// =:= compressed_value(4,0)
			mptr.profile_2_3_4_flags->name.reserved = 0;
			++mptr.profile_2_3_4_flags;

			ip_inner_context.v4->df = base_header_inner.ipv4->df;
			ip_inner_context.v4->last_ip_id_behavior = ip_inner_context.v4->ip_id_behavior;
		}
		else
		{
			// =:= irregular(1)
			c_base_header.esp_co_common->flags_indicator = 0;
		}
		// =:= static_or_irreg(tos_tc_indicator.CVALUE, 8)
		c_base_header.esp_co_common->tos_tc_indicator = c2_static_or_irreg8(
		   &mptr,&ip_inner_context.v4->tos_tc,base_header_inner.ipv4->tos_tc);
		// =:= static_or_irreg(ttl_hopl_indicator.CVALUE, 8)
		c_base_header.esp_co_common->ttl_hopl_indicator = c2_static_or_irreg8(
		   &mptr,&ip_inner_context.v4->ttl_hopl,base_header_inner.ipv4->ttl_hopl);
		// =:= sdvl_sn_lsb(sequence_number.ULENGTH)
		mptr.uint8 = c_sdvl_sn_lsb(mptr,ntohl(esp->sequence_number));
		// =:= ip_id_sequential_variable(ip_id_behavior_innermost.UVALUE,ip_id_indicator.CVALUE)
		c_base_header.esp_co_common->ip_id_indicator = c_ip_id_sequential_variable(
		   &mptr,ip_inner_context.v4->ip_id_behavior,ip_inner_context.v4->last_ip_id,ip_id,
		   v2_context->msn);
	}
	else
	{
		// =:= irregular(1)
		c_base_header.esp_co_common->ip_id_indicator = 0;
		// =:= irregular(1)
		if( ( c_base_header.esp_co_common->flags_indicator = outer_ip_flag ) != 0)
		{
			// See RFC5225 page 76
			// =:= irregular(1)
			mptr.profile_2_3_4_flags->name.ip_outer_indicator = outer_ip_flag;
			// =:= dont_fragment(ip_version)
			mptr.profile_2_3_4_flags->name.df = 0;
			// =:= irregular(2)
			mptr.profile_2_3_4_flags->name.ip_id_behavior = ip_inner_context.v6->ip_id_behavior;
			// =:= compressed_value(4,0)
			mptr.profile_2_3_4_flags->name.reserved = 0;
			++mptr.profile_2_3_4_flags;
		}
		// =:= static_or_irreg(tos_tc_indicator.CVALUE, 8)
		#if __BYTE_ORDER == __LITTLE_ENDIAN
		if(ip_inner_context.v6->tos_tc1 != base_header_inner.ipv6->tos_tc1 ||
		   ip_inner_context.v6->tos_tc2 != base_header_inner.ipv6->tos_tc2)
		{
			c_base_header.esp_co_common->tos_tc_indicator = 1;
			*(mptr.uint8++) = (base_header_inner.ipv6->tos_tc1 << 4) | base_header_inner.ipv6->tos_tc2;
			ip_inner_context.v6->tos_tc1 = base_header_inner.ipv6->tos_tc1;
			ip_inner_context.v6->tos_tc2 = base_header_inner.ipv6->tos_tc2;
		}
		else
		{
			c_base_header.esp_co_common->tos_tc_indicator = 0;
		}
//		c_base_header.esp_co_common->tos_tc_indicator = c2_static_or_irreg8(&mptr,(ip_inner_context.v6->tos_tc1<<4)|ip_inner_context.v6->tos_tc2,(base_header_inner.ipv6->tos_tc1<<4)|base_header_inner.ipv6->tos_tc2);
		#elif __BYTE_ORDER == __BIG_ENDIAN
		c_base_header.esp_co_common->tos_tc_indicator = c2_static_or_irreg8(
		   &mptr,&ip_inner_context.v6->tos_tc,base_header_inner.ipv6->tos_tc);
		#endif
		// =:= static_or_irreg(ttl_hopl_indicator.CVALUE, 8)
		c_base_header.esp_co_common->ttl_hopl_indicator = c2_static_or_irreg8(
		   &mptr,&ip_inner_context.v6->ttl_hopl,base_header_inner.ipv6->ttl_hopl);
		// =:= sdvl_sn_lsb(sequence_number.ULENGTH)
		mptr.uint8 = c_sdvl_sn_lsb(mptr,ntohl(esp->sequence_number));
	}
	rohc_debugf(3, "co_common indicators (%2.2Xh): ip_id %d flags %d ttl_hopl %d tos_tc %d\n",
	            *(c_base_header.uint8 + 2),c_base_header.esp_co_common->ip_id_indicator,
	            c_base_header.esp_co_common->flags_indicator,
	            c_base_header.esp_co_common->ttl_hopl_indicator,
	            c_base_header.esp_co_common->tos_tc_indicator);
	size = mptr.uint8 - c_base_header.uint8;
	// =:= crc7(THIS.UVALUE,THIS.ULENGTH)
	c_base_header.esp_co_common->header_crc = 0;
	c_base_header.esp_co_common->header_crc =
	   crc_calculate(CRC_TYPE_7,c_base_header.uint8,size,CRC_INIT_7,
	                 context->compressor->crc_table_7);
	rohc_debugf(3, "co_common crc7 %2.2Xh length %d\n",c_base_header.esp_co_common->header_crc,size);
	goto end_code_esp_baseheader;

code_esp_pt_0_crc3:
	rohc_debugf(3, "code_esp_pt_0_crc3\n");
	size = sizeof(esp_pt_0_crc3_t);
	// =:= '0'
	c_base_header.esp_pt_0_crc3->discriminator = 0;
	// =:= msn_lsb(4)
	c_base_header.esp_pt_0_crc3->sequence_number = c_msn_lsb(4,v2_context);
	// =:= crc3(THIS.UVALUE,THIS.ULENGTH)
	c_base_header.esp_pt_0_crc3->header_crc = 0;
	c_base_header.esp_pt_0_crc3->header_crc =
	   crc_calculate(CRC_TYPE_3,c_base_header.uint8,size,CRC_INIT_3,
	                 context->compressor->crc_table_3);
	goto end_code_esp_baseheader;

code_esp_pt_0_crc7:
	rohc_debugf(3, "code_esp_pt_0_crc7\n");
	size = sizeof(esp_pt_0_crc7_t);
	// =:= '100'
	c_base_header.esp_pt_0_crc7->discriminator = 0x04;
	// =:= msn_lsb(6)
	#if __BYTE_ORDER == __LITTLE_ENDIAN
	sequence_number = c_msn_lsb(6,v2_context);
	c_base_header.esp_pt_0_crc7->sequence_number1 = sequence_number >> 1;
	c_base_header.esp_pt_0_crc7->sequence_number2 = sequence_number;
	#elif __BYTE_ORDER == __BIG_ENDIAN
	c_base_header.esp_pt_0_crc7->sequence_number = c_msn_lsb(6,v2_context);
	#endif
	// =:= crc7(THIS.UVALUE,THIS.ULENGTH)
	c_base_header.esp_pt_0_crc7->header_crc = 0;
	c_base_header.esp_pt_0_crc7->header_crc =
	   crc_calculate(CRC_TYPE_7,c_base_header.uint8,size,CRC_INIT_7,
	                 context->compressor->crc_table_7);
	goto end_code_esp_baseheader;

code_esp_pt_1_seq_id:
	rohc_debugf(3, "code_esp_pt_1_seq_id\n");
	size = sizeof(esp_pt_1_seq_id_t);
	// =:= '101'
	c_base_header.esp_pt_1_seq_id->discriminator = 0x05;
	// =:= msn_lsb(6)
	#if __BYTE_ORDER == __LITTLE_ENDIAN
	sequence_number = c_msn_lsb(6,v2_context);
	c_base_header.esp_pt_1_seq_id->sequence_number1 = sequence_number >> 4;
	c_base_header.esp_pt_1_seq_id->sequence_number2 = sequence_number;
	#elif __BYTE_ORDER == __BIG_ENDIAN
	c_base_header.esp_pt_1_seq_id->sequence_number = msn_lsb(6,v2_context);
	#endif
	c_base_header.esp_pt_1_seq_id->ip_id =
	   c_ip_id_lsb(ip_inner_context.vx->ip_id_behavior,4,3,ip_inner_context.v4->last_ip_id,ip_id,
	               v2_context->msn);
	// =:= crc3(THIS.UVALUE,THIS.ULENGTH)
	c_base_header.esp_pt_1_seq_id->header_crc = 0;
	c_base_header.esp_pt_1_seq_id->header_crc =
	   crc_calculate(CRC_TYPE_3,c_base_header.uint8,sizeof(esp_pt_1_seq_id_t),CRC_INIT_3,
	                 context->compressor->crc_table_3);
	goto end_code_esp_baseheader;

code_esp_pt_2_seq_id:
	rohc_debugf(3, "code_esp_pt_2_seq_id\n");
	size = sizeof(esp_pt_2_seq_id_t);
	// =:= '110'
	c_base_header.esp_pt_2_seq_id->discriminator = 0x06;
	// =:=ip_id_lsb(ip_id_behavior_innermost.UVALUE, 6)
	#if __BYTE_ORDER == __LITTLE_ENDIAN
	ipid = c_ip_id_lsb(ip_inner_context.v4->ip_id_behavior,6,3,ip_inner_context.v4->last_ip_id,ip_id,
	                   v2_context->msn);
	c_base_header.esp_pt_2_seq_id->ip_id1 = ipid >> 1;
	c_base_header.esp_pt_2_seq_id->ip_id2 = ipid;
	#elif __BYTE_ORDER == __BIG_ENDIAN
	c_base_header.esp_pt_2_seq_id->ip_id =
	   c_ip_id_lsb(ip_inner_context.v4->ip_id_behavior,6,3,ip_inner_context.v4->last_ip_id,ip_id,
	               v2_context->msn);
	#endif
	// =:= msn_lsb(8)
	c_base_header.esp_pt_2_seq_id->sequence_number = c_msn_lsb(8,v2_context);
	// =:= crc7(THIS.UVALUE,THIS.ULENGTH)
	c_base_header.esp_pt_2_seq_id->header_crc = 0;
	c_base_header.esp_pt_2_seq_id->header_crc =
	   crc_calculate(CRC_TYPE_7, c_base_header.uint8, size, CRC_INIT_7,
	                 context->compressor->crc_table_7);

end_code_esp_baseheader:

	// Save value in context
	esp_context->sequence_number = ntohl(esp->sequence_number);

	#if ROHCV2_DEBUG
	rohc_debugf(3, "ESP baseheader, length %d\n",size);
	TraceData(c_base_header.uint8,size);
	#endif

	return c_base_header.uint8 + size;
}


/**
 * @brief Compress the innermost IP header AND the UDP and RTP header
 *
 * See RFC5225 page 78
 *
 * @param context                   The compression context
 * @param v2_context                The specific RoHCv2 context
 * @param ip_inner_context          The specific IP innermost context
 * @param rtp_context               The specific RTP context
 * @param c_base_header             The compressed header
 * @param base_header_inner         The innermost IP header
 * @param rtp                       The decoded RTP base header
 * @param profile_value             The profile value
 * @param outer_ip_flag             The IP flag of the outer IP header
 * @return                          The new pointer in the rohc-packet-under-build buffer
 */

u_int8_t * c_rtp_baseheader(struct c_context *const context,
                            struct sc_v2_context *v2_context,
                            ip_context_ptr_t ip_inner_context,
                            rtp_context_t *rtp_context,
                            multi_ptr_t c_base_header,
                            base_header_ip_t base_header_inner,
                            base_header_rtp_with_csrc_t *rtp,
                            int profile_value, int outer_ip_flag )
{
	profile_1_7_flags1_t *profile_1_7_flags1;
	profile_1_flags2_t *profile_1_flags2;
	multi_ptr_t mptr;
	u_int16_t old_msn;
	WB_t ip_id;
	#if __BYTE_ORDER == __LITTLE_ENDIAN
	u_int8_t msn;
	#endif
	u_int8_t ipid;
	int size;

	rohc_debugf(3, "context %p v2_context %p ip_context %p rtp_context %p\n",context,v2_context,
	            ip_inner_context.uint8,
	            rtp_context);
	rohc_debugf(
	   3,
	   "  c_base_header %p base_header_inner %p rtp_with_csrc %p profile_value %Xh outer_ip_flag %d\n",
	   c_base_header.uint8,base_header_inner.uint8,rtp,profile_value,outer_ip_flag);

	/*
	    format            discriminator   crc   msn   ip_id  ts_scaled  marker

	rtp_pt_0_crc3        1 '0'     0x00    3     4
	rtp_pt_0_crc7        4 '1000'  0x08    7     5

	rtp_pt_1_rnd         3 '101'   0x05    3     4             5          1
	rtp_pt_1_seq_id      4 '1001'  0x09    3     5      4
	rtp_pt_1_seq_ts      3 '101'   0x05    3     4             5          1

	rtp_pt_2_rnd         3 '110'   0x06    7     7             6          1
	rtp_pt_2_seq_id      5 '11000' 0x18    7     7      5
	rtp_pt_2_seq_both    5 '11001' 0x19    7     7      5      7          1
	rtp_pt_2_seq_ts      4 '1101'  0x0D    7     7             5          1
	*/

	if(base_header_inner.ipvx->version == IPV4)
	{
		ip_id.uint16 = ntohs( base_header_inner.ipv4->ip_id );
		if(outer_ip_flag != 0)
		{
			goto code_rtp_co_common;
		}
		if(ip_inner_context.v4->df != base_header_inner.ipv4->df ||
		   ip_inner_context.v4->tos_tc != base_header_inner.ipv4->tos_tc ||
		   ip_inner_context.v4->ttl_hopl != base_header_inner.ipv4->ttl_hopl)
		{
			goto code_rtp_co_common;
		}
		if(ip_inner_context.v4->ip_id_behavior != ip_inner_context.v4->last_ip_id_behavior)
		{
			goto code_rtp_co_common;
		}
		switch(ip_inner_context.v4->ip_id_behavior)
		{
			case IP_ID_BEHAVIOR_SEQUENTIAL:
				// If only 6 bits change
				if( ( ip_inner_context.v4->last_ip_id.uint16 & 0xFFC0 ) == ( ip_id.uint16 & 0xFFC0 ) )
				{
					goto code_rtp_pt_1_seq_id;
				}
				break;
				// If only 8 bits change
				if( ( ip_inner_context.v4->last_ip_id.uint16 & 0xFF00 ) == ( ip_id.uint16 & 0xFF00 ) )
				{
					goto code_rtp_pt_2_seq_id;
				}
				break;
			case IP_ID_BEHAVIOR_SEQUENTIAL_SWAPPED:
				// If only 6 bits change
				if( ( ip_inner_context.v4->last_ip_id.uint16 & 0xC0FF ) == ( ip_id.uint16 & 0xC0FF ) )
				{
					goto code_rtp_pt_1_seq_id;
				}
				break;
				// If only 8 bits change
				if( ( ip_inner_context.v4->last_ip_id.uint16 & 0x00FF ) == ( ip_id.uint16 & 0x00FF ) )
				{
					goto code_rtp_pt_2_seq_id;
				}
				break;
			default:
				goto code_rtp_co_common;
		}
	}
	else
	{
		if(outer_ip_flag != 0)
		{
			goto code_rtp_co_common;
		}
		#if __BYTE_ORDER == __LITTLE_ENDIAN
		if(ip_inner_context.v6->tos_tc1 != base_header_inner.ipv6->tos_tc1 ||
		   ip_inner_context.v6->tos_tc2 != base_header_inner.ipv6->tos_tc2)
		#elif __BYTE_ORDER == __BIG_ENDIAN
		if(ip_inner_context.v6->tos_tc != base_header_inner.ipv6->tos_tc)
		#endif
		{
			goto code_rtp_co_common;
		}
		if(ip_inner_context.v6->ttl_hopl != base_header_inner.ipv6->ttl_hopl)
		{
			goto code_rtp_co_common;
		}

		goto code_rtp_co_common;
	}

	old_msn = v2_context->msn - 1;
	rohc_debugf(3, "old msn %Xh -> %Xh\n",old_msn,v2_context->msn);

	if( ( old_msn & 0xFFF0 ) == ( v2_context->msn & 0xFFF0 ) )
	{
		goto code_rtp_pt_0_crc3;
	}

	if( ( old_msn & 0xFFC0 ) == ( v2_context->msn & 0xFFC0 ) )
	{
		goto code_rtp_pt_0_crc7;
	}

code_rtp_co_common:
	{
		u_int32_t timestamp;
		int tsc_indicator;
		int tss_indicator;


		rohc_debugf(3, "code_rtp_co_common\n");
		mptr.rtp_co_common = c_base_header.rtp_co_common + 1;
		// =:= '11111010'
		c_base_header.rtp_co_common->discriminator = 0xFA;
		// =:= irregular(1)
		c_base_header.rtp_co_common->marker = rtp->marker;

		if(base_header_inner.ipvx->version == IPV4)
		{
			// If outer_ip_indicator
			if(outer_ip_flag != 0 ||
			   // or ttl_hopl_indicator
			   ip_inner_context.v4->ttl_hopl != base_header_inner.ipv4->ttl_hopl ||
			   // or tos_tc_indicator
			   ip_inner_context.v4->tos_tc != base_header_inner.ipv4->tos_tc ||
			   // df change
			   ip_inner_context.v4->df != base_header_inner.ipv4->df ||
			   // ip_id_behavior_innermost change
			   ip_inner_context.v4->ip_id_behavior != ip_inner_context.v4->last_ip_id_behavior)
			{
				// All indicators set to zero
				*mptr.uint8 = 0;
				// profile_1_flags2 present
				profile_1_7_flags1 = mptr.profile_1_7_flags1++;
				// =:= irregular(1)
				c_base_header.rtp_co_common->flags1_indicator = 1;
			}
			else
			{
				// profile_1_flags2 not present
				profile_1_7_flags1 = NULL;
				// =:= irregular(1)
				c_base_header.rtp_co_common->flags1_indicator = 0;
			}
		}
		else
		{
			// If outer_ip_indicator
			if(outer_ip_flag != 0 ||
			   // or ttl_hopl_indicator
			   ip_inner_context.v6->ttl_hopl != base_header_inner.ipv6->ttl_hopl ||
			   // or tos_tc_indicator
			 #if __BYTE_ORDER == __LITTLE_ENDIAN
			   ip_inner_context.v6->tos_tc1 != base_header_inner.ipv6->tos_tc1 ||
			   ip_inner_context.v6->tos_tc2 != base_header_inner.ipv6->tos_tc2)
			 #elif __BYTE_ORDER == __BIG_ENDIAN
			   ip_inner_context.v6->tos_tc != base_header_inner.ipv6->tos_tc)
			 #endif
			{
				// profile_1_flags2 present
				profile_1_7_flags1 = mptr.profile_1_7_flags1++;
				// =:= irregular(1)
				c_base_header.rtp_co_common->flags1_indicator = 1;
			}
			else
			{
				// profile_1_flags2 not present
				profile_1_7_flags1 = NULL;
				// =:= irregular(1)
				c_base_header.rtp_co_common->flags1_indicator = 0;
			}
		}

		// If list_indicator
		if(rtp->cc != 0 ||
		   // or pt_indicator
		   rtp->payload_type != rtp_context->payload_type ||
		   // or time_stride_indicator (tis_indicator)
		   rtp_context->time_stride != rtp_context->last_time_stride ||
		   // pad_bit change
		   rtp->pad_bit != rtp_context->pad_bit ||
		   // or extension change
		   rtp->extension != 0)
		{
			// All indicators set to zero
			*mptr.uint8 = 0;
			// profile_1_flags2 present
			profile_1_flags2 = mptr.profile_1_flags2++;
			// =:= irregular(1)
			c_base_header.rtp_co_common->flags2_indicator = 1;
		}
		else
		{
			// profile_1_flags2 not present
			profile_1_flags2 = NULL;
			// =:= irregular(1)
			c_base_header.rtp_co_common->flags2_indicator = 0;
		}

		// =:= profile_1_7_flags1_enc(flags1_indicator.CVALUE, ip_version.UVALUE)
		if(profile_1_7_flags1 != NULL)
		{
			// See RFC5225 page 75
			// =:= irregular(1)
			profile_1_7_flags1->name.ip_outer_indicator = outer_ip_flag;
			if(base_header_inner.ipvx->version == IPV4)
			{
				// =:= irregular(1) / =:= static_or_irreg(tos_tc_indicator.CVALUE, 8)
				profile_1_7_flags1->name.tos_tc_indicator = c2_static_or_irreg8(
				   &mptr,&ip_inner_context.v4->tos_tc,base_header_inner.ipv4->tos_tc);
				// =:= irregular(1) / =:= static_or_irreg(ttl_hopl_indicator.CVALUE, ttl_hopl.ULENGTH)
				profile_1_7_flags1->name.ttl_hopl_indicator = c2_static_or_irreg8(
				   &mptr,&ip_inner_context.v4->ttl_hopl,base_header_inner.ipv4->ttl_hopl);
				// =:= dont_fragment(ip_version)
				profile_1_7_flags1->name.df = base_header_inner.ipv4->df;
				// =:= irregular(2)
				profile_1_7_flags1->name.ip_id_behavior = ip_inner_context.v4->ip_id_behavior;

				ip_inner_context.v4->df = base_header_inner.ipv4->df;
				ip_inner_context.v4->last_ip_id_behavior = ip_inner_context.v4->ip_id_behavior;
			}
			else
			{
				// =:= irregular(1) / =:= static_or_irreg(tos_tc_indicator.CVALUE, 8)
			#if __BYTE_ORDER == __LITTLE_ENDIAN
				if(ip_inner_context.v6->tos_tc1 != base_header_inner.ipv6->tos_tc1 ||
				   ip_inner_context.v6->tos_tc2 != base_header_inner.ipv6->tos_tc2)
				{
					profile_1_7_flags1->name.tos_tc_indicator = 1;
					*(mptr.uint8++) =
					   (base_header_inner.ipv6->tos_tc1 << 4) | base_header_inner.ipv6->tos_tc2;
					ip_inner_context.v6->tos_tc1 = base_header_inner.ipv6->tos_tc1;
					ip_inner_context.v6->tos_tc2 = base_header_inner.ipv6->tos_tc2;
				}
				else
				{
					profile_1_7_flags1->name.tos_tc_indicator = 0;
				}
//			profile_1_7_flags1->name.tos_tc_indicator = c2_static_or_irreg8(&mptr,(ip_inner_context.v6->tos_tc1<<4)|ip_inner_context.v6->tos_tc2,(base_header_inner.ipv6->tos_tc1<<4)|base_header_inner.ipv6->tos_tc2);
			#elif __BYTE_ORDER == __BIG_ENDIAN
				profile_1_7_flags1->name.tos_tc_indicator = c2_static_or_irreg8(
				   &mptr,&ip_inner_context.v6->tos_tc,base_header_inner.ipv6->tos_tc);
			#endif
				// =:= irregular(1) / static_or_irreg(ttl_hopl_indicator.CVALUE, ttl_hopl.ULENGTH)
				profile_1_7_flags1->name.ttl_hopl_indicator = c2_static_or_irreg8(
				   &mptr,&ip_inner_context.v6->ttl_hopl,base_header_inner.ipv6->ttl_hopl);
				// =:= dont_fragment(ip_version)
				profile_1_7_flags1->name.df = 0;
				// =:= irregular(2)
				profile_1_7_flags1->name.ip_id_behavior = ip_inner_context.v6->ip_id_behavior;
			}
			// =:= irregular(2)
			profile_1_7_flags1->name.reorder_ratio = v2_context->reorder_ratio;
		}
		// payload_type =:= pt_irr_or_static(pt_indicator)
		if(rtp_context->payload_type != rtp->payload_type)
		{
			// =:= irregular(1)
			profile_1_flags2->name.pt_indicator = 1;
			// See RFC5225 page 71
			*(mptr.uint8++) = rtp->payload_type;
			rtp_context->payload_type = rtp->payload_type;
		}
		// sequence_number =:= sdvl_sn_lsb(sequence_number.ULENGTH)
		mptr.uint8 = c_sdvl_sn_lsb(mptr,ntohs(rtp->sequence_number));
		if(base_header_inner.ipvx->version == IPV4)
		{
			// ip_id =:= ip_id_sequential_variable(ip_id_behavior_innermost.UVALUE, ip_id_indicator.CVALUE)
			c_base_header.rtp_co_common->ip_id_indicator = c_ip_id_sequential_variable(
			   &mptr,ip_inner_context.v4->ip_id_behavior,ip_inner_context.v4->last_ip_id,ip_id,
			   v2_context->msn);
			rohc_debugf(3, "ip_id_indicator %d\n",c_base_header.rtp_co_common->ip_id_indicator);
		}
		else
		{
			c_base_header.rtp_co_common->ip_id_indicator = 0;
		}

		tsc_indicator = 0;
		tss_indicator = rtp_context->last_ts_stride & ~rtp_context->ts_stride;

		timestamp = ntohl(rtp->timestamp);
		c_base_header.rtp_co_common->tsc_indicator = tsc_indicator;

		rohc_debugf(3, "code variable_scaled_timestamp\n");
		// ts_scaled =:= variable_scaled_timestamp(tss_indicator.CVALUE, tsc_indicator.CVALUE, ts_stride.UVALUE, time_stride.UVALUE)
// A REVOIR :
		mptr.uint8 = c_variable_scaled_timestamp(
		   mptr,rtp_context,/*c_base_header.rtp_co_common->tss_indicator*/ tss_indicator,
		   /*c_base_header.rtp_co_common->tsc_indicator*/ tsc_indicator,timestamp);
		rohc_debugf(3, "code variable_unscaled_timestamp\n");
		// timestamp =:= variable_unscaled_timestamp(tss_indicator.CVALUE, tsc_indicator.CVALUE)
		mptr.uint8 = c_variable_unscaled_timestamp(
		   mptr,/*c_base_header.rtp_co_common->tss_indicator*/ tss_indicator,
		   /*c_base_header.rtp_co_common->tsc_indicator*/ tsc_indicator,rtp_context->last_timestamp,
		   timestamp);
		rtp_context->last_timestamp = timestamp;
		rohc_debugf(3, "code ts_stride =:= sdvl_or_static(tss_indicator.CVALUE)\n");
		// ts_stride =:= sdvl_or_static(tss_indicator.CVALUE)
		c_base_header.rtp_co_common->tss_indicator =
		   c_sdvl_or_static(&mptr,rtp_context->last_ts_stride,
		                    rtp_context->ts_stride);
		rtp_context->last_ts_stride = rtp_context->ts_stride;
		rohc_debugf(3, "\n");
		// =:= profile_1_flags2_enc(flags2_indicator.CVALUE)
		if(profile_1_flags2 != NULL)
		{
			rohc_debugf(3, "code time_stride\n");
			// time_stride =:= sdvl_or_static(tis_indicator.CVALUE)
			profile_1_flags2->name.time_stride_indicator = c_sdvl_or_static(
			   &mptr,rtp_context->last_time_stride,rtp_context->time_stride);
			rtp_context->last_time_stride = rtp_context->time_stride;
			// =:= irregular(1)
			mptr.profile_1_flags2->name.pad_bit = rtp->pad_bit;
			rtp_context->pad_bit = rtp->pad_bit;
			// =:= irregular(1)
			mptr.profile_1_flags2->name.extension = rtp->extension;
			rtp_context->extension = rtp->extension;

			if(rtp->cc != 0)
			{
				// =:= irregular(1)
				mptr.profile_1_flags2->name.list_indicator = 1;
				// csrc_list =:= csrc_list_presence(list_indicator.CVALUE)
				mptr.uint8 = c_csrc_list_dynchain(rtp_context,mptr,rtp);
			}
		}

		size = mptr.uint8 - c_base_header.uint8;
		// =:= crc7(THIS.UVALUE,THIS.ULENGTH)
		c_base_header.rtp_co_common->header_crc = 0;
		c_base_header.rtp_co_common->header_crc =
		   crc_calculate(CRC_TYPE_7,c_base_header.uint8,size,CRC_INIT_7,
		                 context->compressor->crc_table_7);
		goto end_code_rtp_baseheader;
	}
code_rtp_pt_0_crc3:
	rohc_debugf(3, "code_rtp_pt_0_crc3\n");
	size = sizeof(rtp_pt_0_crc3_t);
	// =:= '0'
	c_base_header.rtp_pt_0_crc3->discriminator = 0;
	// =:= msn_lsb(4)
	c_base_header.rtp_pt_0_crc3->msn = c_msn_lsb(4,v2_context);
	// =:= crc3(THIS.UVALUE,THIS.ULENGTH)
	c_base_header.rtp_pt_0_crc3->header_crc = 0;
	c_base_header.rtp_pt_0_crc3->header_crc =
	   crc_calculate(CRC_TYPE_3,c_base_header.uint8,size,CRC_INIT_3,
	                 context->compressor->crc_table_3);
	goto end_code_rtp_baseheader;

code_rtp_pt_0_crc7:
	rohc_debugf(3, "code_rtp_pt_0_crc7\n");
	size = sizeof(rtp_pt_0_crc7_t);
	// =:= '1000'
	c_base_header.rtp_pt_0_crc7->discriminator = 0x08;
	// =:= msn_lsb(5)
	#if __BYTE_ORDER == __LITTLE_ENDIAN
	msn = c_msn_lsb(5,v2_context);
	c_base_header.rtp_pt_0_crc7->msn1 = msn >> 1;
	c_base_header.rtp_pt_0_crc7->msn2 = msn;
	#elif __BYTE_ORDER == __BIG_ENDIAN
	c_base_header.rtp_pt_0_crc7->msn = c_msn_lsb(5,v2_context);
	#endif
	// =:= crc7(THIS.UVALUE,THIS.ULENGTH)
	c_base_header.rtp_pt_0_crc7->header_crc = 0;
	c_base_header.rtp_pt_0_crc7->header_crc =
	   crc_calculate(CRC_TYPE_7,c_base_header.uint8,size,CRC_INIT_7,
	                 context->compressor->crc_table_7);
	goto end_code_rtp_baseheader;

code_rtp_pt_1_rnd:
	rohc_debugf(3, "code_rtp_pt_1_rnd\n");
	size = sizeof(rtp_pt_1_rnd_t);
	// =:= '101'
	c_base_header.rtp_pt_1_rnd->discriminator = 0x05;
	// =:= irregular(1)
	c_base_header.rtp_pt_1_rnd->marker = rtp->marker;
	// =:= msn_lsb(4)
	c_base_header.rtp_pt_1_rnd->msn = c_msn_lsb(4,v2_context);
	// =:= scaled_ts_lsb(time_stride.UVALUE, 5)
	c_base_header.rtp_pt_1_rnd->ts_scaled = 0; ///               A COMPLETER
	// =:= crc3(THIS.UVALUE,THIS.ULENGTH)
	c_base_header.rtp_pt_1_rnd->header_crc = 0;
	c_base_header.rtp_pt_1_rnd->header_crc =
	   crc_calculate(CRC_TYPE_3,c_base_header.uint8,size,CRC_INIT_3,
	                 context->compressor->crc_table_3);
	goto end_code_rtp_baseheader;

code_rtp_pt_1_seq_id:
	rohc_debugf(3, "code_rtp_pt_1_seq_id\n");
	size = sizeof(rtp_pt_1_seq_id_t);
	// =:= '1001'
	c_base_header.rtp_pt_1_seq_id->discriminator = 0x09;
	// =:= ip_id_lsb(ip_id_behavior_innermost.UVALUE, 4)
	c_base_header.rtp_pt_1_seq_id->ip_id = 0;  // A COMPLETER
	// =:= msn_lsb(5)
	c_base_header.rtp_pt_1_seq_id->msn = c_msn_lsb(5,v2_context);
	// =:= crc3(THIS.UVALUE,THIS.ULENGTH)
	c_base_header.rtp_pt_1_seq_id->header_crc = 0;
	c_base_header.rtp_pt_1_seq_id->header_crc =
	   crc_calculate(CRC_TYPE_3,c_base_header.uint8,size,CRC_INIT_3,
	                 context->compressor->crc_table_3);
	goto end_code_rtp_baseheader;

code_rtp_pt_1_seq_ts:
	rohc_debugf(3, "code_rtp_pt_1_seq_ts\n");
	size = sizeof(rtp_pt_1_seq_ts_t);
	// =:= '101'
	c_base_header.rtp_pt_1_seq_ts->discriminator = 0x05;
	// =:= irregular(1)
	c_base_header.rtp_pt_1_seq_ts->marker = rtp->marker;
	// =:= msn_lsb(4)
	c_base_header.rtp_pt_1_seq_ts->msn = c_msn_lsb(4,v2_context);
	// =:= scaled_ts_lsb(time_stride.UVALUE, 5)
	c_base_header.rtp_pt_1_seq_ts->ts_scaled = 0; ///               A COMPLETER
	// =:= crc3(THIS.UVALUE,THIS.ULENGTH)
	c_base_header.rtp_pt_1_seq_ts->header_crc = 0;
	c_base_header.rtp_pt_1_seq_ts->header_crc =
	   crc_calculate(CRC_TYPE_3,c_base_header.uint8,size,CRC_INIT_3,
	                 context->compressor->crc_table_3);
	goto end_code_rtp_baseheader;

code_rtp_pt_2_rnd:
	rohc_debugf(3, "code_rtp_pt_2_rnd\n");
	size = sizeof(rtp_pt_2_rnd_t);
	// =:= '110'
	c_base_header.rtp_pt_2_rnd->discriminator = 0x06;
	// =:= msn_lsb(7)
	#if __BYTE_ORDER == __LITTLE_ENDIAN
	msn = c_msn_lsb(7,v2_context);
	c_base_header.rtp_pt_2_rnd->msn1 = msn >> 2;
	c_base_header.rtp_pt_2_rnd->msn2 = msn;
	#elif __BYTE_ORDER == __BIG_ENDIAN
	c_base_header.rtp_pt_2_rnd->msn = c_msn_lsb(7,v2_context);
	#endif
	// =:= scaled_ts_lsb(time_stride.UVALUE, 6)
	c_base_header.rtp_pt_2_rnd->ts_scaled = 0; ///               A COMPLETER
	// =:= irregular(1)
	c_base_header.rtp_pt_2_rnd->marker = rtp->marker;
	// =:= crc7(THIS.UVALUE, THIS.ULENGTH)
	c_base_header.rtp_pt_2_rnd->header_crc = 0;
	c_base_header.rtp_pt_2_rnd->header_crc =
	   crc_calculate(CRC_TYPE_7,c_base_header.uint8,size,CRC_INIT_7,
	                 context->compressor->crc_table_7);
	goto end_code_rtp_baseheader;

code_rtp_pt_2_seq_id:
	rohc_debugf(3, "code_rtp_pt_2_seq_id\n");
	size = sizeof(rtp_pt_2_seq_id_t);
	// =:= '11000'
	c_base_header.rtp_pt_2_seq_id->discriminator = 0x18;
	// =:= msn_lsb(7)
	msn = c_msn_lsb(7,v2_context);
	c_base_header.rtp_pt_2_seq_id->msn1 = msn >> 4;
	c_base_header.rtp_pt_2_seq_id->msn2 = msn;
	// =:= ip_id_lsb(ip_id_behavior_innermost.UVALUE, 5)
	ipid = c_ip_id_lsb(ip_inner_context.vx->ip_id_behavior,5,5,ip_inner_context.v4->last_ip_id,ip_id,
	                   v2_context->msn);
	c_base_header.rtp_pt_2_seq_id->ip_id1 = ipid >> 1;
	c_base_header.rtp_pt_2_seq_id->ip_id2 = ipid;
	// =:= crc7(THIS.UVALUE, THIS.ULENGTH)
	c_base_header.rtp_pt_2_seq_id->header_crc = 0;
	c_base_header.rtp_pt_2_seq_id->header_crc =
	   crc_calculate(CRC_TYPE_7,c_base_header.uint8,size,CRC_INIT_7,
	                 context->compressor->crc_table_7);
	goto end_code_rtp_baseheader;

code_rtp_pt_2_seq_both:
	rohc_debugf(3, "code_rtp_pt_2_seq_both\n");
	size = sizeof(rtp_pt_2_seq_both_t);
	// =:= '11001'
	c_base_header.rtp_pt_2_seq_both->discriminator = 0x19;
	// =:= msn_lsb(7)
	msn = c_msn_lsb(7,v2_context);
	c_base_header.rtp_pt_2_seq_both->msn1 = msn >> 4;
	c_base_header.rtp_pt_2_seq_both->msn2 = msn;
	// =:= ip_id_lsb(ip_id_behavior_innermost.UVALUE, 5)
	ipid = c_ip_id_lsb(ip_inner_context.vx->ip_id_behavior,5,5,ip_inner_context.v4->last_ip_id,ip_id,
	                   v2_context->msn);
	c_base_header.rtp_pt_2_seq_both->ip_id1 = ipid >> 1;
	c_base_header.rtp_pt_2_seq_both->ip_id2 = ipid;
	// =:= scaled_ts_lsb(time_stride.UVALUE, 7)
	c_base_header.rtp_pt_2_seq_both->ts_scaled = 0; // .. A COMPLETER
	// =:= irregular(1)
	c_base_header.rtp_pt_2_seq_both->marker = rtp->marker;
	// =:= crc7(THIS.UVALUE, THIS.ULENGTH)
	c_base_header.rtp_pt_2_seq_both->header_crc = 0;
	c_base_header.rtp_pt_2_seq_both->header_crc =
	   crc_calculate(CRC_TYPE_7,c_base_header.uint8,size,CRC_INIT_7,
	                 context->compressor->crc_table_7);
	goto end_code_rtp_baseheader;

code_rtp_pt_2_seq_ts:
	rohc_debugf(3, "code_rtp_pt_2_seq_ts\n");
	size = sizeof(rtp_pt_2_seq_ts_t);
	// =:= '1101'
	c_base_header.rtp_pt_2_seq_ts->discriminator = 0x0D;
	// =:= msn_lsb(7)
	msn = c_msn_lsb(7,v2_context);
	c_base_header.rtp_pt_2_seq_ts->msn1 = msn >> 3;
	c_base_header.rtp_pt_2_seq_ts->msn2 = msn;
	// =:= scaled_ts_lsb(time_stride.UVALUE, 5)
	c_base_header.rtp_pt_2_seq_ts->ts_scaled = 0; // .. A COMPLETER
	// =:= irregular(1)
	c_base_header.rtp_pt_2_seq_ts->marker = rtp->marker;
	// =:= crc7(THIS.UVALUE, THIS.ULENGTH)
	c_base_header.rtp_pt_2_seq_id->header_crc = 0;
	c_base_header.rtp_pt_2_seq_id->header_crc =
	   crc_calculate(CRC_TYPE_7,c_base_header.uint8,size,CRC_INIT_7,
	                 context->compressor->crc_table_7);
	goto end_code_rtp_baseheader;

end_code_rtp_baseheader:
	#if ROHCV2_DEBUG
	rohc_debugf(3, "RTP baseheader, length %d\n",size);
	TraceData(c_base_header.uint8,size);
	#endif

	return c_base_header.uint8 + size;
}


// IP-only profile
// Page 89
u_int8_t * c_iponly_baseheader( multi_ptr_t mptr, base_header_ip_t base_header, int profile_value,
                                int outer_ip_flag, int ip_id_behavior_value,
                                int reorder_ratio_value )
{
	return mptr.uint8;
}


// UDP-lite/RTP profile
// Page 92
u_int8_t * c_udplite_rtp_baseheader( multi_ptr_t mptr, base_header_ip_t base_header,
                                     int profile_value, int ts_stride_value, int time_stride_value,
                                     int outer_ip_flag,
                                     int ip_id_behavior_value, int reorder_ratio_value,
                                     int coverage_behavior_value )
{
	return mptr.uint8;
}


/**
 * @brief Build the static part of the IPv6 option header.
 *
 * @param v2_context     The specific RoHCv2 compression context
 * @param ip_context     The specific IP compression context
 * @param mptr           The current pointer in the rohc-packet-under-build buffer
 * @param protocol       The IPv6 protocol option
 * @param base_header    The IP header
 * @param packet_size    The size of packet
 * @return               The new pointer in the rohc-packet-under-build buffer
 */
u_int8_t * rohc_v2_code_static_ipv6_option_part(struct sc_v2_context *v2_context,
                                                ip_context_ptr_t ip_context,
                                                multi_ptr_t mptr,
                                                u_int8_t protocol,
                                                base_header_ip_t base_header,
                                                const int packet_size)
{
	u_int8_t size;

	rohc_debugf(3, "v2_context %p ip_context %p protocol %d base_header_ip %p\n",v2_context,
	            ip_context.uint8,protocol,
	            base_header.uint8);

	// Common to all options
	mptr.ip_opt_static->next_header = base_header.ipv6_opt->next_header;

	switch(protocol)
	{
		case IPPROTO_HOPOPTS:  // IPv6 Hop-by-Hop options
			mptr.ip_hop_opt_static->length = base_header.ipv6_opt->length;
			size = sizeof(ip_hop_opt_static_t);
			break;
		case IPPROTO_ROUTING:  // IPv6 routing header
			mptr.ip_hop_opt_static->length = base_header.ipv6_opt->length;
			size = (base_header.ipv6_opt->length + 1) << 3;
			memcpy(mptr.ip_rout_opt_static->value,base_header.ipv6_opt->value,size - 2);
			break;
		case IPPROTO_GRE:
			if(ntohs(base_header.ip_gre_opt->protocol) == 0x0800)
			{
				mptr.ip_gre_opt_static->protocol = 0;
			}
			else
			{
				assert( ntohs(base_header.ip_gre_opt->protocol) == 0x86DD );
				mptr.ip_gre_opt_static->protocol = 1;
			}
			mptr.ip_gre_opt_static->c_flag = base_header.ip_gre_opt->c_flag;
			mptr.ip_gre_opt_static->s_flag = base_header.ip_gre_opt->s_flag;
			mptr.ip_gre_opt_static->padding = 0;
			if( ( mptr.ip_gre_opt_static->k_flag = base_header.ip_gre_opt->k_flag ) != 0)
			{
				mptr.ip_gre_opt_static->key =
				   base_header.ip_gre_opt->datas[base_header.ip_gre_opt->c_flag];
				size = sizeof(ip_gre_opt_static_t);
			}
			else
			{
				size = sizeof(ip_gre_opt_static_t) - sizeof(u_int32_t);
			}
			break;
		case IPPROTO_DSTOPTS:  // IPv6 destination options
			mptr.ip_dest_opt_static->length = base_header.ipv6_opt->length;
			size = sizeof(ip_dest_opt_static_t);
			break;
		case IPPROTO_MIME:
			mptr.ip_mime_opt_static->s_bit = base_header.ip_mime_opt->s_bit;
			mptr.ip_mime_opt_static->res_bits = base_header.ip_mime_opt->res_bits;
			mptr.ip_mime_opt_static->orig_dest = base_header.ip_mime_opt->orig_dest;
			if(base_header.ip_mime_opt->s_bit != 0)
			{
				mptr.ip_mime_opt_static->orig_src = base_header.ip_mime_opt->orig_src;
				size = sizeof(ip_mime_opt_static_t);
				break;
			}
			size = sizeof(ip_mime_opt_static_t) - sizeof(u_int32_t);
			break;
		case IPPROTO_AH:
			mptr.ip_ah_opt_static->length = base_header.ip_ah_opt->length;
			mptr.ip_ah_opt_static->spi = base_header.ip_ah_opt->spi;
			size = sizeof(ip_ah_opt_static_t);
			break;
		default:
			size = 0;
			break;
	}

	#if ROHCV2_DEBUG
	rohc_debugf(3, "IPv6 option %d static part, length %d\n",base_header.ipv6_opt->next_header,size);
	TraceData(mptr.uint8,size);
	#endif
	fflush(stdout);
	return mptr.uint8 + size;
}


/**
 * @brief Build the dynamic part of the IPv6 option header.
 *
 * @param v2_context     The specific RoHCv2 compression context
 * @param ip_context     The specific IP compression context
 * @param mptr           The current pointer in the rohc-packet-under-build buffer
 * @param protocol       The IPv6 protocol option
 * @param base_header    The IP header
 * @param packet_size    The size of packet
 * @return               The new pointer in the rohc-packet-under-build buffer
 */
u_int8_t * rohc_v2_code_dynamic_ipv6_option_part(struct sc_v2_context *v2_context,
                                                 ip_context_ptr_t ip_context,
                                                 multi_ptr_t mptr,
                                                 u_int8_t protocol,
                                                 base_header_ip_t base_header,
                                                 const int packet_size)
{
	int size;

	rohc_debugf(3, "v2_context %p ip_context %p protocol %d base_header %p\n",v2_context,
	            ip_context.uint8,protocol,
	            base_header.uint8);

	switch(protocol)
	{
		case IPPROTO_HOPOPTS:  // IPv6 Hop-by-Hop options
		case IPPROTO_DSTOPTS:  // IPv6 destination options
			size = ( (base_header.ipv6_opt->length + 1) << 3 ) - 2;
			memcpy(ip_context.v6_option->value,base_header.ipv6_opt->value,size);
			memcpy(mptr.ip_opt_dynamic->value,base_header.ipv6_opt->value,size);
			break;
		case IPPROTO_ROUTING:  // IPv6 routing header
			size = 0;
			break;
		case IPPROTO_GRE:
			size = 0;
			// checksum_and_res =:= optional_checksum(c_flag.UVALUE)
			if(base_header.ip_gre_opt->c_flag != 0)
			{
				u_int8_t *ptr = (u_int8_t*) base_header.ip_gre_opt->datas;
				*(mptr.uint8++) = *ptr++;
				*(mptr.uint8++) = *ptr;
				size += sizeof(u_int16_t);
			}
			// sequence_number =:= optional_32(s_flag.UVALUE)
			if(base_header.ip_gre_opt->s_flag != 0)
			{
				ip_context.v6_gre_option->sequence_number =
				   base_header.ip_gre_opt->datas[base_header.ip_gre_opt->c_flag];
				WRITE32_TO_MPTR(mptr,base_header.ip_gre_opt->datas[base_header.ip_gre_opt->c_flag]);
				size += sizeof(u_int32_t);
			}
			mptr.uint8 -= size;
			break;
		case IPPROTO_MIME:
			size = 0;
			break;
		case IPPROTO_AH:
			ip_context.v6_ah_option->sequence_number = ntohl(base_header.ip_ah_opt->sequence_number);
			mptr.ip_ah_opt_dynamic->sequence_number = base_header.ip_ah_opt->sequence_number;
			size = (base_header.ip_ah_opt->length - 1) << 2;
			memcpy(mptr.ip_ah_opt_dynamic->icv,base_header.ip_ah_opt->icv,
			       (base_header.ip_ah_opt->length - 1) << 2);
			size += sizeof(u_int32_t);
			break;
		default:
			size = 0;
			break;
	}

	#if ROHCV2_DEBUG
	rohc_debugf(3, "IPv6 option %d dynamic part, length %d\n",base_header.ipv6_opt->next_header,size);
	TraceData(mptr.uint8,size);
	#endif
	fflush(stdout);
	return mptr.uint8 + size;
}


/**
 * @brief Build the irregular part of the IPv6 option header.
 *
 * @param v2_context     The specific RoHCv2 compression context
 * @param ip_context     The specific IP compression context
 * @param mptr           The current pointer in the rohc-packet-under-build buffer
 * @param protocol       The IPv6 protocol option
 * @param base_header    The IP header
 * @param packet_size    The size of packet
 * @return               The new pointer in the rohc-packet-under-build buffer
 */
u_int8_t * rohc_v2_code_irregular_ipv6_option_part(struct sc_v2_context *v2_context,
                                                   ip_context_ptr_t ip_context,
                                                   multi_ptr_t mptr,
                                                   u_int8_t protocol,
                                                   base_header_ip_t base_header,
                                                   const int packet_size)
{
	#if ROHCV2_DEBUG
	u_int8_t *ptr = mptr.uint8;
	#endif
	u_int32_t sequence_number;
	int size;

	rohc_debugf(3, "v2_context %p ip_context %p protocol %d base_header_ip %p\n",v2_context,
	            ip_context.uint8,protocol,
	            base_header.uint8);

	switch(protocol)
	{
		case IPPROTO_GRE:
			// checksum_and_res =:= optional_checksum(c_flag.UVALUE)
			if(base_header.ip_gre_opt->c_flag != 0)
			{
				u_int8_t *ptr = (u_int8_t*) base_header.ip_gre_opt->datas;
				*(mptr.uint8++) = *ptr++;
				*(mptr.uint8++) = *ptr;
			}
			// sequence_number =:= optional_lsb_7_or_31(s_flag.UVALUE)
			if(base_header.ip_gre_opt->s_flag != 0)
			{
				sequence_number = ntohl(base_header.ip_gre_opt->datas[base_header.ip_gre_opt->c_flag]);
				if( ( sequence_number & 0xFFFFFF80 ) ==
				    ( ip_context.v6_gre_option->sequence_number & 0xFFFFFF80 ) )
				{
					// discriminator =:= '0'
					*(mptr.uint8++) = sequence_number & 0x7F;
				}
				else
				{
					// discriminator =:= '1'
					WRITE32_TO_MPTR(mptr,htonl(0x80000000 | sequence_number));
				}
				ip_context.v6_gre_option->sequence_number =
				   base_header.ip_gre_opt->datas[base_header.ip_gre_opt->c_flag];
			}
			break;
		case IPPROTO_AH:
			sequence_number = ntohl(base_header.ip_ah_opt->sequence_number);
			if( ( sequence_number & 0xFFFFFF80 ) ==
			    ( ip_context.v6_ah_option->sequence_number & 0xFFFFFF80 ) )
			{
				// discriminator =:= '0'
				*(mptr.uint8++) = sequence_number & 0x7F;
			}
			else
			{
				// discriminator =:= '1'
				WRITE32_TO_MPTR(mptr,htonl(0x80000000 | sequence_number));
			}
			ip_context.v6_ah_option->sequence_number = sequence_number;
			size = (base_header.ip_ah_opt->length - 1) << 3;
			memcpy(mptr.uint8,base_header.ip_ah_opt->icv,size);
			mptr.uint8 += size;
			break;
		default:
			break;
	}

	#if ROHCV2_DEBUG
	rohc_debugf(3, "IPv6 option %d irregular part, length %d\n",base_header.ipv6_opt->next_header,
	            (int)(mptr.uint8 - ptr));
	TraceData(ptr,mptr.uint8 - ptr);
	#endif
	fflush(stdout);
	return mptr.uint8;
}


/**
 * @brief Build the static part of the IP header.
 *
 * @param v2_context     The specific RoHCv2 compression context
 * @param ip_context     The specific IP compression context
 * @param mptr           The current pointer in the rohc-packet-under-build buffer
 * @param base_header    The IP header
 * @param packet_size    The size of packet
 * @param is_innermost   True if IP header is the innermost of the packet
 * @return               The new pointer in the rohc-packet-under-build buffer
 */
u_int8_t * rohc_v2_code_static_ip_part(struct sc_v2_context *v2_context,
                                       ip_context_ptr_t ip_context,
                                       multi_ptr_t mptr,
                                       base_header_ip_t base_header,
                                       const int packet_size,
                                       int is_innermost)
{
	int size;

	rohc_debugf(3, "v2_context %p ip_context %p base_header %p\n",v2_context,ip_context.uint8,
	            base_header.uint8);

	if(base_header.ipvx->version == IPV4)
	{
		mptr.ipv4_static->version_flag = 0;
		mptr.ipv4_static->innermost_ip = is_innermost;
		mptr.ipv4_static->reserved = 0;
		mptr.ipv4_static->protocol = base_header.ipv4->protocol;
		rohc_debugf(3, "protocol %d\n",mptr.ipv4_static->protocol);
		mptr.ipv4_static->src_addr = base_header.ipv4->src_addr;
		mptr.ipv4_static->dst_addr = base_header.ipv4->dest_addr;
		size = sizeof(ipv4_static_t);
	}
	else
	{
		if(base_header.ipv6->flow_label1 == 0 && base_header.ipv6->flow_label2 == 0)
		{
			mptr.ipv6_static1->version_flag = 1;
			mptr.ipv6_static1->innermost_ip = is_innermost;
			mptr.ipv6_static1->flow_label_enc_discriminator = 0;
			mptr.ipv6_static1->reserved = 0;
			mptr.ipv6_static1->next_header = base_header.ipv6->next_header;
			memcpy(mptr.ipv6_static1->src_addr,base_header.ipv6->src_addr,sizeof(u_int32_t) * 4 * 2);
			size = sizeof(ipv6_static1_t);
		}
		else
		{
			mptr.ipv6_static2->version_flag = 1;
			mptr.ipv6_static1->innermost_ip = is_innermost;
			mptr.ipv6_static2->reserved = 0;
			mptr.ipv6_static2->flow_label_enc_discriminator = 1;
			mptr.ipv6_static2->flow_label1 = base_header.ipv6->flow_label1;
			mptr.ipv6_static2->flow_label2 = base_header.ipv6->flow_label2;
			mptr.ipv6_static2->next_header = base_header.ipv6->next_header;
			memcpy(mptr.ipv6_static2->src_addr,base_header.ipv6->src_addr,sizeof(u_int32_t) * 4 * 2);
			size = sizeof(ipv6_static2_t);
		}
		rohc_debugf(3, "next_header %d\n",base_header.ipv6->next_header);
	}

	#if ROHCV2_DEBUG
	rohc_debugf(3, "IP v%d static part, length %d\n",base_header.ipvx->version,size);
	TraceData(mptr.uint8,size);
	#endif

	return mptr.uint8 + size;
}


/**
 * @brief Build the dynamic part of the IP header.
 *
 * @param v2_context        The specific RoHCv2 compression context
 * @param ip_inner_context  The specific IP compression context of the innermost IP header
 * @param ip_context        The specific IP compression context
 * @param mptr              The current pointer in the rohc-packet-under-build buffer
 * @param base_header       The IP header
 * @param packet_size       The size of packet
 * @param profile_value     The profile ID
 * @return                  The new pointer in the rohc-packet-under-build buffer
 */
u_int8_t * rohc_v2_code_dynamic_ip_part(struct sc_v2_context *v2_context,
                                        ip_context_ptr_t ip_inner_context,
                                        ip_context_ptr_t ip_context,
                                        multi_ptr_t mptr,
                                        base_header_ip_t base_header,
                                        const int packet_size,
                                        int profile_value)
{
//	struct c_generic_context *g_context;
//	struct sc_v2_context *v2_context;
	int size;

//	g_context = (struct c_generic_context *) context->specific;
//	v2_context = (struct sc_v2_context *) g_context->specific;

//	rohc_debugf(3, "context %p ip_inner_context %p ip_context %p base_header %p\n",context,ip_inner_context.uint8,ip_context.uint8,base_header.uint8);
	rohc_debugf(3,
	            "v2_context %p ip_inner_context %p ip_context %p base_header %p profile_value %Xh\n",
	            v2_context,
	            ip_inner_context.uint8,ip_context.uint8,base_header.uint8,
	            profile_value);
	rohc_debugf(3, "ip_context %p version %d\n",ip_context.uint8,ip_context.v6->version);
	fflush(stdout);
	if(base_header.ipvx->version == IPV4)
	{
		assert( ip_context.v4->version == IPV4 );

		// If is innermost
		if(ip_context.uint8 == ip_inner_context.uint8)
		{
			rohc_debugf(3, "ip_context %p ip_id_behavior %d\n",ip_context.uint8,
			            ip_context.v4->ip_id_behavior);
			if(profile_value == PROFILE_IP_0104)
			{
				mptr.ipv4_endpoint_innermost_dynamic1->reserved = 0;
				mptr.ipv4_endpoint_innermost_dynamic1->reorder_ratio = v2_context->reorder_ratio;
				mptr.ipv4_endpoint_innermost_dynamic1->df = base_header.ipv4->df;
				mptr.ipv4_endpoint_innermost_dynamic1->ip_id_behavior_innermost =
				   ip_context.v4->ip_id_behavior;
				mptr.ipv4_endpoint_innermost_dynamic1->tos_tc = base_header.ipv4->tos_tc;
				mptr.ipv4_endpoint_innermost_dynamic1->ttl_hopl = base_header.ipv4->ttl_hopl;
				if(ip_context.v4->ip_id_behavior == IP_ID_BEHAVIOR_ZERO)
				{
					mptr.ipv4_endpoint_innermost_dynamic1->msn = htons(v2_context->msn);
					size = sizeof(ipv4_endpoint_innermost_dynamic1_t);
				}
				else
				{
					// ip_id =:= ip_id_enc_dyn(behavior)
					mptr.ipv4_endpoint_innermost_dynamic2->ip_id = base_header.ipv4->ip_id;
					mptr.ipv4_endpoint_innermost_dynamic2->msn = htons(v2_context->msn);
					size = sizeof(ipv4_endpoint_innermost_dynamic2_t);
				}
			}
			else
			{
				mptr.ipv4_regular_innermost_dynamic1->reserved = 0;
				mptr.ipv4_regular_innermost_dynamic1->df = base_header.ipv4->df;
				mptr.ipv4_regular_innermost_dynamic1->ip_id_behavior_innermost =
				   ip_context.v4->ip_id_behavior;
				mptr.ipv4_regular_innermost_dynamic1->tos_tc = base_header.ipv4->tos_tc;
				mptr.ipv4_regular_innermost_dynamic1->ttl_hopl = base_header.ipv4->ttl_hopl;
				if(ip_context.v4->ip_id_behavior == IP_ID_BEHAVIOR_ZERO)
				{
					size = sizeof(ipv4_regular_innermost_dynamic1_t);
				}
				else
				{
					// ip_id =:= ip_id_enc_dyn(behavior)
					mptr.ipv4_regular_innermost_dynamic2->ip_id = base_header.ipv4->ip_id;
					size = sizeof(ipv4_regular_innermost_dynamic2_t);
				}
			}
		}
		else
		{
			// IP outer dynamic
			mptr.ipv4_outer_dynamic1->reserved = 0;
			mptr.ipv4_outer_dynamic1->df = base_header.ipv4->df;
			mptr.ipv4_outer_dynamic1->ip_id_behavior_outer = ip_context.v4->ip_id_behavior;
			mptr.ipv4_outer_dynamic1->tos_tc = base_header.ipv4->tos_tc;
			mptr.ipv4_outer_dynamic1->ttl_hopl = base_header.ipv4->ttl_hopl;
			if(ip_context.v4->ip_id_behavior == IP_ID_BEHAVIOR_ZERO)
			{
				size = sizeof(ipv4_outer_dynamic1_t);
			}
			else
			{
				// ip_id =:= ip_id_enc_dyn(behavior)
				mptr.ipv4_outer_dynamic2->ip_id = base_header.ipv4->ip_id;
				size = sizeof(ipv4_outer_dynamic2_t);
			}
		}
		// Save values in context
		ip_context.v4->tos_tc = base_header.ipv4->tos_tc;
		ip_context.v4->ttl_hopl = base_header.ipv4->ttl_hopl;
		ip_context.v4->df = base_header.ipv4->df;
		ip_context.v4->last_ip_id.uint16 = ntohs(base_header.ipv4->ip_id);
	}
	else
	{
		rohc_debugf(3, "ip_context %p version %d\n",ip_context.uint8,ip_context.v6->version);
		fflush(stdout);
		assert( ip_context.v6->version == IPV6 );
		rohc_debugf(3, "---\n");
		fflush(stdout);

		// If is innermost
		if(ip_context.uint8 == ip_inner_context.uint8 &&
		   profile_value == PROFILE_IP_0104)
		{
			#if __BYTE_ORDER == __LITTLE_ENDIAN
			mptr.ipv6_endpoint_dynamic->tos_tc =
			   ( base_header.ipv6->tos_tc1 << 4 ) | base_header.ipv6->tos_tc2;
			#elif __BYTE_ORDER == __BIG_ENDIAN
			mptr.ipv6_endpoint_dynamic->tos_tc = base_header.ipv6->tos_tc;
			#endif
			mptr.ipv6_endpoint_dynamic->ttl_hopl = base_header.ipv6->ttl_hopl;
			mptr.ipv6_endpoint_dynamic->reserved = 0;
			mptr.ipv6_endpoint_dynamic->reorder_ratio = v2_context->reorder_ratio;
			mptr.ipv6_endpoint_dynamic->msn = htons(v2_context->msn);
			size = sizeof(ipv6_endpoint_dynamic_t);
		}
		else
		{
			#if __BYTE_ORDER == __LITTLE_ENDIAN
			mptr.ipv6_regular_dynamic->tos_tc =
			   ( base_header.ipv6->tos_tc1 << 4 ) | base_header.ipv6->tos_tc2;
			#elif __BYTE_ORDER == __BIG_ENDIAN
			mptr.ipv6_regular_dynamic->tos_tc = base_header.ipv6->tos_tc;
			#endif
			mptr.ipv6_regular_dynamic->ttl_hopl = base_header.ipv6->ttl_hopl;
			size = sizeof(ipv6_regular_dynamic_t);
		}

		// Save values in context
		#if __BYTE_ORDER == __LITTLE_ENDIAN
		ip_context.v6->tos_tc1 = base_header.ipv6->tos_tc1;
		ip_context.v6->tos_tc2 = base_header.ipv6->tos_tc2;
		#elif __BYTE_ORDER == __BIG_ENDIAN
		ip_context.v6->tos_tc = base_header.ipv6->tos_tc;
		#endif
		ip_context.v6->ttl_hopl = base_header.ipv6->ttl_hopl;
	}

	#if ROHCV2_DEBUG
	rohc_debugf(3, "IP v%d dynamic part, length %d\n",base_header.ipvx->version,size);
	TraceData(mptr.uint8,size);
	#endif
	fflush(stdout);

	return mptr.uint8 + size;
}


/**
 * @brief Build the irregular part of the IP header.
 *
 * See RFC 5225 page 59 and 62
 *
 * @param ip_context     The specific IP compression context
 * @param mptr           The current pointer in the rohc-packet-under-build buffer
 * @param base_header    The IP header
 * @param packet_size    The size of packet
 * @param is_innermost   True if IP header is the innermost of the packet
 * @param outer_ip_flag  True if TOS/TC or TTL/Hop Limit in outer IP headers has changed
 * @return               The new pointer in the rohc-packet-under-build buffer
 */
u_int8_t * rohc_v2_code_irregular_ip_part(ip_context_ptr_t ip_context,
                                          multi_ptr_t mptr,
                                          base_header_ip_t base_header,
                                          const int packet_size,
                                          int is_innermost,
                                          int outer_ip_flag)
{
	#if ROHCV2_DEBUG
	u_int8_t *ptr = mptr.uint8;
	#endif

	rohc_debugf(3, "ip_context %p base_header %p is_innermost %d outer_ip_flag %d\n",
	            ip_context.uint8,base_header.uint8,is_innermost,
	            outer_ip_flag);
	rohc_debugf(3, "IP version %d ip_id_behavior %d\n",base_header.ipvx->version,
	            ip_context.v4->ip_id_behavior);

	if(base_header.ipvx->version == IPV4)
	{
		if(is_innermost == 0)
		{
			// ipv4_outer_irregular
			// ip_id =:= ip_id_enc_irreg(ip_id_behavior_innermost.UVALUE)
			#ifdef USE_ROHC_V2_MACROS
			c_ip_id_enc_irreg(mptr,ip_context.v4->ip_id_behavior,base_header.ipv4->ip_id);
			#else
			mptr.uint8 = c_ip_id_enc_irreg(mptr,ip_context.v4->ip_id_behavior,base_header.ipv4->ip_id);
			#endif
			if(outer_ip_flag != 0)
			{
				// tos_tc =:= static_or_irreg(outer_ip_flag, 8)
				*(mptr.uint8++) = base_header.ipv4->tos_tc;
				// ttl_hopl =:= static_or_irreg(outer_ip_flag, 8)
				*(mptr.uint8++) = base_header.ipv4->ttl_hopl;
				rohc_debugf(3, "IPv4 tos and ttl_hopl saved\n");
				// Save values in context
				ip_context.v4->tos_tc = base_header.ipv4->tos_tc;
				ip_context.v4->ttl_hopl = base_header.ipv4->ttl_hopl;
			}
		}
		else
		{
			// ipv4_innermost_irregular
			// ip_id =:= ip_id_enc_irreg(ip_id_behavior_innermost.UVALUE)
			#ifdef USE_ROHC_V2_MACROS
			c_ip_id_enc_irreg(mptr,ip_context.v4->ip_id_behavior,base_header.ipv4->ip_id);
			#else
			mptr.uint8 = c_ip_id_enc_irreg(mptr,ip_context.v4->ip_id_behavior,base_header.ipv4->ip_id);
			#endif
		}
		// Save values in context
//		ip_context.v4->tos_tc = base_header.ipv4->tos_tc;
//		ip_context.v4->ttl_hopl = base_header.ipv4->ttl_hopl;
//		ip_context.v4->df = base_header.ipv4->df;
		ip_context.v4->last_ip_id.uint16 = ntohs(base_header.ipv4->ip_id);
	}
	else
	{
		// IPv6
		if(is_innermost == 0)
		{
			// ipv6_outer_irregular
			if(outer_ip_flag != 0)
			{
				// tos_tc =:= static_or_irreg(outer_ip_flag,8)
				#if __BYTE_ORDER == __LITTLE_ENDIAN
				*(mptr.uint8++) = ( base_header.ipv6->tos_tc1 << 4 ) | base_header.ipv6->tos_tc2;
				#elif __BYTE_ORDER == __BIG_ENDIAN
				*(mptr.uint8++) = base_header.ipv6->tos_tc;
				#endif
				// ttl_hopl =:= static_or_irreg(outer_ip_flag,8)
				*(mptr.uint8++) = base_header.ipv6->ttl_hopl;
				rohc_debugf(3, "IPv6 tos and ttl_hopl saved\n");
				// Save values in context
				#if __BYTE_ORDER == __LITTLE_ENDIAN
				ip_context.v6->tos_tc1 = base_header.ipv6->tos_tc1;
				ip_context.v6->tos_tc2 = base_header.ipv6->tos_tc2;
				#elif __BYTE_ORDER == __BIG_ENDIAN
				ip_context.v6->tos_tc = base_header.ipv6->tos_tc;
				#endif
				ip_context.v6->ttl_hopl = base_header.ipv6->ttl_hopl;
			}
		}
		else
		{
			// ipv6_innermost_irregular
		}
#ifdef LMJKQSLMQJKs
		// Save values in context
		#if __BYTE_ORDER == __LITTLE_ENDIAN
		ip_context.v6->tos_tc1 = base_header.ipv6->tos_tc1;
		ip_context.v6->tos_tc2 = base_header.ipv6->tos_tc2;
		#elif __BYTE_ORDER == __BIG_ENDIAN
		ip_context.v6->tos_tc = base_header.ipv6->tos_tc;
		#endif
		ip_context.v6->ttl_hopl = base_header.ipv6->ttl_hopl;
#endif
	}

#if ROHCV2_DEBUG
	rohc_debugf(3, "IP v%d irregular part length %d\n",ip_context.vx->version,(int)(mptr.uint8 - ptr));
	TraceData(ptr,mptr.uint8 - ptr);
#endif
	return mptr.uint8;
}


/**
 * @brief Compress the IP-only header
 *
 * See RFC5225 page 89
 *
 * @param context                   The compression context
 * @param v2_context                The specific RoHCv2 context
 * @param ip_inner_context          The specific IP innermost context
 * @param c_base_header             The compressed header
 * @param base_header_inner         The innermost IP header
 * @param outer_ip_flag             The IP flag of the outer IP header
 * @return                          The new pointer in the rohc-packet-under-build buffer
 */


u_int8_t * c_ip_baseheader(struct c_context *const context,
                           struct sc_v2_context *v2_context,
                           ip_context_ptr_t ip_inner_context,
                           multi_ptr_t c_base_header,
                           base_header_ip_t base_header_inner,
                           int outer_ip_flag )
{
	multi_ptr_t mptr;
	u_int16_t old_msn;
	WB_t ip_id;
	#if __BYTE_ORDER == __LITTLE_ENDIAN
	u_int8_t msn;
	u_int8_t ipid;
	#endif
	int size;

	rohc_debugf(3, "context %p v2_context %p ip_inner_context %p\n",context,v2_context,
	            ip_inner_context.uint8);
	rohc_debugf(3, "c_base_header %p base_header_inner %p outer_ip_flag %d\n",c_base_header.uint8,
	            base_header_inner.uint8,
	            outer_ip_flag);

	fflush(stdout);

	if(base_header_inner.ipvx->version == IPV4)
	{
		rohc_debugf(3, "ip_inner_context.v4->df %d base_header_inner.ipv4->df %d\n",
		            ip_inner_context.v4->df,
		            base_header_inner.ipv4->df);

		ip_id.uint16 = ntohs( base_header_inner.ipv4->ip_id );
		if(outer_ip_flag != 0)
		{
			rohc_debugf(3, "outer_ip_flag != 0\n");
			goto code_ip_co_common;
		}
		if(ip_inner_context.v4->df != base_header_inner.ipv4->df ||
		   ip_inner_context.v4->tos_tc != base_header_inner.ipv4->tos_tc ||
		   ip_inner_context.v4->ttl_hopl != base_header_inner.ipv4->ttl_hopl)
		{
			rohc_debugf(3, "df %d %d tos_tc %Xh %Xh ttl_hopl %Xh %Xh\n",
			            ip_inner_context.v4->df,base_header_inner.ipv4->df,
			            ip_inner_context.v4->tos_tc,base_header_inner.ipv4->tos_tc,
			            ip_inner_context.v4->ttl_hopl,base_header_inner.ipv4->ttl_hopl);
			goto code_ip_co_common;
		}
		if(ip_inner_context.v4->ip_id_behavior != ip_inner_context.v4->last_ip_id_behavior)
		{
			rohc_debugf(3, "ip_id_behavior %d %d\n",
			            ip_inner_context.v4->ip_id_behavior,ip_inner_context.v4->last_ip_id_behavior);
			goto code_ip_co_common;
		}
		switch(ip_inner_context.v4->ip_id_behavior)
		{
			case IP_ID_BEHAVIOR_SEQUENTIAL:
				// If only 6 bits change
				if( ( ip_inner_context.v4->last_ip_id.uint16 & 0xFFC0 ) == ( ip_id.uint16 & 0xFFC0 ) )
				{
					goto code_ip_pt_1_seq_id;
				}
				break;
				// If only 8 bits change
				if( ( ip_inner_context.v4->last_ip_id.uint16 & 0xFF00 ) == ( ip_id.uint16 & 0xFF00 ) )
				{
					goto code_ip_pt_2_seq_id;
				}
				break;
			case IP_ID_BEHAVIOR_SEQUENTIAL_SWAPPED:
				// If only 6 bits change
				if( ( ip_inner_context.v4->last_ip_id.uint16 & 0xC0FF ) == ( ip_id.uint16 & 0xC0FF ) )
				{
					goto code_ip_pt_1_seq_id;
				}
				break;
				// If only 8 bits change
				if( ( ip_inner_context.v4->last_ip_id.uint16 & 0x00FF ) == ( ip_id.uint16 & 0x00FF ) )
				{
					goto code_ip_pt_2_seq_id;
				}
				break;
			default:
				goto code_ip_co_common;
		}
	}
	else
	{
		if(outer_ip_flag != 0)
		{
			rohc_debugf(3, "outer_ip_flag != 0\n");
			goto code_ip_co_common;
		}
		#if __BYTE_ORDER == __LITTLE_ENDIAN
		if(ip_inner_context.v6->tos_tc1 != base_header_inner.ipv6->tos_tc1 ||
		   ip_inner_context.v6->tos_tc2 != base_header_inner.ipv6->tos_tc2)
		#elif __BYTE_ORDER == __BIG_ENDIAN
		if(ip_inner_context.v6->tos_tc != base_header_inner.ipv6->tos_tc)
		#endif
		{
			rohc_debugf(3, "tos_tc %Xh %Xh\n",
							#if __BYTE_ORDER == __LITTLE_ENDIAN
			            (ip_inner_context.v6->tos_tc1 << 4) | ip_inner_context.v6->tos_tc2,
			            (base_header_inner.ipv6->tos_tc1 << 4) | base_header_inner.ipv6->tos_tc2);
			#elif __BYTE_ORDER == __BIG_ENDIAN
			            ip_inner_context.v6->tos_tc,base_header_inner.ipv6->tos_tc);
			#endif
			goto code_ip_co_common;
		}
		if(ip_inner_context.v6->ttl_hopl != base_header_inner.ipv6->ttl_hopl)
		{
			rohc_debugf(3, "tos_tc %Xh %Xh\n",ip_inner_context.v6->ttl_hopl,
			            base_header_inner.ipv6->ttl_hopl);
			goto code_ip_co_common;
		}
	}

	old_msn = v2_context->msn - 1;
	rohc_debugf(3, "old msn %Xh -> %Xh\n",old_msn,v2_context->msn);

	if( ( old_msn & 0xFFF0 ) == ( v2_context->msn & 0xFFF0 ) )
	{
		goto code_ip_pt_0_crc3;
	}

	if( ( old_msn & 0xFFC0 ) == ( v2_context->msn & 0xFFC0 ) )
	{
		goto code_ip_pt_0_crc7;
	}

code_ip_co_common:
	rohc_debugf(3, "code ip_co_common\n");
	mptr.ip_co_common = c_base_header.ip_co_common + 1;
	// =:= '11111010'
	c_base_header.ip_co_common->discriminator = 0xFA;
	// =:= irregular(2)
	c_base_header.ip_co_common->reorder_ratio = v2_context->reorder_ratio;
	// =:= control_crc3_encoding
	c_base_header.ip_co_common->control_crc3 = 0;
	if(base_header_inner.ipvx->version == IPV4)
	{
		// =:= profile_2_3_4_flags_enc(flags_indicator.CVALUE, ip_version.UVALUE)
		if(outer_ip_flag != 0 ||
		   base_header_inner.ipv4->df != ip_inner_context.v4->df ||
		   ip_inner_context.v4->ip_id_behavior != ip_inner_context.v4->last_ip_id_behavior)
		{
			// =:= irregular(1)
			c_base_header.ip_co_common->flags_indicator = 1;
			// See RFC5225 page 76
			// =:= irregular(1)
			mptr.profile_2_3_4_flags->name.ip_outer_indicator = outer_ip_flag;
			// =:= dont_fragment(ip_version)
			mptr.profile_2_3_4_flags->name.df = base_header_inner.ipv4->df;
			// =:= irregular(2)
			mptr.profile_2_3_4_flags->name.ip_id_behavior = ip_inner_context.v4->ip_id_behavior;
			// =:= compressed_value(4,0)
			mptr.profile_2_3_4_flags->name.reserved = 0;
			++mptr.profile_2_3_4_flags;

			ip_inner_context.v4->df = base_header_inner.ipv4->df;
			ip_inner_context.v4->last_ip_id_behavior = ip_inner_context.v4->ip_id_behavior;
		}
		else
		{
			// =:= irregular(1)
			c_base_header.ip_co_common->flags_indicator = 0;
		}
		c_base_header.ip_co_common->tos_tc_indicator = c2_static_or_irreg8(
		   &mptr,&ip_inner_context.v4->tos_tc,base_header_inner.ipv4->tos_tc);
		// =:= static_or_irreg(ttl_hopl_indicator.CVALUE, 8)
		c_base_header.ip_co_common->ttl_hopl_indicator = c2_static_or_irreg8(
		   &mptr,&ip_inner_context.v4->ttl_hopl,base_header_inner.ipv4->ttl_hopl);
		// =:= msn_lsb(8)
		*(mptr.uint8++) = c_msn_lsb(8,v2_context);
		// =:= ip_id_sequential_variable(ip_id_behavior_innermost.UVALUE,ip_id_indicator.CVALUE)
		c_base_header.ip_co_common->ip_id_indicator = c_ip_id_sequential_variable(
		   &mptr,ip_inner_context.v4->ip_id_behavior,ip_inner_context.v4->last_ip_id,ip_id,
		   v2_context->msn);
	}
	else
	{
		// =:= irregular(1)
		if( ( c_base_header.ip_co_common->flags_indicator = outer_ip_flag ) != 0)
		{
			// See RFC5225 page 76
			// =:= irregular(1)
			mptr.profile_2_3_4_flags->name.ip_outer_indicator = outer_ip_flag;
			// =:= dont_fragment(ip_version)
			mptr.profile_2_3_4_flags->name.df = 0;
			// =:= irregular(2)
			mptr.profile_2_3_4_flags->name.ip_id_behavior = ip_inner_context.v6->ip_id_behavior;
			// =:= compressed_value(4,0)
			mptr.profile_2_3_4_flags->name.reserved = 0;
			++mptr.profile_2_3_4_flags;
		}
		// =:= static_or_irreg(tos_tc_indicator.CVALUE, 8)
		#if __BYTE_ORDER == __LITTLE_ENDIAN
		if(ip_inner_context.v6->tos_tc1 != base_header_inner.ipv6->tos_tc1 ||
		   ip_inner_context.v6->tos_tc2 != base_header_inner.ipv6->tos_tc2)
		{
			c_base_header.ip_co_common->tos_tc_indicator = 1;
			*(mptr.uint8++) = (base_header_inner.ipv6->tos_tc1 << 4) | base_header_inner.ipv6->tos_tc2;
			ip_inner_context.v6->tos_tc1 = base_header_inner.ipv6->tos_tc1;
			ip_inner_context.v6->tos_tc2 = base_header_inner.ipv6->tos_tc2;
		}
		else
		{
			c_base_header.ip_co_common->tos_tc_indicator = 0;
		}
//		c_base_header.ip_co_common->tos_tc_indicator = c2_static_or_irreg8(&mptr,(ip_inner_context.v6->tos_tc1<<4)|ip_inner_context.v6->tos_tc2,(base_header_inner.ipv6->tos_tc1<<4)|base_header_inner.ipv6->tos_tc2);
		#elif __BYTE_ORDER == __BIG_ENDIAN
		c_base_headerip_co_common->tos_tc_indicator = c2_static_or_irreg8(
		   &mptr,&ip_inner_context.v6->tos_tc,base_header_inner.ipv6->tos_tc);
		#endif
		// =:= static_or_irreg(ttl_hopl_indicator.CVALUE, 8)
		c_base_header.ip_co_common->ttl_hopl_indicator = c2_static_or_irreg8(
		   &mptr,&ip_inner_context.v6->ttl_hopl,base_header_inner.ipv6->ttl_hopl);
		// =:= ip_id_sequential_variable(ip_id_behavior_innermost.UVALUE,ip_id_indicator.CVALUE)
		c_base_header.ip_co_common->ip_id_indicator = 0;
		// =:= msn_lsb(8)
		*(mptr.uint8++) = c_msn_lsb(8,v2_context);
	}
	size = mptr.uint8 - c_base_header.uint8;
	// =:= crc7(THIS.UVALUE,THIS.ULENGTH)
	c_base_header.ip_co_common->header_crc = 0;
	c_base_header.ip_co_common->header_crc =
	   crc_calculate(CRC_TYPE_7,c_base_header.uint8,size,CRC_INIT_7,
	                 context->compressor->crc_table_7);
	goto end_code_ip_baseheader;

code_ip_pt_0_crc3:
	rohc_debugf(3, "code ip_pt_0_crc3\n");
	size = sizeof(ip_pt_0_crc3_t);
	// =:= '0'
	c_base_header.ip_pt_0_crc3->discriminator = 0;
	// =:= msn_lsb(4)
	c_base_header.ip_pt_0_crc3->msn = c_msn_lsb(4,v2_context);
	// =:= crc3(THIS.UVALUE,THIS.ULENGTH)
	c_base_header.ip_pt_0_crc3->header_crc = 0;
	c_base_header.ip_pt_0_crc3->header_crc =
	   crc_calculate(CRC_TYPE_3,c_base_header.uint8,size,CRC_INIT_3,
	                 context->compressor->crc_table_3);
	goto end_code_ip_baseheader;

code_ip_pt_0_crc7:
	rohc_debugf(3, "code ip_pt_0_crc7\n");
	size = sizeof(ip_pt_0_crc7_t);
	// =:= '100'
	c_base_header.ip_pt_0_crc7->discriminator = 0x04;
	// =:= msn_lsb(6)
	#if __BYTE_ORDER == __LITTLE_ENDIAN
	msn = c_msn_lsb(6,v2_context);
	c_base_header.ip_pt_0_crc7->msn1 = msn >> 1;
	c_base_header.ip_pt_0_crc7->msn2 = msn;
	#elif __BYTE_ORDER == __BIG_ENDIAN
	c_base_header.ip_pt_0_crc7->msn = c_msn_lsb(6,v2_context);
	#endif
	// =:= crc7(THIS.UVALUE,THIS.ULENGTH)
	c_base_header.ip_pt_0_crc7->header_crc = 0;
	c_base_header.ip_pt_0_crc7->header_crc =
	   crc_calculate(CRC_TYPE_7,c_base_header.uint8,size,CRC_INIT_7,
	                 context->compressor->crc_table_7);
	goto end_code_ip_baseheader;

code_ip_pt_1_seq_id:
	rohc_debugf(3, "code ip_pt_1_seq_id\n");
	size = sizeof(ip_pt_1_seq_id_t);
	// =:= '101'
	c_base_header.ip_pt_1_seq_id->discriminator = 0x05;
	// =:= msn_lsb(6)
	#if __BYTE_ORDER == __LITTLE_ENDIAN
	msn = c_msn_lsb(6,v2_context);
	c_base_header.ip_pt_1_seq_id->msn1 = msn >> 4;
	c_base_header.ip_pt_1_seq_id->msn2 = msn;
	#elif __BYTE_ORDER == __BIG_ENDIAN
	c_base_header.ip_pt_1_seq_id->msn = msn_lsb(6,v2_context);
	#endif
	c_base_header.ip_pt_1_seq_id->ip_id =
	   c_ip_id_lsb(ip_inner_context.vx->ip_id_behavior,4,3,ip_inner_context.v4->last_ip_id,ip_id,
	               v2_context->msn);
	// =:= crc3(THIS.UVALUE,THIS.ULENGTH)
	c_base_header.ip_pt_1_seq_id->header_crc = 0;
	c_base_header.ip_pt_1_seq_id->header_crc =
	   crc_calculate(CRC_TYPE_3,c_base_header.uint8,size,CRC_INIT_3,
	                 context->compressor->crc_table_3);
	goto end_code_ip_baseheader;

code_ip_pt_2_seq_id:
	rohc_debugf(3, "code ip_pt_2_seq_id\n");
	size = sizeof(ip_pt_2_seq_id_t);
	// =:= '110'
	c_base_header.ip_pt_2_seq_id->discriminator = 0x06;
	// =:=ip_id_lsb(ip_id_behavior_innermost.UVALUE, 6)
	#if __BYTE_ORDER == __LITTLE_ENDIAN
	ipid = c_ip_id_lsb(ip_inner_context.v4->ip_id_behavior,6,3,ip_inner_context.v4->last_ip_id,ip_id,
	                   v2_context->msn);
	c_base_header.ip_pt_2_seq_id->ip_id1 = ipid >> 1;
	c_base_header.ip_pt_2_seq_id->ip_id2 = ipid;
	#elif __BYTE_ORDER == __BIG_ENDIAN
	c_base_header.ip_pt_2_seq_id->ip_id =
	   c_ip_id_lsb(ip_inner_context.v4->ip_id_behavior,6,3,ip_inner_context.v4->last_ip_id,ip_id,
	               v2_context->msn);
	#endif
	// =:= msn_lsb(8)
	c_base_header.ip_pt_2_seq_id->msn = c_msn_lsb(8,v2_context);
	// =:= crc7(THIS.UVALUE,THIS.ULENGTH)
	c_base_header.ip_pt_2_seq_id->header_crc = 0;
	c_base_header.ip_pt_2_seq_id->header_crc =
	   crc_calculate(CRC_TYPE_7,c_base_header.uint8,size,CRC_INIT_7,
	                 context->compressor->crc_table_7);

end_code_ip_baseheader:
	#if ROHCV2_DEBUG
	rohc_debugf(3, "IP baseheader length %d\n",size);
	TraceData(c_base_header.uint8,size);
	#endif

	return c_base_header.uint8 + size;
}


/**
 * @brief Decide the state that should be used for the next packet compressed
 *        with the ROHCv2 profile.
 *
 * The three states are:
 *  - Initialization and Refresh (IR),
 *  - First Order (FO),
 *  - Second Order (SO).
 *
 * @param context The compression context
 */
#ifdef LMKQSJMQLJs
void tcp_decide_state(struct c_context *const context)
{
	struct c_generic_context *g_context;
	struct sc_v2_context *v2_context;

	g_context = (struct c_generic_context *) context->specific;
	v2_context = (struct sc_v2_context *) g_context->specific;

	rohc_debugf(3, "current state %d\n", context->state);

	if(v2_context->tmp.send_tcp_dynamic)
	{
		change_state(context, IR);
	}
	else
	{
		/* generic function used by the IP-only, UDP and UDP-Lite profiles */
		decide_state(context);
	}

	rohc_debugf(3, "next state %d\n", context->state);
}


#endif


/**
 * @brief Check if the dynamic part of the RoHCv2 header changed.
 *
 * @param context The compression context
 * @param tcp     The RoHCv2 header
 * @return        The number of RoHCv2 fields that changed
 */
#ifdef mLKJSQLMJS
int tcp_changed_tcp_dynamic(const struct c_context *context,
                            const tcphdr_t *tcp)
{
	const struct c_generic_context *g_context;
	struct sc_v2_context *v2_context;
	u_int32_t seq_number;

	g_context = (struct c_generic_context *) context->specific;
	v2_context = (struct sc_v2_context *) g_context->specific;

	// A REVOIR  DBX

//        rohc_debugf(3, "state %d\n",v2_context->state);
	rohc_debugf(3, "state %d\n",context->state);

	switch(context->state)
	{
		case IR:  /* The Initialization and Refresh (IR) state */
			break;
		case FO:  /* The First Order (FO) state */
			break;
		case SO:  /* The Second Order (SO) state */
			break;
	}

	seq_number = ntohl(tcp->seq_number);

	rohc_debugf(3, "seq_number 0x%08x last 0x%08x old 0x%08x change_count %d\n",
	            seq_number, v2_context->tcp_last_seq_number,
	            ntohl(v2_context->old_tcphdr.seq_number),
	            v2_context->tcp_seq_number_change_count);
/*
   if( seq_number != (v2_context->tcp_last_seq_number+1) )
   {
                v2_context->tcp_last_seq_number = seq_number;
      v2_context->tcp_seq_number_change_count = 0;
                rohc_debugf(3, "seq_number_change_count = 0 return 1\n");
                return 1;
   }
*/
	if(v2_context->tcp_seq_number_change_count < MAX_IR_COUNT)
	{
		rohc_debugf(3, "seq_number_change_count = %d return 1\n",
		            v2_context->tcp_seq_number_change_count);
		return 1;
	}

	v2_context->tcp_last_seq_number = seq_number;
	rohc_debugf(3, "seq_number_change_count = %d return 0\n",
	            v2_context->tcp_seq_number_change_count);

	return 0;
}


#endif


/**
 * @brief Define the compression part of the UDP profile as described
 *        in the RFC 3095.
 */
//struct c_profile c_rohc_v2_ip_profile =
c_v2_profile_t c_rohc_v2_ip_profile =
{
	{
		IPPROTO_IP,            /* IP protocol */
		NULL,             /* list of UDP ports, not relevant for UDP */
		ROHCv2_PROFILE_IP, /* profile ID (see RFC 5225 page 4 or 46) */
		"ROHCv2 IP / Compressor", /* profile description */
		c_rohc_v2_create,     /* profile handlers */
		c_generic_destroy,
		c_rohc_v2_check_context,
		c_rohc_v2_encode,
		c_generic_feedback,
	},
	sizeof(ip_context_t)
};

/**
 * @brief Define the compression part of the UDP profile as described
 *        in the RFC 3095.
 */
//struct c_profile c_rohc_v2_udp_profile =
c_v2_profile_t c_rohc_v2_udp_profile =
{
	{
		IPPROTO_UDP,      /* IP protocol */
		NULL,             /* list of UDP ports, not relevant for UDP */
		ROHCv2_PROFILE_UDP, /* profile ID (see RFC 5225 page 4 or 46) */
		"ROHCv2 UDP / Compressor", /* profile description */
		c_rohc_v2_create,     /* profile handlers */
		c_generic_destroy,
		c_rohc_v2_check_context,
		c_rohc_v2_encode,
		c_generic_feedback,
	},
	sizeof(udp_context_t)
};

/**
 * @brief Define the compression part of the RTP profile as described
 *        in the RFC 3095.
 */
extern int rtp_ports[];
//struct c_profile c_rohc_v2_rtp_profile =
c_v2_profile_t c_rohc_v2_rtp_profile =
{
	{
		IPPROTO_UDP,      /* IP protocol */
		rtp_ports,             /* list of UDP ports, not relevant for UDP */
		ROHCv2_PROFILE_RTP, /* profile ID (see RFC 5225 page 4 or 46) */
		"ROHCv2 RTP / Compressor", /* profile description */
		c_rohc_v2_create,     /* profile handlers */
		c_rohc_v2_destroy, // c_generic_destroy,
		c_rohc_v2_check_context,
		c_rohc_v2_encode,
		c_generic_feedback,
	},
	sizeof(udp_context_t) + sizeof(rtp_context_t)
};

/**
 * @brief Define the compression part of the UDP-Lite profile as described
 *        in the RFC 3095.
 */
//struct c_profile c_rohc_v2_udp_lite_profile =
c_v2_profile_t c_rohc_v2_udp_lite_profile =
{
	{
		IPPROTO_UDPLITE,      /* IP protocol */
		NULL,             /* list of UDP ports, not relevant for UDP */
		ROHCv2_PROFILE_UDPLITE, /* profile ID (see RFC 5225 page 4 or 46) */
		"ROHCv2 UDP-Lite / Compressor", /* profile description */
		c_rohc_v2_create,     /* profile handlers */
		c_generic_destroy,
		c_rohc_v2_check_context,
		c_rohc_v2_encode,
		c_generic_feedback,
	},
	sizeof(udp_lite_context_t)
};

/**
 * @brief Define the compression part of the ESP profile as described
 *        in the RFC 3095.
 */
//struct c_profile c_rohc_v2_esp_profile =
c_v2_profile_t c_rohc_v2_esp_profile =
{
	{
		IPPROTO_ESP,      /* IP protocol */
		NULL,             /* list of UDP ports, not relevant for UDP */
		ROHCv2_PROFILE_ESP, /* profile ID (see RFC 5225 page 4 or 46) */
		"ROHCv2 ESP / Compressor", /* profile description */
		c_rohc_v2_create,     /* profile handlers */
		c_generic_destroy,
		c_rohc_v2_check_context,
		c_rohc_v2_encode,
		c_generic_feedback,
	},
	sizeof(esp_context_t)
};

/**
 * @brief Define the compression part of the RTP profile as described
 *        in the RFC 3095.
 */
//struct c_profile c_rohc_v2_udp_lite_rtp_profile =
c_v2_profile_t c_rohc_v2_udp_lite_rtp_profile =
{
	{
		IPPROTO_UDPLITE,      /* IP protocol */
		NULL,             /* list of UDP ports, not relevant for UDP */
		ROHCv2_PROFILE_UDPLITE_RTP, /* profile ID (see RFC 5225 page 4 or 46) */
		"ROHCv2 RTP/UDP-Lite / Compressor", /* profile description */
		c_rohc_v2_create,     /* profile handlers */
		c_rohc_v2_destroy, // c_generic_destroy,
		c_rohc_v2_check_context,
		c_rohc_v2_encode,
		c_generic_feedback,
	},
	sizeof(udp_lite_context_t) + sizeof(rtp_context_t)
};

#endif // SUPPORT_ROHC_V2

